/*******************************************************************
 * This file is part of the Emulex Linux Device Driver for         *
 * Fibre Channel Host Bus Adapters.                                *
 * Copyright (C) 2003-2006 Emulex.  All rights reserved.           *
 * EMULEX and SLI are trademarks of Emulex.                        *
 * www.emulex.com                                                  *
 *                                                                 *
 * This program is free software; you can redistribute it and/or   *
 * modify it under the terms of version 2 of the GNU General       *
 * Public License as published by the Free Software Foundation.    *
 * This program is distributed in the hope that it will be useful. *
 * ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND          *
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE      *
 * DISCLAIMED, EXCEPT TO THE EXTENT THAT SUCH DISCLAIMERS ARE HELD *
 * TO BE LEGALLY INVALID.  See the GNU General Public License for  *
 * more details, a copy of which can be found in the file COPYING  *
 * included with this package.                                     *
 *******************************************************************/

/* 
 * $Id: mpdevid.h 1005 2006-01-27 17:27:22Z hdaschbach $
 */

#ifndef  EMULEX_MP_DEVICE_ID_INCLUDED
#define  EMULEX_MP_DEVICE_ID_INCLUDED

#ifdef __cplusplus
extern "C" {
#endif


#define  ELX_DI_INVALID_DEVID_LENGTH         (-1)    /* return value if length is invalid            */
#define  ELX_DI_SUCCESS                       0      /* return value if operation was successful     */
#define  ELX_DI_NOT_ENOUGH_MEMORY             1      /* return value if not enough memory for dev id */
#define  ELX_DI_LOOKUP_NOT_AVAILABLE          2      /* a required lookup value was not available    */
#define  ELX_DI_NOT_A_DEVICE_IDENTIFIER       3      /* the string was not a device identifier       */
#define  ELX_DI_UNSUPPORTED_DEVICE_IDENTIFIER 4      /* the dev id is not valid for this operation   */

#define  ELX_DI_DEVID_INQ_LEN        32
#define  ELX_DI_DEVID_MAX_LEN        64

#define  ELX_DI_MAX_DEVID_STRING_LEN   1024

/* m_type values */
#define ELX_DI_DEVID_INQ             0x01            /* Device by Inquiry information                */
#define ELX_DI_DEVID_INQSN           0x02            /* Device by Inquiry Serial Number              */
#define ELX_DI_DEVID_WWPN            0x03            /* Device by World Wide Port Name               */
#define ELX_DI_DEVID_WWNN            0x04            /* Device by World Wide Node Name               */
#define ELX_DI_DEVID_NPORTID         0x05            /* Device by NPort Id                           */
#define ELX_DI_DEVID_SCSI_NEXUS      0x06            /* Device by adapter / bus / tgt / lun          */
#define ELX_DI_DEVID_ASCII           0x07            /* ASCII string                                 */
#define ELX_DI_DEVID_NEXUS_LONG      0x08            /* SCSI nexus by fibre channel names            */
#define ELX_DI_DEVID_ISCSI           0x09            /* Device by iSCSI address string               */
#define ELX_DI_DEVID_ISCSI_NEXUS     0x0a            /* Device by iSCSI nexus                        */
#define ELX_DI_DEVID_ISCSI_SERIAL    0x0b            /* Device by iSCSI serial number                */
#define ELX_DI_DEVID_ARRAY           0xff            /* this entry describes an array of DevIds      */

typedef unsigned short ELX_DI_DEVID_TYPE;

typedef struct ElxDiDevId {
   ELX_DI_DEVID_TYPE m_type;             /* type of identification used          */
   unsigned short m_dflags;              /* Device flags                         */
#define  ELX_DI_DEVID_FLAGS_BINARY         0x0001
#define  ELX_DI_DEVID_FLAGS_LOOKUP         0x0002
   unsigned short m_len;                 /* Length of this Devid                 */
   union {
      unsigned char m_InquiryData[ELX_DI_DEVID_INQ_LEN];    /* device serial number */
      struct {
         unsigned short m_adapter;       /* adapter number device is accessed thru                    */
         unsigned short m_bus;           /* bus number adapter uses to access device                  */
         unsigned short m_target;        /* scsi target number assigned to device accessed thru path  */
         unsigned short m_lun;           /* scsi LUN number assigned to device accessed thru  path    */
      } m_ScsiId;
      unsigned char m_wwn[8];            /* World Wide Name of device                                 */
      unsigned int m_nportid;            /* Fibre Channel NPort Id of device                          */
      unsigned char m_Raw[ELX_DI_DEVID_MAX_LEN];
      struct {
         unsigned char m_adapter_wwpn[8];
         unsigned char m_target_wwpn[8];
         unsigned char m_lun[8];
      } m_NexusLong;
      unsigned char m_session_name[1];       /* variable length session             */

      /* other types of device authentication, for future use, can be added here    */
   } un;
} elxDiDeviceIdentifier;


/*
// structure for a single element of the list of key to data translations 
// for device identifiers
*/

typedef struct ElxDiDevIdLookup {
   struct ElxDiDevIdLookup *Next;
   unsigned long DataLength;
   unsigned long KeyLength;
   char Key[32];
   char Data[1];
} ELX_DI_DEVID_LOOKUP, *PELX_DI_DEVID_LOOKUP;


/***************************************************************************
 *
 * elxDiDevIdCompare
 *
 * Compare two device identifiers 
 *
 * Parameters:
 *
 * Id1 -    Pointer to the first Device Identifier
 *
 * Id2 -    Pointer to the second Device Identifier
 *
 * Return Values:
 *
 * Returns nonzero if the Device Identifiers are identical, returns 0 if the
 * Device Identfiers differ.
 *
 * Comments:
 *
 * elxDiDevIdCompare does not do a binary compare of the entire buffer. Only
 * the pertinent information is checked, thus the two buffers for the Device
 * Identifiers may contain somewhat different data and still contain 
 * equivalent Device Identifiers.
 *
 ***************************************************************************/

int elxDiDevIdCompare(elxDiDeviceIdentifier *Id1, 
                      elxDiDeviceIdentifier *Id2);



                      
/***************************************************************************
 *
 * elxDiFormatDevId
 *
 * Convert the Device Identifier to a zero terminated ASCII string.
 *
 * Parameters:
 *
 * buf -       Buffer to contain the formatted Device Identifier.
 *
 * DevId -     A pointer to the Device Identifier to be formatted.
 *
 * buflen -    The number of bytes at buf to contain the result.
 *
 * NoLookUp -  Set to TRUE if the lookup function is not to be performed
 *             for Device Identifiers with the ELX_DI_DEVID_FLAGS_LOOKUP
 *             set.
 *
 * Return Values:
 *
 * Returns ELX_DI_SUCCESS if the Device Identifier was successfully formatted,
 * and ELX_DI_NOT_ENOUGH_MEMORY if the buffer at buf is not large enough to 
 * hold the formatted Device Identifier.
 *
 * Comments:
 *
 * The formatted Device Identifier is displayed as follows:
 *
 * <type>:<length>:<flags>:<data>
 *
 * <type> is the Device Identifier type. <length> is the length in bytes of 
 * the raw data field. <flags> are the flags for this Device Identifier. 
 * <data> is the raw data formatted according to <type>.
 *
 * If ELX_DI_DEVID_FLAGS_LOOKUP is set, then the NoLookUp parameter is 
 * checked. If the NoLookUp parameter is TRUE, elxDiFormatDevId will format
 * the lookup handle that is stored as part of the Device Identifier. If
 * the NoLookUp parameter is FALSE, elxDiFormatDevId formats the data stored 
 * in the lookup data for this Device Identifier, assuming it is present.
 *
 ***************************************************************************/

int elxDiFormatDevId(char *buf, 
                     elxDiDeviceIdentifier *DevId, 
                     int buflen, 
                     int NoLookup);



                      
/***************************************************************************
 *
 * elxDiCopyDevId
 *
 * Copy a Device Identifier to another Device Identifer.
 *
 * Parameters:
 *
 * Src -    A pointer to the source Device Identifier.
 *
 * Dst -    A pointer to a buffer to contain the copied Device Identifier.
 *
 * DestinationSize - The size of the buffer at Dst.
 *
 * Return Values:
 *
 * Returns ELX_DI_SUCCESS if the Device Identifier was successfully copied,
 * and ELX_DI_NOT_ENOUGH_MEMORY if the buffer at Dst is not large enough to 
 * hold the Device Identifier.
 *
 * Comments:
 *
 * Because elxDiCopyDevId only copies the pertinent information to duplicate 
 * the Device Identifier, a binary compare of the two buffers after a 
 * successful completion of elxDiCopyDevId may show some differences.
 *
 ***************************************************************************/

int elxDiCopyDevId(elxDiDeviceIdentifier *Src,
                   elxDiDeviceIdentifier *Dst,
                   int DestinationSize);




                      
/***************************************************************************
 *
 * elxDiStringToDevId
 *
 * Convert a zero terminated ASCII string that is a formatted Device
 * Identifier into a Device Identifier.
 *
 * Parameters:
 *
 * buf -    A pointer to the formatted Device Identifier string.
 *
 * DevId -  A pointer to a buffer to contain the Device Identifier.
 *
 * Return Values:
 *
 * Returns ELX_DI_SUCCESS if the Device Identifier was successfully copied,
 * and an error code if elxDiStringToDevId was not successful.
 *
 *
 ***************************************************************************/

int elxDiStringToDevId(const unsigned char *buf, 
                       elxDiDeviceIdentifier *DevId);






                      
/***************************************************************************
 *
 * elxDiInitializeDevId
 *
 * Initialize a Device Identifier to an empty Device Identifier.
 *
 * Parameters:
 *
 * _d -     A pointer to an empty Device Identifier structure.
 *
 * _l -     The size of the buffer at _d.
 *
 * Return Values:
 *
 * None.
 *
 * Comments:
 *
 * elxDiInitializeDevId sets the control fields in the Device Identifier
 * to indicate an empty Device Identifier. The m_buf_len field is set to 
 * indicate the amount of memory the Device Identifier has available.
 *
 ***************************************************************************/

#define elxDiInitializeDevId(_d, _l) {_d->m_type = 0; _d->m_len = _l;}






                      
/***************************************************************************
 *
 * elxDiDevIdActualLength
 *
 * Get the minimum amount of memory in bytes needed to contain the Device 
 * Identifier.
 *
 * Parameters:
 *
 * _d -     A pointer to the Device Identifier structure.
 *
 * Return Values:
 *
 * The minimum amount of memory in bytes needed to contain the Device 
 * Identifier.
 *
 ***************************************************************************/

#define elxDiDevIdActualLength(_d) \
               (((_d)->m_len > ELX_DI_DEVID_MAX_LEN) ? \
                  (_d)->m_len + (int)(&((elxDiDeviceIdentifier *)0)->un) : \
                  sizeof(elxDiDeviceIdentifier))





                      
/***************************************************************************
 *
 * elxDiAddLookupInfo
 *
 * Add a lookup key and data to the list of elements to be searched when
 * a lookup is required.
 *
 * Parameters:
 *
 * Lookup -     A pointer to a ELX_DI_DEVID_LOOKUP structure containing the 
 *              information to be added to the lookup list
 *
 * Return Values:
 *
 * None.
 *
 * Comments:
 *
 * elxDiAddLookupInfo and elxDiFreeLookupInfo must not be called 
 * simultaneously by different threads. Serialization must be performed by
 * the caller to prevent different threads from entering these functions
 * simultaneously.
 *
 ***************************************************************************/

void elxDiAddLookupInfo(PELX_DI_DEVID_LOOKUP Lookup);





                      
/***************************************************************************
 *
 * elxDiFreeLookupInfo
 *
 * Free the lookup info list.
 *
 * Parameters:
 *
 * FreeFunc -     A pointer to a function that is called for each Lookup
 *                entry, so the caller can free the memory containing the
 *                entry.
 *
 * Return Values:
 *
 * None.
 *
 * Comments:
 *
 * FreeFunc is called for each entry. The caller should release the memory
 * for the entry.
 *
 ***************************************************************************/

void elxDiFreeLookupInfo(void (*FreeFunc)(PELX_DI_DEVID_LOOKUP Lookup));





                      
/***************************************************************************
 *
 * elxDiFindLookupByKey
 *
 * Find a Lookup information record by its key.
 *
 * Parameters:
 *
 * Key -       Pointer to the key
 *
 * KeyLength - Length, in bytes, of the key.
 *
 * Return Values:
 *
 * A pointer to the lookup information record, if one is found that matches
 * the key, NULL if no record is found.
 *
 * Comments:
 *
 * The caller should take care to serialize this function in respect to
 * elxDiAddLookupInfo and elxDiFreeLookupInfo.
 *
 ***************************************************************************/

PELX_DI_DEVID_LOOKUP elxDiFindLookupByKey(const unsigned char *Key, 
                                          unsigned long KeyLength);





                      
/***************************************************************************
 *
 * elxDiFindLookupByData
 *
 * Find a Lookup information record by its data.
 *
 * Parameters:
 *
 * Data -       Pointer to the data.
 *
 * DataLength - Length, in bytes, of the data.
 *
 * Return Values:
 *
 * A pointer to the lookup information record, if one is found that matches
 * the data, NULL if no record is found.
 *
 * Comments:
 *
 * The caller should take care to serialize this function in respect to
 * elxDiAddLookupInfo and elxDiFreeLookupInfo.
 *
 ***************************************************************************/

PELX_DI_DEVID_LOOKUP elxDiFindLookupByData(const unsigned char *Data, 
                                           unsigned long DataLength);





                      
/***************************************************************************
 *
 * elxDiComparePortDIDtoPathDID
 *
 * Check to see if a Path Device Identifer belongs to a port Device 
 * Identifier.
 *
 * Parameters:
 *
 * PortId -    Pointer to the Device Identifier for the Port
 *
 * PathId -    Pointer to the Device Identifier for the Path
 *
 * Return Values:
 *
 * Returns non-zero if the path is a child of the port, returns 0 if the
 * path is not a child of the port.
 *
 ***************************************************************************/

int elxDiComparePortDIDtoPathDID(elxDiDeviceIdentifier *PortId,
                                 elxDiDeviceIdentifier *PathId);





                      
/***************************************************************************
 *
 * elxDiComparePortTargettoPathTarget
 *
 * Check to see if a Path Device Identifer belongs to a port Device 
 * Identifier via a specific target..
 *
 * Parameters:
 *
 * PortId -    Pointer to the Device Identifier for the Port
 *
 * PathId -    Pointer to the Device Identifier for the Path
 *
 * Return Values:
 *
 * Returns non-zero if the path is a child of the port, returns 0 if the
 * path is not a child of the port.
 *
 ***************************************************************************/

int elxDiComparePortTargettoPathTarget(elxDiDeviceIdentifier *Dev1,
                                       elxDiDeviceIdentifier *Dev2);






int elxDiMakeDevId(elxDiDeviceIdentifier *DevId,        
                   ELX_DI_DEVID_TYPE Type,
                   unsigned char Flags,
                   unsigned short RawDataLength,
                   unsigned long DevIdBufferLength,
                   const unsigned char *DevIdData);

int elxDiRequiredDevIdLength(ELX_DI_DEVID_TYPE Type, ...);

int elxDiRequiredDevIdLengthFromString(const unsigned char *DevId);

#ifdef __cplusplus
}
#endif
#endif
