/*******************************************************************
 * This file is part of the Emulex Linux Device Driver for         *
 * Fibre Channel Host Bus Adapters.                                *
 * Copyright (C) 2003-2006 Emulex.  All rights reserved.           *
 * EMULEX and SLI are trademarks of Emulex.                        *
 * www.emulex.com                                                  *
 *                                                                 *
 * This program is free software; you can redistribute it and/or   *
 * modify it under the terms of version 2 of the GNU General       *
 * Public License as published by the Free Software Foundation.    *
 * This program is distributed in the hope that it will be useful. *
 * ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND          *
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE      *
 * DISCLAIMED, EXCEPT TO THE EXTENT THAT SUCH DISCLAIMERS ARE HELD *
 * TO BE LEGALLY INVALID.  See the GNU General Public License for  *
 * more details, a copy of which can be found in the file COPYING  *
 * included with this package.                                     *
 *******************************************************************/

/* 
 * $Id: mpl_hpaa.c 1005 2006-01-27 17:27:22Z hdaschbach $
 */

#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/blk.h>
#include <linux/utsname.h>

/* From drivers/scsi */
#include <sd.h>
#include <hosts.h>
#include <scsi.h>
#else
#include <linux/pci.h>
#include <scsi/scsi.h>
#include <scsi/scsi_device.h>
#include <scsi/scsi_cmnd.h>
#endif

#include <linux/string.h>

#include "mpl.h"
#include "mpl_crtn.h"
#include "mpl_log.h"

extern struct scsi_id_search_values mpl_id_search_list[12];
extern struct list_head mpl_hosts;

/* For now MSA (active-active), and EVA (active-active)
 * will share the same plugin. A default HP active-active plugin.
 * Hopefully we can migrate XP into here as well.
 */
static uint8_t msa0_str[] = "MSA CONTROLLER";
static uint8_t msaX_str[] = "MSA VOLUME";
static uint8_t hsv200_str[] = "HSV200";
static uint8_t hsv210_str[] = "HSV210";
static uint8_t hsv101_str[] = "HSV101";
static uint8_t hsv111_str[] = "HSV111";

extern int mpl_enable_nopath;

static int
mpl_hpaa_probe(struct scsi_cmnd *cmd)
{
	uint8_t vendor[8 + 1];
	uint8_t product[16 + 1];
	uint8_t rev[4 + 1];
	uint8_t *prod = (uint8_t *)cmd->request_buffer+16;

	if (!mpl_vendor_hp_check(cmd)) {
		return 0;
	}

	if ((memcmp(prod, msa0_str, strlen(msa0_str)-1) == 0) ||
	    (memcmp(prod, msaX_str, strlen(msaX_str)-1) == 0) ||
	    (memcmp(prod, hsv101_str, strlen(hsv101_str)-1) == 0) ||
	    (memcmp(prod, hsv111_str, strlen(hsv111_str)-1) == 0) ||
	    (memcmp(prod, hsv200_str, strlen(hsv200_str)-1) == 0) ||
	    (memcmp(prod, hsv210_str, strlen(hsv210_str)-1) == 0)) {
		trim_to(vendor, cmd->request_buffer + 8, sizeof(vendor));
		trim_to(product, cmd->request_buffer + 16, sizeof(product));
		trim_to(rev, cmd->request_buffer + 32, sizeof(rev));
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
		mpl_print(MPL_INIT, MPL_INFO " hpaa: plugin support for: "
		  	"tgt %d lun %d: %s, %s, %s\n",
		  	cmd->target, cmd->lun,
			vendor, product, rev);
#else
		mpl_print(MPL_INIT, MPL_INFO " hpaa: plugin support for: "
		  	"tgt %d lun %d: %s, %s, %s\n",
		  	cmd->device->id, cmd->device->lun,
			vendor, product, rev);
#endif
		return 1;
	} else {
		return 0;
	}
}

static void
mpl_hpaa_path_down(struct mpl_path *path)
{
	mpl_print(MPL_RECOVERY, MPL_INFO "Path Down Plugin %s %s\n",
		mpl_print_device(path->device), mpl_print_path(path));
	path->flag &= ~(PATH_FAIL_OVER_EVT|PATH_FAIL_BACK_EVT);
	if ((path->sas == MPL_ALUA_ACTIVE_OPT) &&
	    (atomic_read(&path->device->opt_path_cnt))) {
		atomic_dec(&path->device->opt_path_cnt);
	}
	path->sas = MPL_ALUA_UNAVAILABLE;
	return;
}

static int
mpl_hpaa_path_up(struct mpl_path *path)
{
	struct mpl_device *device = path->device;
	int fail = 0;

	mpl_print(MPL_RECOVERY, MPL_INFO "Path Up Plugin %s %s\n",
		mpl_print_device(device), mpl_print_path(path));

	/* Check if path_up is being called as a result of a
	 * Failover or Failback.
	 */
	if (path->flag & (PATH_FAIL_OVER_EVT|PATH_FAIL_BACK_EVT))
		fail = 1;
	path->flag &= ~(PATH_FAIL_OVER_EVT|PATH_FAIL_BACK_EVT);

	if (device->scsi_type) {
		path->sas = MPL_ALUA_ACTIVE_NOT_OPT;
		return 1;
	}

	if ((path->sas == MPL_ALUA_ACTIVE_OPT) ||
	     device->tpg_cmd_inp ||
	    (path != device->last_path)) {
		return 1;
	}

#if 0
/* Wait till EVA / MSA firmware TPG issues are addressed before 
 * we turn this code on.
 */
	if (fail) {
		/* Defer the PSM action till we set the new path to optimized */

		mpl_print(MPL_ALUA, MPL_INFO
			"Issue SET TGT PORT GROUP %d:  %s %s sas %d\n",
			path->tgt_port_group, mpl_print_device(device),
			mpl_print_path(path), path->sas);

		path->flag |= (PATH_NEEDS_SET_TGT_PORT | PATH_DEV_QUIESCE |
		       PATH_OUT_CHG_LUN_OWN_CMD);
		path->retries = 1;
		device->tpg_cmd_inp = 1;
		mpl_restart_sched_cmd(path, jiffies + 1);

		/* We will still bring the path up and start IO before this
		 * command completes.
		 */
	}
#endif

	return 1;
}


static void
mpl_hpaa_process_stpg(struct io_stub *stub)
{
	struct mpl_path *path = stub->path;
	struct mpl_device *device = path->device;
	struct scsi_cmnd *cmd = stub->cmd;
	struct io_stub *stpg_stub;
	struct mpl_host *h;
	struct mpl_path *p;
	struct mpl_device *d;

	mpl_print(MPL_ALUA, MPL_INFO
		"Cmpl SET TGT PORT GROUP %d:  %s %s result %d\n",
		path->tgt_port_group, mpl_print_device(device),
		mpl_print_path(path), cmd->result);

	if (cmd->result == 0) {

		mpl_print(MPL_INTERNAL_CMD, MPL_INFO "Cmpl Set TPG: %s %s\n",
		  	mpl_print_device(device),
		  	mpl_print_path(path));
		path->retries = 0;

		/* Kick off hbeat to update path->sas */
		list_for_each_entry(h, &mpl_hosts, host_list) {
			list_for_each_entry(d, &h->devices, devices) {
				list_for_each_entry(p, &d->paths, path_list) {
					/* Kick off hbeat to update p->sas */
					if (path->hbstate == HB_WAITING) {
						mpl_hbeat_state_machine(p, HB_START);
					}
				}
			}
		}
	}
	else {
		mpl_print(MPL_INTERNAL_CMD,
			 MPL_INFO "Bad Set TPG status x%x: %s %s retry %d\n",
			cmd->result,
		  	mpl_print_device(device),
		  	mpl_print_path(path), path->retries);

		if ((path->sas != MPL_ALUA_ACTIVE_OPT) &&
	    	    (path == device->last_path) &&
		    (path->retries < 3) && 
			 (cmd->result != ScsiResult(DID_NO_CONNECT, 0))) {
			stpg_stub = mpl_setup_set_tgt_port(path,
				path->tgt_port_group, MPL_ALUA_ACTIVE_OPT);
			if(stpg_stub) {
				path->retries++;
				path->flag |= PATH_OUT_CHG_LUN_OWN_CMD;
				mpl_add_deferral(stpg_stub, MPL_INTERNAL_CMD);
			}
			return;
		}
		path->retries = 0;
	}
	return;
}

static int
mpl_hpaa_diagnose(struct io_stub *stub, int flag)
{
	struct scsi_cmnd *cmd = stub->cmd;
	struct mpl_path *path = stub->path;
	struct mpl_device *device = path->device;
	uint8_t sk, asc, ascq;
	int rv = 0;

	mpl_print(MPL_RECOVERY, MPL_INFO "Diagnose Plugin "
		"%s %s%scmd %x result %x\n",
		mpl_print_device(device), mpl_print_path(path),
		(stub->done) ? " " : " internal ",
		(unsigned int)cmd->cmnd[0], cmd->result);

	sk = cmd->sense_buffer[2] & 0xF;
	asc = cmd->sense_buffer[12];
	ascq = cmd->sense_buffer[13];

	/* Setup alua / active-passive on first command */
	device->alua_support = 1; /* supports ALUA mode */

	switch (flag) {
	case PLUG_INTERNAL_CMD:
		/* mpl driver ignores the return value for this case */
		switch (cmd->cmnd[0]) {
		case SET_TGT_PORT_GROUP:
			if (path->flag & PATH_OUT_CHG_LUN_OWN_CMD) {
				path->flag &= ~PATH_OUT_CHG_LUN_OWN_CMD;
				mpl_hpaa_process_stpg(stub);
			}
			break;
		}
		break;

	case PLUG_HBEAT_CMD:
		/* mpl driver assumes a return value of 0 means the hbeat
		 * command succeeded.
		 */
		switch (cmd->cmnd[0]) {
		case INQUIRY:
			/* Setup tgt qdepth after first INQUIRY */
			/* Assume 80% utilization of max (2047) */
			device->tgt_queue_depth = 1637; 
			device->active_passive = 0;  /* active - active */
 			if (mpl_enable_nopath && !device->nopath_tmo) {
				/* This only applies to EVAs */
				if ((memcmp(device->product, "HSV", 3) == 0)) {
					device->nopath_tmo = 60;
				}
			}
			/* Drop thru to set rv */
		case TEST_UNIT_READY:
		case REPORT_TGT_PORT_GROUP:

			if(cmd->result)
				rv = 1; /* error */
		}
		break;

	case PLUG_SCSI_CMD:
		/* mpl driver assumes a return value of 1 means to initiate
		 * a failover and retry the command.
		 */
		break;
	}
	return rv;
}

/* Return 0 if we find a valid validation id, else return 1 */
int mpl_hpaa_validation_id(
	struct mpl_path *path,
	uint8_t *page_83,
	uint8_t *serial, int *serial_size)
{
	uint32_t offset = 0;
	uint32_t size = 0, relport;
	uint16_t total_size, id_size;
	uint8_t *idptr, type, tpcnt, old_sas;
	uint16_t tpg;
	uint32_t *idwptr;

	total_size = *((uint16_t *) &page_83[2]);
	total_size = be16_to_cpu(total_size);
	offset = 4;

	path->flag &= ~PATH_ASSIGN_TPG;
	while (offset < total_size) {
		idptr = page_83 + offset;
		id_size = (uint16_t) *(idptr+3);
		offset += (id_size + 4);
		
		idwptr = (uint32_t *)(idptr + 4);

		type = *(idptr+1);
		if (((type & 0xf) == 5) && (id_size == sizeof(uint32_t))) {
			size = *idwptr;
			path->tgt_port_group = be32_to_cpu(size);
			path->flag |= PATH_ASSIGN_TPG;
			mpl_print(MPL_ALUA, MPL_INFO "Tgt Port Group: x%x"
				" %s %s\n",
				path->tgt_port_group,
				mpl_print_device(path->device),
				mpl_print_path(path));
		}
		if (((type & 0xf) == 4) && (id_size == sizeof(uint32_t))) {
			size = *idwptr;
			path->relative_tgt_port = be32_to_cpu(size);
			mpl_print(MPL_ALUA, MPL_INFO "Relative Tgt Port: x%x"
				" %s %s\n",
				path->relative_tgt_port,
				mpl_print_device(path->device),
				mpl_print_path(path));
		}

		if (id_size == 0)
			break;
	}

	old_sas = path->sas;
	if (path->tgt_port_info) {
		idwptr = (uint32_t *)path->tgt_port_info;
		size = *idwptr;
		size = be32_to_cpu(size);
		offset = 0;
		idptr = (uint8_t *)(idwptr + 1);
		while (offset < size) {
			type = *(idptr + offset);
			offset += 2;
			tpg = *((uint16_t *)(idptr + offset));
			tpg = be16_to_cpu(tpg);
			offset += 5;
			tpcnt = *(idptr + offset);
			offset++;
			/* For our tpg, set sas if paths
			 * asymmetric access state is optimised
			 */
			if (tpg == path->tgt_port_group) {
				path->sas = (type & MPL_ALUA_SAS_MASK);
				mpl_print(MPL_ALUA, MPL_INFO "Asymetric Access State "
		  		" %s %s: %d\n",
		  		mpl_print_device(path->device), mpl_print_path(path),
		  		path->sas);
			}
			idwptr = (uint32_t *)(idptr + offset);
			while (tpcnt--) {
				relport = *idwptr++;
				relport = be32_to_cpu(relport);
				offset += 4;
			}
		}
	}

	if (old_sas != path->sas) {
		if (path->sas == MPL_ALUA_ACTIVE_OPT) {
			atomic_inc(&path->device->opt_path_cnt);
		}
		else {
			if ((old_sas == MPL_ALUA_ACTIVE_OPT) &&
			    (atomic_read(&path->device->opt_path_cnt))) {
				atomic_dec(&path->device->opt_path_cnt);
			}
		}
		mpl_print(MPL_ALUA, MPL_INFO "Asymetric Access State Change"
		  " %s %s: tpg %d From %d to %d optcnt %d\n",
		  mpl_print_device(path->device), mpl_print_path(path),
		  path->tgt_port_group, old_sas, path->sas,
		  atomic_read(&path->device->opt_path_cnt));
	}

	return mpl_get_scsi_page83_serial(page_83, mpl_id_search_list,
		sizeof(mpl_id_search_list)/sizeof(mpl_id_search_list[0]),
		serial, serial_size);
}

/* DEFINE_PLUGIN(id, name, probe, prequeue, diagnose, down, up, valid)
 */

DEFINE_PLUGIN(hpaa, "hpaa", mpl_hpaa_probe, NULL, mpl_hpaa_diagnose, 
		mpl_hpaa_path_down, mpl_hpaa_path_up, mpl_hpaa_validation_id)



