/*******************************************************************
 * This file is part of the Emulex Linux Device Driver for         *
 * Fibre Channel Host Bus Adapters.                                *
 * Copyright (C) 2003-2006 Emulex.  All rights reserved.           *
 * EMULEX and SLI are trademarks of Emulex.                        *
 * www.emulex.com                                                  *
 *                                                                 *
 * This program is free software; you can redistribute it and/or   *
 * modify it under the terms of version 2 of the GNU General       *
 * Public License as published by the Free Software Foundation.    *
 * This program is distributed in the hope that it will be useful. *
 * ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND          *
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE      *
 * DISCLAIMED, EXCEPT TO THE EXTENT THAT SUCH DISCLAIMERS ARE HELD *
 * TO BE LEGALLY INVALID.  See the GNU General Public License for  *
 * more details, a copy of which can be found in the file COPYING  *
 * included with this package.                                     *
 *******************************************************************/

/* 
 * $Id: mpl_jbod.c 1005 2006-01-27 17:27:22Z hdaschbach $
 */

#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/blk.h>
#include <linux/utsname.h>

/* From drivers/scsi */
#include <sd.h>
#include <hosts.h>
#include <scsi.h>
#else
#include <linux/pci.h>
#include <scsi/scsi.h>
#include <scsi/scsi_device.h>
#include <scsi/scsi_cmnd.h>
#endif

#include <linux/string.h>

#include "mpl.h"
#include "mpl_crtn.h"
#include "mpl_log.h"
#include "mpl_mem.h"

extern void mpl_fail_device(struct mpl_device *);
extern void mpl_recovery_on_next_path(struct mpl_path *);
extern void mpl_set_path_state_n_hbeat(struct mpl_path *, 
	enum mpl_path_state);

extern int  stub_hash_size;
extern struct io_stub **stub_hash;
extern spinlock_t stub_hash_table_lock;
extern volatile int driver_unloading;

static uint8_t jbod1_str[] = "SEAGATE ST318451FC";
static uint8_t jbod2_str[] = "SEAGATE ST318452FC";
static uint8_t jbod3_str[] = "SEAGATE ST318304FCV";

static int
mpl_jbod_probe(struct scsi_cmnd *cmd)
{
	uint8_t vendor[8 + 1];
	uint8_t product[16 + 1];
	uint8_t rev[4 + 1];
	uint8_t *prod = (uint8_t *)cmd->request_buffer+8;

	if ((memcmp(prod, jbod1_str, strlen(jbod1_str)-1) == 0) || 
	    (memcmp(prod, jbod2_str, strlen(jbod2_str)-1) == 0) ||
	    (memcmp(prod, jbod3_str, strlen(jbod2_str)-1) == 0)) { 
		trim_to(vendor, cmd->request_buffer + 8, sizeof(vendor));
		trim_to(product, cmd->request_buffer + 16, sizeof(product));
		trim_to(rev, cmd->request_buffer + 32, sizeof(rev));
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
		mpl_print(MPL_INIT, MPL_INFO " jbod: plugin support for: "
		  	"tgt %d lun %d: %s, %s, %s\n",
		  	cmd->target, cmd->lun,
			vendor, product, rev);
#else
		mpl_print(MPL_INIT, MPL_INFO " jbod: plugin support for: "
		  	"tgt %d lun %d: %s, %s, %s\n",
		  	cmd->device->id, cmd->device->lun,
			vendor, product, rev);
#endif
		return 1;
	} else {
		return 0;
	}
}

#if 0
/* Get rid of abort for now, assume tgt reset will abort any outstanding IO */
static void
mpl_jbod_abort_cmds(struct mpl_path *path)
{
	struct io_stub *stub, *prev;
	unsigned long flags;
	int i;

	/* Abort any outstanding commands */
	spin_lock_irqsave(&stub_hash_table_lock, flags);
	for(i=0;i<stub_hash_size;i++) {
		stub = stub_hash[i];
		while(stub) {
			prev = stub;
			stub = stub->next;
			if (prev && prev->path == path) {
				spin_unlock_irqrestore(&stub_hash_table_lock,
					flags);
				mpl_add_eh_deferral(prev, MPL_DO_ABORT);
				spin_lock_irqsave(&stub_hash_table_lock, flags);
			}
		}
	}
	spin_unlock_irqrestore(&stub_hash_table_lock, flags);
	return;
}
#endif

static void
mpl_jbod_path_down(struct mpl_path *down_path)
{
	struct mpl_device *device = down_path->device;
	struct io_stub *stub = NULL;
	struct mpl_path *path;
	struct scsi_cmnd *cmd;

	mpl_print(MPL_RECOVERY, MPL_INFO "Path Down Plugin %s %s\n",
		mpl_print_device(down_path->device), mpl_print_path(down_path));

	if (driver_unloading)
		return;

	cmd = kmalloc(sizeof(struct scsi_cmnd), GFP_ATOMIC);
	if(cmd) {
		list_for_each_entry(path, &device->paths, path_list) {
			/* First reset the device */
			if ((path->state != PATH_FAILED) && (path != down_path)) {
				memset(cmd, 0, sizeof(struct scsi_cmnd));
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
				cmd->host = path->host->scsi_host;
				cmd->channel = path->channel;
				cmd->target = path->target;
				cmd->lun = path->lun;
#else
				cmd->device = path->sdev;
#endif
				if((stub = mpl_stub_alloc(cmd, path, 0))) {
					goto abtit;
				}
				break;
			}
		}
		kfree(cmd);
	}

abtit:
#if 0
	/* Get rid of abort for now, assume tgt reset will abort any outstanding IO */

	/* next abort any outstanding commands */
	list_for_each_entry(path, &device->paths, path_list) {
		mpl_jbod_abort_cmds(path);
	}
#endif
	if(stub) {
		mpl_add_eh_deferral(stub, MPL_DO_DEV_RESET);
	}

	return;
}

static int
mpl_jbod_path_up(struct mpl_path *path)
{
	mpl_print(MPL_RECOVERY, MPL_INFO "Path Up Plugin %s %s\n",
		mpl_print_device(path->device), mpl_print_path(path));

	return 1;
}


static int
mpl_jbod_diagnose(struct io_stub *stub, int flag)
{
	struct mpl_path *path = stub->path;
	struct scsi_cmnd *cmd = stub->cmd;
	int rv = 0;

	mpl_print(MPL_RECOVERY, MPL_INFO "Diagnose Plugin "
		"%s %s%scmd %x result %x\n",
		mpl_print_device(path->device), mpl_print_path(path),
		(stub->done) ? " " : " internal ",
		(unsigned int)cmd->cmnd[0], cmd->result);


	/* Setup alua / active-passive on first command */
	/* Make it think we support active-active ALUA, but don't
	 * set the alua support flag in the device structure.
	 */
	path->tpgs = MPL_ALUA_ACTIVE_ACTIVE;
	path->sas  = MPL_ALUA_ACTIVE_OPT;
	path->device->active_passive = 0;  /* active - active */

	switch (flag) {
	case PLUG_INTERNAL_CMD:
		/* mpl driver ignores the return value for this case */
		break;

	case PLUG_HBEAT_CMD:
		/* mpl driver assumes a return value of 0 means the hbeat
		 * command succeeded.
		 */
		switch (cmd->cmnd[0]) {
		case INQUIRY:
			/* Setup tgt qdepth after first INQUIRY */
			path->device->tgt_queue_depth = 0; /* no tgt qdepth */
		case TEST_UNIT_READY:
			if(cmd->result)
				rv = 1; /* error */
		}
		break;

	case PLUG_SCSI_CMD:
		/* mpl driver assumes a return value of 1 means to initiate
		 * a failover and retry the command.
		 */
		break;
	}
	return rv;
}

/* In non-default plugin files the plugin table is defined as:
 * DEFINE_PLUGIN(id, name, probe, prequeue, diagnose, down, up, valid)
 * instead of the table below
 */

DEFINE_PLUGIN(jbod, "jbod", mpl_jbod_probe, NULL, mpl_jbod_diagnose, 
		mpl_jbod_path_down, mpl_jbod_path_up, NULL)



