/*******************************************************************
 * This file is part of the Emulex Linux Device Driver for         *
 * Fibre Channel Host Bus Adapters.                                *
 * Copyright (C) 2003-2006 Emulex.  All rights reserved.           *
 * EMULEX and SLI are trademarks of Emulex.                        *
 * www.emulex.com                                                  *
 *                                                                 *
 * This program is free software; you can redistribute it and/or   *
 * modify it under the terms of version 2 of the GNU General       *
 * Public License as published by the Free Software Foundation.    *
 * This program is distributed in the hope that it will be useful. *
 * ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND          *
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE      *
 * DISCLAIMED, EXCEPT TO THE EXTENT THAT SUCH DISCLAIMERS ARE HELD *
 * TO BE LEGALLY INVALID.  See the GNU General Public License for  *
 * more details, a copy of which can be found in the file COPYING  *
 * included with this package.                                     *
 *******************************************************************/

/* 
 * $Id: mpl_util.c 1027 2006-04-07 23:26:53Z jinfante $
 */

#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/blk.h>
#include <linux/utsname.h>
#include <linux/list.h>

/* From drivers/scsi */
#include <sd.h>
#include <hosts.h>
#include <scsi.h>
#else
#include <linux/spinlock.h>
#include <linux/pci.h>
#include <linux/blkdev.h>
#include <scsi/scsi.h>
#include <scsi/scsi_device.h>
#include <scsi/scsi_host.h>
#include <scsi/scsi_cmnd.h>
#include <scsi/scsi_transport_fc.h>
#endif

#include <linux/ctype.h>
#include <linux/string.h>

#include "mpl.h"
#include "mpl_mem.h"
#include "mpl_log.h"
#include "mpl_crtn.h"

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
static struct scsi_device mpl_dummy_scsi_device;
#endif

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
extern int   (* mpl_base_slave_alloc)(struct scsi_device *sdev);
#endif
extern int (* mpl_base_abort_handler)(struct scsi_cmnd *);

int
mpl_vendor_hp_check(struct scsi_cmnd *cmd)
{
	uint8_t hp_str[] = "HP";
	uint8_t compaq_str[] = "COMPAQ";
	uint8_t *vendor = (uint8_t *)cmd->request_buffer+8;

	if ((memcmp(vendor, hp_str, strlen(hp_str)-1) == 0) ||
	    (memcmp(vendor, compaq_str, strlen(compaq_str)-1) == 0)) {
		return 1;
	}
	return 0;
}

void
mpl_restart_sched_cmd(struct mpl_path *path, unsigned long expires)
{
	if (!driver_unloading && path->state != PATH_DISABLED) {
		mpl_module_unlock();
		mod_timer(&path->sched_cmd, expires);
		mpl_module_lock();
	}
}

void
mpl_set_cmd_path(struct scsi_cmnd *cmd, struct mpl_path *path)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	if (!cmd->device) {
		cmd->device = &mpl_dummy_scsi_device;
	}
	cmd->host = path->host->scsi_host;
	cmd->target = path->target;
	cmd->lun = path->lun;
	cmd->channel = path->channel;
#else
	cmd->device = path->sdev;
#endif
	return;
}

#ifdef NOT_USED
void
mpl_set_cmd_path_lun0(struct scsi_cmnd *cmd, struct mpl_path *path)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	if (!cmd->device) {
		cmd->device = &mpl_dummy_scsi_device;
	}
	cmd->host = path->host->scsi_host;
	cmd->target = path->target;
	cmd->lun = 0;
	cmd->channel = path->channel;
#else
	struct mpl_target *tgtp;
	struct mpl_path *pathp;

	cmd->device = path->sdev;
	tgtp = path->tgt;
	if (tgtp) {
		list_for_each_entry(pathp, &tgtp->paths, tgt_path_list) {
			if (pathp->lun == 0) {
				cmd->device = pathp->sdev;
				return;
			}
		}
	}
#endif
	return;
}
#endif  /* NOT_USED */

void
mpl_cmd_timeout(unsigned long data)
{
	struct io_stub *stub = (struct io_stub *)data;
	struct io_stub *s;

	mpl_module_lock();
	s = mpl_find_stub_remove_quiet(stub->cmd);
	if (s != stub) {
		mpl_print(0, MPL_ERR "Timed out command not in hash, "
			  "stub = %p, s = %p cmd = %02x, reason = %x, %s\n",
			  stub, s, stub->cmd->cmnd[0], stub->reason,
			  mpl_print_path(stub->path));
	} else if ( stub->reason == MPL_HEARTBEAT_CMD ) {
		mpl_add_eh_deferral(stub, MPL_DO_HBEAT_TMO);
	} else {
		mpl_add_eh_deferral(stub, MPL_DO_INT_TMO);
	}
	mpl_module_unlock();
	return;
}

void
mpl_set_cmd_timer(struct io_stub *stub, int tmo)
{
	struct scsi_cmnd *cmd = stub->cmd;

	if (tmo) {
		cmd->timeout_per_command = jiffies + (HZ * tmo);
		cmd->eh_timeout.data = (unsigned long) stub;
		mpl_module_unlock();
		mod_timer(&cmd->eh_timeout, jiffies + HZ * tmo);
		mpl_module_lock();
	}
	return;
}

/* 
 *
 */
struct io_stub *
mpl_setup_inquiry(struct mpl_path *path, unsigned char page)
{
	struct scsi_cmnd *cmd;
	struct io_stub *stub;
	uint8_t *databuf;

	cmd = mpl_get_scsi_cmd();

	if (!cmd) 
		goto cmd_alloc_err;

	databuf = mpl_get_scsi_buf(); 
	if (!databuf)
		goto buf_alloc_err;

	wmb();

	/* internal cmd so no done route (param 3) */
	stub = mpl_stub_alloc(cmd, path, 0);
	if (!stub) 
		goto stub_alloc_err;

	mpl_set_cmd_path(cmd, path);
	mpl_set_cmd_timer(stub, MPL_INQUIRY_CMD_TMO); 

	cmd->cmd_len = 6;               /* 6 byte CDB */
	cmd->cmnd[0] = INQUIRY; 
	cmd->cmnd[1] = 1;               /* evpd */
	cmd->cmnd[2] = page;            /* serial # page */
	cmd->cmnd[4] = 0xff;            /* allocation len */

	/* with memset above, use_sg and underflow are all properly setup */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	cmd->sc_data_direction = SCSI_DATA_READ;
#else
	cmd->sc_data_direction = DMA_FROM_DEVICE;
#endif
	cmd->request_bufflen = 0xff;
	cmd->request_buffer = databuf;
	memset(cmd->request_buffer, 0, cmd->request_bufflen);

	return stub;

stub_alloc_err:
	mpl_put_scsi_buf(databuf);
buf_alloc_err:
	mpl_put_scsi_cmd(cmd);
cmd_alloc_err:
	return NULL;
}


struct io_stub *
mpl_setup_report_tgt_port(struct mpl_path *path)
{
	struct scsi_cmnd *cmd;
	struct io_stub *stub;
	uint8_t *databuf;

	cmd = mpl_get_scsi_cmd();

	if (!cmd) 
		goto cmd_alloc_err;

	databuf = mpl_get_scsi_buf(); 
	if (!databuf)
		goto buf_alloc_err;

	wmb();

	/* internal cmd so no done route (param 3) */
	stub = mpl_stub_alloc(cmd, path, 0);
	if (!stub) 
		goto stub_alloc_err;

	mpl_set_cmd_path(cmd, path);
	mpl_set_cmd_timer(stub, MPL_INTERNAL_CMD_TMO); 

	cmd->cmd_len = 12;              /* 12 byte CDB */
	cmd->cmnd[0] = REPORT_TGT_PORT_GROUP;
	cmd->cmnd[1] = 0xa;             /* service action */
	cmd->cmnd[9] = 0xff;            /* allocation len */

	/* with memset above, use_sg and underflow are all properly setup */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	cmd->sc_data_direction = SCSI_DATA_READ;
#else
	cmd->sc_data_direction = DMA_FROM_DEVICE;
#endif
	cmd->request_bufflen = 0xff;
	cmd->request_buffer = databuf;
	memset(cmd->request_buffer, 0, cmd->request_bufflen);

	return stub;

stub_alloc_err:
	mpl_put_scsi_buf(databuf);
buf_alloc_err:
	mpl_put_scsi_cmd(cmd);
cmd_alloc_err:
	return NULL;
}

struct io_stub *
mpl_setup_set_tgt_port(struct mpl_path *path,
	int port_group, int state)
{
	struct mpl_device *device = path->device;
	struct scsi_cmnd *cmd;
	struct io_stub *stub;
	uint8_t *databuf;

	cmd = mpl_get_scsi_cmd();

	if (!cmd) 
		goto cmd_alloc_err;

	databuf = mpl_get_scsi_buf(); 
	if (!databuf)
		goto buf_alloc_err;

	wmb();

	/* internal cmd so no done route (param 3) */
	stub = mpl_stub_alloc(cmd, path, 0);
	if (!stub) 
		goto stub_alloc_err;

	mpl_set_cmd_path(cmd, path);
	mpl_set_cmd_timer(stub, MPL_INTERNAL_CMD_TMO); 

	cmd->cmd_len = 12;              /* 12 byte CDB */
	cmd->cmnd[0] = SET_TGT_PORT_GROUP;
	cmd->cmnd[1] = 0xa;             /* service action */
	cmd->cmnd[9] = 0x8;             /* allocation len */

	/* with memset above, use_sg and underflow are all properly setup */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	cmd->sc_data_direction = SCSI_DATA_WRITE;
#else
	cmd->sc_data_direction = DMA_TO_DEVICE;
#endif
	cmd->request_bufflen = 0x8;
	cmd->request_buffer = databuf;
	memset(cmd->request_buffer, 0, cmd->request_bufflen);
	*(databuf+4) = (uint8_t)state;
	*(databuf+7) = (uint8_t)port_group;

	/* This is implemented in a non-standard way for DS400 */
	if (device->plugin && (memcmp(device->plugin->name, "ds400", 5) == 0)) {
		cmd->cmnd[9] = 0;             /* allocation len */
		cmd->request_bufflen = 0;
	}
	return stub;

stub_alloc_err:
	mpl_put_scsi_buf(databuf);
buf_alloc_err:
	mpl_put_scsi_cmd(cmd);
cmd_alloc_err:
	return NULL;
}

struct io_stub *
mpl_setup_startstop(struct mpl_path *path)
{
	struct scsi_cmnd *cmd;
	struct io_stub *stub;

	cmd = mpl_get_scsi_cmd();

	if (!cmd) 
		goto cmd_alloc_err;

	wmb();

	/* internal cmd so no done route (param 3) */
	stub = mpl_stub_alloc(cmd, path, 0);
	if (!stub) 
		goto stub_alloc_err;

	mpl_set_cmd_path(cmd, path);
	mpl_set_cmd_timer(stub, MPL_INTERNAL_CMD_TMO); 

	cmd->cmd_len = 6;               /* 6 byte CDB */
	cmd->cmnd[0] = START_STOP; 
	cmd->cmnd[4] = 1;               /* start */

	/* with memset above, request_bufflen, use_sg and underflow
	 * are all properly setup */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	cmd->sc_data_direction = SCSI_DATA_NONE;
#else
	cmd->sc_data_direction = DMA_NONE;
#endif

	return stub;

stub_alloc_err:
	mpl_put_scsi_cmd(cmd);
cmd_alloc_err:
	return NULL;
}


struct io_stub *
mpl_setup_tur(struct mpl_path *path)
{
	struct scsi_cmnd *cmd;
	struct io_stub *stub;

	cmd = mpl_get_scsi_cmd();

	if (!cmd) 
		goto cmd_alloc_err;

	wmb();

	/* internal cmd so no done route (param 3) */
	stub = mpl_stub_alloc(cmd, path, 0);
	if (!stub) 
		goto stub_alloc_err;

	mpl_set_cmd_path(cmd, path);
	mpl_set_cmd_timer(stub, MPL_TUR_CMD_TMO); 

	cmd->cmd_len = 6;               /* 6 byte CDB */
	cmd->cmnd[0] = TEST_UNIT_READY; 

	/* with memset above, request_bufflen, use_sg and underflow
	 * are all properly setup */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	cmd->sc_data_direction = SCSI_DATA_NONE;
#else
	cmd->sc_data_direction = DMA_NONE;
#endif
	return stub;

stub_alloc_err:
	mpl_put_scsi_cmd(cmd);
cmd_alloc_err:
	return NULL;
}

void
mpl_send_internal_completion(struct mpl_path *path, struct io_stub *stub,
			     void (*done)(struct scsi_cmnd *))
{
	struct scsi_cmnd *cmd = stub->cmd;
	int  rv;
	unsigned long sflag;
        struct Scsi_Host *shost;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
	spinlock_t    *host_lock;
#endif

	atomic_inc(&int_cmd_cnt);
	atomic_inc(&path->internal_count);

	if (driver_unloading || path->state == PATH_DISABLED) {
		goto int_cmd_not_sent;
	}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	if (!cmd->device) {
		cmd->device = &mpl_dummy_scsi_device;
	}
#else
	if (path->sdev) {
		if ((cmd->cmnd[0] != REPORT_TGT_PORT_GROUP) &&
		    (cmd->cmnd[0] != SET_TGT_PORT_GROUP))
			/* This assumes the SCSI cmd is not specifically directed to LUN 0,
			 * like rtpg and stpg should be (in the long run).
			 */
			cmd->device = path->sdev;
	} else {
		path->sdev = kmalloc(sizeof(struct scsi_device), GFP_ATOMIC);
		if (!path->sdev)
			goto int_cmd_not_sent;
	       	memset(path->sdev, 0, sizeof(struct scsi_device));
		path->sdev->host =  path->host->scsi_host;
		path->sdev->id = path->target;
		path->sdev->lun = path->lun;
		path->sdev->channel = path->channel;
		path->sdev->queue_depth = mpl_get_lun_q_depth(path->host);
		if (mpl_base_slave_alloc(path->sdev)) {
			kfree(path->sdev);
			path->sdev = NULL;
			goto int_cmd_not_sent;
		}
		path->private_sdev = 1;
	}
#endif

	stub->path = path;
	shost =  path->host->scsi_host;
	if ((stub->reason == MPL_HEARTBEAT_CMD) &&
		(shost->host_busy > shost->can_queue) &&
		(atomic_read(&int_cmd_cnt) > 10)) {
		mpl_print(MPL_INTERNAL_CMD, MPL_INFO
			"Defer cmd 0x%x for %s %s: IOs %d > %d\n",
			cmd->cmnd[0],
			mpl_print_device(path->device),
			mpl_print_path(path),
			shost->host_busy, shost->can_queue);

		del_timer(&cmd->eh_timeout);
		atomic_dec(&path->internal_count);
		if (cmd->cmnd[0] == REPORT_TGT_PORT_GROUP)  {
			path->device->tpg_cmd_inp = 0;
		}
		atomic_dec(&int_cmd_cnt);
		mpl_free_internal_stub(stub);
		mpl_hbeat_state_machine(path, HB_BUSY);
		return;
	}

	mpl_print(MPL_INTERNAL_CMD, MPL_INFO "Issue cmd 0x%x for %s %s blk %d\n",
		  cmd->cmnd[0],
		  mpl_print_device(path->device),
		  mpl_print_path(path),
		  (path->host->blocked || path->lld_blocked));

	mpl_add_stub_to_hash(stub);

	if (path->host->blocked || path->lld_blocked) {
		cmd->result = ScsiResult(DID_BUS_BUSY, SAM_STAT_BUSY);
		cmd->retries = 0;
		mpl_module_unlock();
		done(stub->cmd);
		mpl_module_lock();
		return;
	}

	mpl_module_unlock();
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	spin_lock_irqsave(&io_request_lock, sflag);
	rv = mpl_base_queuecommand(cmd, done);
	spin_unlock_irqrestore(&io_request_lock, sflag);
#else
	host_lock = cmd->device->host->host_lock;
	spin_lock_irqsave(host_lock, sflag);
	rv = mpl_base_queuecommand(cmd, done);
	spin_unlock_irqrestore(host_lock, sflag);
	/* If 8.xx driver return SCSI_MLQUEUE_HOST_BUSY retry the command */
	if ( rv == SCSI_MLQUEUE_HOST_BUSY) {
		cmd->result = ScsiResult(0,SAM_STAT_TASK_SET_FULL);
		done(cmd);
		rv = 0;
	}
#endif
	mpl_module_lock();
	if (rv) {
		if (cmd->owner == HBEAT_SIG) {
			cmd->result = ScsiResult(DID_BUS_BUSY, SAM_STAT_BUSY);
			cmd->retries = 0;
			mpl_module_unlock();
			done(stub->cmd);
			mpl_module_lock();
			return;
		}

		WARN_ON(mpl_find_stub_remove(stub->cmd) != stub);

		if ((cmd->cmnd[0] == INQUIRY) && (cmd->cmnd[2] == 0x83)) {
			path->flag |= PATH_NEEDS_INQUIRYP0;
		} else if (cmd->cmnd[0] == TEST_UNIT_READY) {
			path->flag |= PATH_NEEDS_TUR;
		} else if (cmd->cmnd[0] == START_STOP) {
			path->flag |= PATH_NEEDS_STARTSTOP;
		}
		path->retries = cmd->retries + 1;
		mpl_restart_sched_cmd(path, jiffies + HZ);

int_cmd_not_sent:
		if (cmd->request_buffer) {
			mpl_put_scsi_buf(cmd->request_buffer);
		}

		mpl_module_unlock();
		del_timer_sync(&cmd->eh_timeout);
		mpl_module_lock();

		mpl_inc_can_queue(stub);
		mpl_put_scsi_cmd(cmd);
		mpl_stub_free(stub);
		atomic_dec(&int_cmd_cnt);
		atomic_dec(&path->internal_count);
	}
	return;
}

void
mpl_path_schedule_cmd(unsigned long data)
{
	struct mpl_path *path = (struct mpl_path *)data;
	struct mpl_device *device = path->device;
	struct io_stub *stub;
	unsigned long expires = jiffies + 1;

	mpl_module_lock();
	mpl_print(MPL_TMO, MPL_INFO "Schedule (%d) Cmd flg:x%x io:%d dev:%d on %s %s\n",
		path->retries, path->flag, 
		atomic_read(&device->io_count), device->state,
		mpl_print_device(device), mpl_print_path(path));

	if ((path->flag & PATH_DEV_QUIESCE) && atomic_read(&device->io_count)) {
		/* If we need to quiesce, and there are outstanding IOs,
		 * reschedule this command for later.
		 */
		mpl_restart_sched_cmd(path, jiffies + HZ);
		mpl_module_unlock();
		return;
	}
	path->flag &= ~PATH_DEV_QUIESCE;

	if (path->flag & PATH_NEEDS_TUR) {
		stub = mpl_setup_tur(path);
		if (stub) {
			path->flag &= ~PATH_NEEDS_TUR;
			expires = 0;
			stub->cmd->retries = path->retries;
			path->retries = 0;
			mpl_add_deferral(stub, MPL_INTERNAL_CMD);
		}
	} else if (path->flag & PATH_NEEDS_INQUIRYP0) {
		stub = mpl_setup_inquiry(path, 0);
		if (stub) {
			path->flag &= ~PATH_NEEDS_INQUIRYP0;
			expires = 0;
			stub->cmd->retries = path->retries;
			path->retries = 0;
			mpl_add_deferral(stub, MPL_INTERNAL_CMD);
		}
	} else if (path->flag & PATH_NEEDS_STARTSTOP) {
		stub = mpl_setup_startstop(path);
		if (stub) {
			path->flag &= ~PATH_NEEDS_STARTSTOP;
			expires = 0;
			stub->cmd->retries = path->retries;
			path->retries = 0;
			mpl_add_deferral(stub, MPL_INTERNAL_CMD);
		}
	} else if (path->flag & PATH_NEEDS_SET_TGT_PORT) {
		stub = mpl_setup_set_tgt_port(path, path->tgt_port_group, MPL_ALUA_ACTIVE_OPT);
		if (stub) {
			path->flag &= ~PATH_NEEDS_SET_TGT_PORT;
			expires = 0;
			stub->cmd->retries = path->retries;
			path->retries = 0;
			mpl_add_deferral(stub, MPL_INTERNAL_CMD);
		}
	} else {
		/* @TLE. this should not happen */
		expires = 0;
	}

	if (expires) {
		path->retries++;
		mpl_restart_sched_cmd(path, expires);
	}
	mpl_module_unlock();
	return;
}

int
mpl_do_base_queuecommand(struct mpl_device *device, struct mpl_path *path,
			 struct io_stub *stub)
{
	int rv;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
	unsigned long sflag;
	spinlock_t    *prev_host_lock, *new_host_lock ;
#endif

				/* resubmit stub -- need to re-add to hash */
	stub->path = path;
	mpl_add_stub_to_hash(stub);

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
	if (path->sdev) {
		stub->cmd->device = path->sdev;
	} else {
		path->sdev = kmalloc(sizeof(struct scsi_device), GFP_ATOMIC);
		if (!path->sdev) {
			mpl_find_stub_remove(stub->cmd);
			return SCSI_MLQUEUE_HOST_BUSY;
		}
		memset(path->sdev, 0, sizeof(struct scsi_device));
		path->sdev->host =  path->host->scsi_host;
		path->sdev->id = path->target;
		path->sdev->lun = path->lun;
		path->sdev->channel = path->channel;
		path->sdev->queue_depth = mpl_get_lun_q_depth(path->host);
		if (mpl_base_slave_alloc(path->sdev)) {
			kfree(path->sdev);
			path->sdev = NULL;
			stub->cmd->result = ScsiResult(DID_NO_CONNECT, 0);
			mpl_module_unlock();
			mpl_done(stub->cmd);
			mpl_module_lock();
			return 0;
		}
		path->private_sdev = 1;
		stub->cmd->device = path->sdev;
	}
#endif

	if (path->host->blocked || path->lld_blocked) {
		stub->cmd->result = ScsiResult(DID_BUS_BUSY, SAM_STAT_BUSY);
		stub->cmd->retries = 0;
		mpl_module_unlock();
		mpl_done(stub->cmd);
		mpl_module_lock();
		return 0;
	}

	mpl_module_unlock();
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
	prev_host_lock = stub->sdev->host->host_lock;
	new_host_lock  = stub->cmd->device->host->host_lock;
	spin_unlock_irq(prev_host_lock);
	spin_lock_irqsave(new_host_lock, sflag);
#endif
	stub->start_time = jiffies;
	rv = mpl_base_queuecommand(stub->cmd, mpl_done);
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
	spin_unlock_irqrestore(new_host_lock, sflag);
	spin_lock_irq(prev_host_lock);
#endif
			/* if base driver was busy, put it back on 
			 * pendq, re-process this queue 1 sec later */
	mpl_module_lock();

	if (rv) {
		mpl_find_stub_remove(stub->cmd);
	}
	return rv;
}

int
mpl_queuecommand_alternate_path(struct mpl_device *device,
				struct mpl_path *path,
				struct io_stub *stub)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	unsigned long sflag;
#else
	struct scsi_device *sdev;
#endif
	int rc;

	mpl_set_cmd_path(stub->cmd, path);
	atomic_inc(&path->host->io_count);
	atomic_inc(&device->io_count);
	atomic_inc(&path->tgt->io_count);
	atomic_inc(&path->io_count);

	mpl_module_unlock();
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	spin_lock_irqsave(&io_request_lock, sflag);
	mpl_module_lock();
	rc =  mpl_do_base_queuecommand(device, path, stub);
	mpl_module_unlock();
	spin_unlock_irqrestore(&io_request_lock, sflag);
#else
	sdev = stub->sdev;
	spin_lock_irq(sdev->host->host_lock);
	mpl_module_lock();
	rc =  mpl_do_base_queuecommand(device, path, stub);
	mpl_module_unlock();
	spin_unlock_irq(sdev->host->host_lock);
#endif
	if (rc) {
		atomic_dec(&path->host->io_count);
		atomic_dec(&device->io_count);
		atomic_dec(&path->tgt->io_count);
		atomic_dec(&path->io_count);
	}
	mpl_module_lock();
	return rc;
}

void
mpl_path_set_active(struct mpl_path *path, int boolean)
{
	struct mpl_host *host = path->device->host;
	struct mpl_device *device;
	struct mpl_path *p;
	int  target = path->target;
	int  port_group = path->tgt_port_group;

	if (path->device->scsi_type) {
		path->active_controller = 0;
		return;
	}

	if (path->active_controller == boolean) {
		return;
	}

	list_for_each_entry(device, &host->devices, devices) {
		list_for_each_entry(p, &device->paths, path_list) {
			if (p->host->host_no == path->host->host_no &&
			    p->target == target &&
			    p->tgt_port_group == port_group) {
				/* LUN 0 on MSA1000 is never available */
				if (p->lun == 0)
					continue;
				p->active_controller = boolean;
			}
		}
	}
	return;
}

void
mpl_path_flip_all_active(struct mpl_path *path)
{
	struct mpl_host *host = path->device->host;
	struct mpl_device *device;
	struct mpl_device *flip_device;
	struct mpl_path *p;
	struct mpl_path *flip_p;
	int  target = path->target;
	int  port_group = path->tgt_port_group;

	if (path->active_controller) {
		return;
	}

	list_for_each_entry(device, &host->devices, devices) {
		list_for_each_entry(p, &device->paths, path_list) {
			if (p->host->host_no == path->host->host_no &&
			    p->target == target &&
			    p->tgt_port_group == port_group) {
				/* LUN 0 on MSA1000 is never available */
				if (p->lun == 0)
					continue;

				if(p->active_controller < 2) {
					/* Now mark all paths in this device
					 * to be flipped
					 */
					flip_device = p->device;
					list_for_each_entry(flip_p, &flip_device->paths, path_list) {
						if(flip_p->active_controller < 2) {
							flip_p->active_controller += 2; 
						}
					}
				}
			}
		}
	}

	list_for_each_entry(device, &host->devices, devices) {
		list_for_each_entry(p, &device->paths, path_list) {
			/* Flip all marked paths */
			if(p->active_controller == 2) {
				p->active_controller = 1; 
			}
			else if(p->active_controller == 3) {
				p->active_controller = 0; 
			}
		}
	}
	return;
}

void
mpl_device_process_pendq(struct mpl_device *device)
{
	struct io_stub *stub;

	while ((stub = mpl_get_first_pend_io(device)) != NULL) {
		mpl_add_deferral(stub, MPL_RETRY_CMD);
	}
}

#include "lpfc_hw.h"
#include "lpfc_sli.h"
#include "lpfc_mem.h"
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#define  lpfcHBA lpfc_hba
#include "lpfc_sched.h"
#else
#include "lpfc_scsi.h"
#endif
#include "lpfc_disc.h"
#include "lpfc.h"


void
lpfc_driver_lock(struct lpfc_hba *phba, unsigned long *iflag)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	unsigned long flag;

	flag = 0;
	spin_lock_irqsave(&phba->drvrlock, flag);
	*iflag = flag;
	phba->iflag = flag;
#else
	spin_lock_irq(phba->host->host_lock);
#endif
	return;
}

void
lpfc_driver_unlock(struct lpfc_hba *phba, unsigned long *iflag)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	unsigned long flag;

	flag = phba->iflag;
	spin_unlock_irqrestore(&phba->drvrlock, flag);
#else
	spin_unlock_irq(phba->host->host_lock);
#endif
	return;
}

/* Search for a nodelist entry on a specific list */
int
mpl_get_tgt_wwinfo(struct lpfc_hba *phba, uint32_t tgt_id, uint8_t *wwpn_ptr,
                 uint8_t *wwnn_ptr)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	LPFC_NODELIST_t *ndlp;
	LPFCSCSITARGET_t *targetp;
	unsigned long iflag;
#else
	struct lpfc_nodelist *ndlp;
	struct lpfc_target *targetp;
#endif
	struct list_head *pos;

	targetp = phba->device_queue_hash[tgt_id];
	/* First see if the SCSI ID has an allocated LPFCSCSITARGET_t */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	lpfc_driver_lock(phba, &iflag);
#endif

	if (targetp) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
		ndlp =  targetp->pcontext;
#else
		ndlp =  targetp->pnode;
#endif
		if (ndlp) {
			goto found;
		}
	}

	/* Now try the hard way */
	list_for_each(pos, &phba->fc_nlpmap_list) {
		ndlp = list_entry(pos, struct lpfc_nodelist, nlp_listp);
		if (tgt_id == ndlp->nlp_sid) {
			goto found;
		}
	}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	lpfc_driver_unlock(phba, &iflag);
#endif
	/* no match found */
	return 0;
found:
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	lpfc_driver_unlock(phba, &iflag);
#endif
	if(wwpn_ptr)
		memcpy(wwpn_ptr, (uint8_t *) &ndlp->nlp_portname, 8);
	if(wwnn_ptr)
		memcpy(wwnn_ptr, (uint8_t *) &ndlp->nlp_nodename, 8);
	return 1;
}

void
mpl_get_hba_wwpn(struct lpfc_hba *phba, uint8_t *ptr)
{
	memcpy(ptr, (uint8_t *) &phba->fc_portname, 8);
	return;
}

int
mpl_host_is_offline(struct lpfc_hba *phba)
{
	if (phba->fc_flag & FC_OFFLINE_MODE) {
		return 1;
	}
	return 0;
}

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
unsigned short
mpl_get_lun_q_depth( struct mpl_host *host)
{
	unsigned short depth = 0;

	if (host->hba)
		depth = (unsigned short)(host->hba->cfg_lun_queue_depth);
	return depth;
}
#endif

