/*******************************************************************
 * This file is part of the Emulex Linux Device Driver for         *
 * Fibre Channel Host Bus Adapters.                                *
 * Copyright (C) 2003-2006 Emulex.  All rights reserved.           *
 * EMULEX and SLI are trademarks of Emulex.                        *
 * www.emulex.com                                                  *
 *                                                                 *
 * This program is free software; you can redistribute it and/or   *
 * modify it under the terms of version 2 of the GNU General       *
 * Public License as published by the Free Software Foundation.    *
 * This program is distributed in the hope that it will be useful. *
 * ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND          *
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE      *
 * DISCLAIMED, EXCEPT TO THE EXTENT THAT SUCH DISCLAIMERS ARE HELD *
 * TO BE LEGALLY INVALID.  See the GNU General Public License for  *
 * more details, a copy of which can be found in the file COPYING  *
 * included with this package.                                     *
 *******************************************************************/

/* 
 * $Id: mplib.h 1005 2006-01-27 17:27:22Z hdaschbach $
 */

#ifndef _H_MPLIB
#define _H_MPLIB

#include "mpdevid.h"

#ifdef __cplusplus
extern "C" {
#endif

/*
 * The maximum number of paths comprising one route
 */

#define  MPL_MAX_PATHS              4



/***************************************************************************
 *
 * STATUS CODES
 *
 ***************************************************************************/

   /* The request completed successfully */
#define MPL_STATUS_SUCCESS                                             0

   /* The request completed unsuccessfully for an unspecified reason */
#define MPL_STATUS_UNSUCCESSFUL                                        1

   /* The output buffer was not large enough to contain the result */
#define MPL_STATUS_BUFFER_TOO_SMALL                                    2

   /* The specified device was not found */
#define MPL_STATUS_NO_DEVICE                                           3

   /* There was not enough memory available to complete the request */
#define MPL_STATUS_NO_MEMORY                                           4

   /* The name for the object being created already exists */
#define MPL_STATUS_OBJECT_EXISTS                                       5

   /* The specified port was not found */
#define MPL_STATUS_PORT_NOT_FOUND                                      6

   /* The specified path was not found */
#define MPL_STATUS_PATH_NOT_FOUND                                      7

   /* The specified route was not found */
#define MPL_STATUS_ROUTE_NOT_FOUND                                     8

   /* The was an incorrect parameter for the requested operation */
#define MPL_STATUS_INVALID_PARAMETER                                   9

   /* A specified path is already in use */
#define MPL_STATUS_PATH_IN_USE                                         10

   /* The request has not been implemented */
#define MPL_STATUS_UNIMPLEMENTED                                       11

   /* Unable to open the non-volatile storage */
#define MPL_STATUS_NV_OPEN_FAILED                                      12

   /* The write to non-volatile storage failed */
#define MPL_STATUS_NV_WRITE_FAILED                                     13

   /* The specified event mask is invalid */
#define MPL_STATUS_INVALID_EVENT_MASK                                  14

   /* The specified path is not filtered */
#define MPL_STATUS_PATH_IS_UNFILTERED                                  15

   /* The MPULSE subsystem is not enabled */
#define MPL_STATUS_MPULSE_DISABLED                                     16

typedef unsigned int MPL_STATUS, *PMPL_STATUS;





/***************************************************************************
 *
 * Port List structure
 *
 ***************************************************************************/

typedef unsigned int MPL_PORT_STATE, *PMPL_PORT_STATE;

typedef struct MplPortListEntry {
   elxDiDeviceIdentifier DevId;     /* Device identifier to identify the port */
   MPL_PORT_STATE     State;
   unsigned int Pad;
} MPL_PORT_LIST_ENTRY, *PMPL_PORT_LIST_ENTRY;

typedef struct MplGetPortList {
   unsigned int Count;
   unsigned int Pad;
   MPL_PORT_LIST_ENTRY Ports[1];
} MPL_GET_PORT_LIST, *PMPL_GET_PORT_LIST;


#define MPL_PORT_LINK_DOWN          1
#define MPL_PORT_LINK_UP            2
#define MPL_PORT_DISCOVERY_COMPLETE 3
#define MPL_PORT_DIAG_MODE          4
#define MPL_PORT_OTHER              5



/***************************************************************************
 *
 * Route List structure
 *
 ***************************************************************************/

typedef struct MplRouteListEntry {
   elxDiDeviceIdentifier RouteId;   /* Device identifier to identify the route */
   elxDiDeviceIdentifier ValidationId; /* Device identifier to validate route */
} MPL_ROUTE_LIST_ENTRY, *PMPL_ROUTE_LIST_ENTRY;

typedef struct MplGetRouteList {
   unsigned int Count;
   unsigned int Pad;
   MPL_ROUTE_LIST_ENTRY Routes[1];
} MPL_GET_ROUTE_LIST, *PMPL_GET_ROUTE_LIST;



/***************************************************************************
 *
 * Path List structure
 *
 ***************************************************************************/

typedef struct MplPathListEntry {
   elxDiDeviceIdentifier DevId;        /* Device identifier to identify the
                                     * device to the driver    */
   elxDiDeviceIdentifier ValidationId; /* Device identifier to validate the
                                     * identity of the device  */
   unsigned int FailoverBitMask;             /* bitmask to indicate path availability */
   unsigned int Default:             1;      /* Path is the default path */
   unsigned int FcpTarget:           1;      /* Path is to a FCP target */
   unsigned int Bound:               1;      /* Path is bound to a route */
   unsigned int Pad:                29;
} MPL_PATH_LIST_ENTRY, *PMPL_PATH_LIST_ENTRY;

/* path FailoverBitMask events */
#define MPL_PATH_LINK_DOWN         0x1   /* Link is down */
#define MPL_PATH_DISCONNECTED      0x2   /* noactive connection to remote dev */
#define MPL_PATH_INVALID           0x4   /* Path invalid by Inquiry SN */
#define MPL_PATH_MAINT_MODE        0x8   /* HBA is in maintance mode */
#define MPL_PATH_INACTIVE          0x10  /* Path inactive by drvr internally */
#define MPL_PATH_USER_INITIATED    0x20  /* Path taken offline by admin */

#define mpl_path_phys_connected(fmask) ((fmask & \
   (MPL_PATH_LINK_DOWN | MPL_PATH_DISCONNECTED | \
    MPL_PATH_INVALID | MPL_PATH_MAINT_MODE)) == 0)


typedef struct MplGetPathList {
   elxDiDeviceIdentifier RouteId;      /* Device identifier to indicate the path
                                     * for which routes should be returned */
   unsigned int Count;
   unsigned int Pad;
   MPL_PATH_LIST_ENTRY Paths[1];
} MPL_GET_PATH_LIST, *PMPL_GET_PATH_LIST;




/***************************************************************************
 *
 * Path control structures
 *
 ***************************************************************************/

/* Path Control Types */

#define MPL_PC_TYPE_FAILOVER       0x1  /* alternate paths are standby paths */
#define MPL_PC_TYPE_TRAFFIC_RATIO  0x2  /* alt paths balanced by traffic ratio */
#define MPL_PC_TYPE_ROUND_ROBIN    0x2  /* Round Robin algorithm used */
#define MPL_PC_TYPE_DYNAMIC_LUNQ   0x3  /* Dynamic balancing algorithm used */
#define MPL_PC_TYPE_DYNAMIC_TGTQ   0x4  /* Dynamic balancing algorithm used */
#define MPL_PC_TYPE_DYNAMIC_HBAQ   0x5  /* Dynamic balancing algorithm used */
#define MPL_PC_TYPE_DYNAMIC_TIME   0x6  /* least time algorithm used */
#define MPL_PC_TYPE_COUNT          0x6  /* cnt of different path control types */
#define MPL_PC_TYPE_MASK           0xf  /* mask to obtain path control types */

typedef unsigned int MPL_PATH_CONTROL_TYPE, *PMPL_PATH_CONTROL_TYPE;

/* Path Control Configuration identifiers */

#define MPL_PC_CFG_AUTO_FAILBACK       0x10
#define MPL_PC_CFG_INQUIRY_HEARTBEAT   0x40
#define MPL_PC_CFG_FAIL_ABORT_TASK_SET 0x80
#define MPL_PC_CFG_FAIL_LUN_RESET      0x100
#define MPL_PC_CFG_FAIL_TARGET_RESET   0x200
#define MPL_PC_CFG_FAIL_MASK           0x380
#define MPL_PC_CFG_OPTIMIZED           0x400 /* only balance across active-optimized paths */
#define MPL_PC_CFG_MASK                0x7f7


typedef struct MplPathControlHeader {
   MPL_PATH_CONTROL_TYPE Type;
   unsigned int AutoFailBack:            1;  /* failback to primary path is automatic */
   unsigned int InquiryHeartbeat:        1;  /* Validate idle path with inq heartbeat */
   unsigned int FailoverAbortTaskSet:    1;  /* On path failover, issue Abort Task Set */
   unsigned int FailoverLunReset:        1;  /* On path failover, issue LUN Reset */
   unsigned int FailoverTgtReset:        1;  /* On path failover, issue Target Reset */
   unsigned int Pad:                    25;
   unsigned int OptimizedOnly:           1;  /* load balance on optimized paths only */
   unsigned int ActivePassive:           1;  /* Array is active/passive type */
   unsigned int Length;
   unsigned int PathCount;                   /* number of paths this route supports */
} MPL_PATH_CONTROL_HEADER, *PMPL_PATH_CONTROL_HEADER;

typedef struct MplFailoverControl {
   MPL_PATH_CONTROL_HEADER Header;
   elxDiDeviceIdentifier PrimaryPathDevId;
   elxDiDeviceIdentifier PrimaryPathValidationId;
} MPL_FAILOVER_CONTROL, *PMPL_FAILOVER_CONTROL;


typedef struct MplTrafficRatioEntry {
   unsigned int Ratio;
   unsigned int Pad;
   elxDiDeviceIdentifier DeviceId;
   elxDiDeviceIdentifier ValidationId;
} MPL_TRAFFIC_RATIO_ENTRY, *PMPL_TRAFFIC_RATIO_ENTRY;

typedef struct MplTrafficRatioControl {
   MPL_PATH_CONTROL_HEADER Header;
   unsigned int Count;
   unsigned int Pad;
   MPL_TRAFFIC_RATIO_ENTRY PathRatios[MPL_MAX_PATHS];
} MPL_TRAFFIC_RATIO_CONTROL, *PMPL_TRAFFIC_RATIO_CONTROL;

#define  MPL_MAX_PATH_CONTROL_SIZE \
   ((sizeof(MPL_FAILOVER_CONTROL) < sizeof(MPL_TRAFFIC_RATIO_CONTROL)) \
   ? sizeof(MPL_TRAFFIC_RATIO_CONTROL) : sizeof(MPL_FAILOVER_CONTROL))

/***************************************************************************
 *
 * Path status structure
 *
 * Only MPL_PS_STATE_ONLINE and MPL_PS_STATE_OFFLINE can be used for
 * setting the Path state.
 *
 ***************************************************************************/

#define MPL_PS_STATE_ONLINE              1   /* path is available */
#define MPL_PS_STATE_AVAILABLE           1   /* path is available */
#define MPL_PS_STATE_DEGRADED            2   /* path tested ready, but has 
                                                failed */
#define MPL_PS_STATE_FAILED              3   /* path has failed */
#define MPL_PS_STATE_FAILING             4   /* path has failed, with I/O
                                                still outstanding */
#define MPL_PS_STATE_OFFLINE             5   /* user has taken path offline */
#define MPL_PS_STATE_REMOVED             6   /* PNP has removed the path */
#define MPL_PS_STATE_STOPPING            7   /* path is transitioning to 
                                                offline */
#define MPL_PS_STATE_STARTING            8   /* path is transitioning from 
                                                offline to online */
#define MPL_PS_STATE_DISABLED            9   /* path is not enabled for 
                                                use by MultiPulse */
#define MPL_PS_STATE_STANDBY             10  /* path is in a standby state, 
                                                not ready for use */
#define MPL_PS_STATE_NEED_REBOOT         11  /* path needs reboot before
                                                a valid state can be 
                                                determined */

typedef unsigned int MPL_PATH_STATE, *PMPL_PATH_STATE;

typedef struct MplPathStatus {
   unsigned int Length;
   unsigned int Mask;
#define  MPL_PATH_MASK_STATE           0x00000001
#define  MPL_PATH_MASK_ERROR_COUNT     0x00000002
#define  MPL_PATH_MASK_QUEUE_FULL_CNT  0x00000004
#define  MPL_PATH_MASK_TOTAL_CMD_CNT   0x00000008
#define  MPL_PATH_MASK_TOTAL_COMP_CNT  0x00000010
#define  MPL_PATH_MASK_ACTIVE_COUNT    0x00000020
#define  MPL_PATH_MASK_QUEUE_DEPTH     0x00000040
#define  MPL_PATH_MASK_RATIO_COUNT     0x00000080
#define  MPL_PATH_MASK_FLAGS           0x00000100
#define  MPL_PATH_MASK_FAILOVER_MASK   0x00000200
#define  MPL_PATH_MASK_ALL             0x000003ff
   MPL_PATH_STATE State;

   /* The following counters can be zero/changed with mplSetPathStatus */
   unsigned int ErrorCount;                  /* Current number of errors */
   unsigned int QueueFullCount;              /* Current number of Qfulls */
   unsigned int TotalCmdCount;               /* Current number of FCP cmds initiated */
   unsigned int TotalCmplCount;              /* Current number of FCP cmds completed */

   /* The following counters are ignored by mplSetPathStatus */
   unsigned int ActiveCount;                 /* Current count of IO on path */
   unsigned int QueueDepth;                  /* Current queue depth set for device */
   unsigned int RatioCount;                  /* Current counter for traffic ratio */
   unsigned int Flags;
   unsigned int FailoverBitMask;
} MPL_PATH_STATUS_V1, *PMPL_PATH_STATUS_V1;

#define  MPL_PATH_STATUS   MPL_PATH_STATUS_V1
#define  PMPL_PATH_STATUS  PMPL_PATH_STATUS_V1


#define  MPL_PATH_MASK_SETABLE_ALL                              \
   (MPL_PATH_MASK_ERROR_COUNT | MPL_PATH_MASK_QUEUE_FULL_CNT |  \
   MPL_PATH_MASK_TOTAL_CMD_CNT | MPL_PATH_MASK_TOTAL_COMP_CNT)


typedef struct MplPathInfo {
   unsigned int Length;
   unsigned int Port;
   unsigned int Bus;
   unsigned int Target;
   unsigned int Lun;
   unsigned int Pad;
   char Vendor[16];
   char Product[20];
   char Rev[8];
   elxDiDeviceIdentifier RouteId;
} MPL_PATH_INFO, *PMPL_PATH_INFO;


/***************************************************************************
 *
 * Event handling definitions and structures
 *
 ***************************************************************************/

typedef void *MPL_EVENT_HANDLE, *PMPL_EVENT_HANDLE;

#define MPL_EVT_LINK_EVENT                0x01
#define MPL_EVT_MULTIPULSE_EVENT          0x08   
#define MPL_EVT_PATH_EVENT                0x10

typedef unsigned int MPL_EVENT_TYPE, *PMPL_EVENT_TYPE;

#define MPL_EVT_MASK_LINK_EVENT       MPL_EVT_LINK_EVENT
#define MPL_EVT_MASK_PATH_EVENT       MPL_EVT_PATH_EVENT
#define MPL_EVT_MASK_MULTIPULSE_EVENT      MPL_EVT_MULTIPULSE_EVENT

typedef unsigned int MPL_EVENT_MASK, *PMPL_EVENT_MASK;


typedef struct MplLinkEvent {
   elxDiDeviceIdentifier DevId;
   MPL_PORT_STATE State;
   unsigned int Pad;
} MPL_LINK_EVENT, *PMPL_LINK_EVENT;

typedef struct MplPathEvent {
   elxDiDeviceIdentifier DevId;
   elxDiDeviceIdentifier ValidationId;
   MPL_PATH_STATUS Status;
   unsigned int Pad;
} MPL_PATH_EVENT, *PMPL_PATH_EVENT;

typedef struct MplRouteEvent {
   elxDiDeviceIdentifier MpDevId;
} MPL_MULTIPULSE_EVENT, *PMPL_MULTIPULSE_EVENT;




/***************************************************************************
 *
 * Route Control Types
 *
 ***************************************************************************/

typedef struct MplRouteControlType {
   char Name[36];
   unsigned int IndexValue;
} MPL_ROUTE_CONTROL_TYPE, *PMPL_ROUTE_CONTROL_TYPE;

typedef struct MplRouteControlTypes {
   unsigned int Count;
   unsigned int Pad;
   MPL_ROUTE_CONTROL_TYPE ControlTypes[MPL_PC_TYPE_COUNT];
} MPL_ROUTE_CONTROL_TYPES, *PMPL_ROUTE_CONTROL_TYPES;

/***************************************************************************
 *
 * Global Preferences
 *
 ***************************************************************************/

typedef struct MplGlobalOption {
   char Name[64];
   char HelpString[128];
   unsigned int MinValue;
   unsigned int MaxValue;
   unsigned int CurrentValue;
   int RebootRequired;           /* non zero if a reboot is required to */
                                 /* activate                            */
   int AdvancedCommand;          /* non zero if this is considered an   */
                                 /* advanced parameter                  */
} MPL_GLOBAL_OPTION, *PMPL_GLOBAL_OPTION;

typedef struct MplGlobalPreferences {
   int Count;
   int Pad;
   MPL_GLOBAL_OPTION Options[1];
} MPL_GLOBAL_PREFERENCES, *PMPL_GLOBAL_PREFERENCES;

/*
 * An identifier initialized to the value of a NULL identifier
 */

extern elxDiDeviceIdentifier mplNullIdentifier; /* = {0} */

                                             
int mplDebugOutput(char *fmt, ...);


/* Configuration parameter handling used for
 * mplGetCfgParam and mplSetCfgParam
 */
typedef struct MplCfgParam {
	char a_string[32];
	uint32_t a_low;
	uint32_t a_hi;
	uint32_t a_default;
	uint32_t a_current;
	uint16_t a_flag;
	uint16_t a_changestate;
	char a_help[80];
} MPL_CFGPARAM;

#define MPL_CFG_LOG_VERBOSE         0  /* mpl_log_verbose */
#define MPL_CFG_HBEAT_INTERVAL      1  /* mpl_hbeat_interval */
#define MPL_CFG_HBEAT_FAIL_INTERVAL 2  /* mpl_fail_hbeat_interval */
#define MPL_CFG_AUTO_CONFIGURE      3  /* mpl_auto_configure */
#define MPL_CFG_AUTO_CONTROLLER     4  /* mpl_auto_controller */
#define MPL_CFG_AUTO_LUN_BALANCE    5  /* mpl_auto_lun_balance */
#define MPL_CFG_HBEAT_RETRY         6  /* mpl_max_hbeat_retry */
#define MPL_CFG_ENABLE_TGT_QDEPTH   7  /* mpl_enable_tgt_qdepth */
#define MPL_CFG_HANDLE_QFULL        8  /* mpl_handle_qfull */
#define MPL_MAX_CFG_PARAM           9

/* values for a_flag */
#define MPL_CFG_EXPORT         1 /* export parameter to end-user */
#define MPL_CFG_IGNORE         2 /* ignore parameter */

/* values for a_changestate */
#define MPL_CFG_REBOOT         0 /* effective after reboot */
#define MPL_CFG_DYNAMIC        1 /* effective immediately */

#ifdef __cplusplus
}
#endif

#endif  /* _H_MPLIB */
