#!/usr/bin/perl -w
#*******************************************************************
#* This file is part of the Emulex Linux Device Driver for         *
#* Fibre Channel Host Bus Adapters.                                *
#* Copyright (C) 2003-2006 Emulex.  All rights reserved.           *
#* EMULEX and SLI are trademarks of Emulex.                        *
#* www.emulex.com                                                  *
#*                                                                 *
#* This program is free software; you can redistribute it and/or   *
#* modify it under the terms of version 2 of the GNU General       *
#* Public License as published by the Free Software Foundation.    *
#* This program is distributed in the hope that it will be useful. *
#* ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND          *
#* WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY,  *
#* FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE      *
#* DISCLAIMED, EXCEPT TO THE EXTENT THAT SUCH DISCLAIMERS ARE HELD *
#* TO BE LEGALLY INVALID.  See the GNU General Public License for  *
#* more details, a copy of which can be found in the file COPYING  *
#* included with this package.                                     *
#*******************************************************************

# Add conditional compilation directives to /etc/lpfc.conf file (if this
# has not already been done) to compile the mpl, but not lpfc driver,
# configuration from the confguration file.

$lpfc_conditional = "#ifndef IGNORE_LPFC_CONFIGURATION\n";
$mpulse_found = 0;
$exclusion_found = 0;
$config_found = 0;

$file = $ARGV[0];

if (-l $file) {
  $file = readlink($file);
}
die "Cannot access $file" if (!defined $file || ! -f $file);
$new_file = $file . ".mplized";
open(STDIN, "< $file") || die "Cannot open $file: $!";
open(STDOUT, "> $new_file") || die "Cannot open $new_file: $!";

while (<>) {
  m/GLOBAL PARMS/ && print $lpfc_conditional;
  m/MPULSE/ && ($mpulse_found = 1);
  m/lpfc_multipulse/ && ($config_found = 1);
  m/mpl_exclusion/ && ($exclusion_found = 1);
  if (m/$lpfc_conditional/) {
    # if the file already has been mplized, just copy the rest of the file.
    # indicate whether we need to add an exclusion list.
    print;
    while (<>) {
      m/MPULSE/ && ($mpulse_found = 1);
      m/mpl_exclusion/ && ($exclusion_found = 1);
      print;
    }
    last;
  }
  if (m/#endif/) {
    print "#endif\n";
    if (! $mpulse_found) {
      print "\n#ifdef MPULSE\n";
      # NOTE: we don't want to include mpl.h, so hard code 1024
      print "char *lpfc_multipulse[1024] = { 0 };\n" if (!$config_found);
      print "char *mpl_exclusion[1024] = { 0 };\n" if (!$exclusion_found);
      print "#endif /* MPULSE */\n\n";
      $exclusion_found = 1;
    }
  }
  print;
}

close STDIN;
close STDOUT;
if ($mpulse_found && !$exclusion_found) {
  open(STDIN, "< $new_file") || die "Cannot open $new_file: $!";
  open(STDOUT, "> $file") || die "Cannot open $file: $!";
  while (<>) {
    print;
    print "char *mpl_exclusion[1024] = { 0 };\n" if (m/char .lpfc_multipulse/);
  }
  close STDIN;
  close STDOUT;
  unlink $new_file;
} elsif (! $mpulse_found) {
  rename $new_file, $file;
}
