/*******************************************************************
 * This file is part of the Emulex Linux Device Driver for         *
 * Fibre Channel Host Bus Adapters.                                *
 * Copyright (C) 2003-2006 Emulex.  All rights reserved.           *
 * EMULEX and SLI are trademarks of Emulex.                        *
 * www.emulex.com                                                  *
 *                                                                 *
 * This program is free software; you can redistribute it and/or   *
 * modify it under the terms of version 2 of the GNU General       *
 * Public License as published by the Free Software Foundation.    *
 * This program is distributed in the hope that it will be useful. *
 * ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND          *
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE      *
 * DISCLAIMED, EXCEPT TO THE EXTENT THAT SUCH DISCLAIMERS ARE HELD *
 * TO BE LEGALLY INVALID.  See the GNU General Public License for  *
 * more details, a copy of which can be found in the file COPYING  *
 * included with this package.                                     *
 *******************************************************************/

/* 
 * $Id: mpl_shim.c 1128 2006-09-13 20:19:15Z jinfante $
 */

#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#ifndef EXPORT_SYMTAB
#define EXPORT_SYMTAB
#endif
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kmod.h>
#include <linux/kernel.h>
#include <linux/blk.h>
#include <linux/utsname.h>
#include <linux/list.h>
#include <linux/delay.h>
#include <linux/smp_lock.h>

/* From drivers/scsi */
#include <sd.h>
#include <hosts.h>
#include <scsi.h>
#else
#include <linux/spinlock.h>
#include <linux/pci.h>
#include <linux/blkdev.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <scsi/scsi.h>
#include <scsi/scsi_device.h>
#include <scsi/scsi_host.h>
#include <scsi/scsi_cmnd.h>
#include <scsi/scsi_transport_fc.h>
#include <linux/smp_lock.h>
extern int scsi_internal_device_block(struct scsi_device *sdev);
extern int scsi_internal_device_unblock(struct scsi_device *sdev);
extern int scsi_queue_insert(struct scsi_cmnd *cmd, int reason);
int mpl_eh_strategy_handler(struct Scsi_Host *shost);
extern int scsi_delete_timer(struct scsi_cmnd *);
#endif

#include "mpl.h"
#include "mpl_mem.h"
#include "mpl_log.h"
#include "mpl_crtn.h"
#include "mpl_version.h"

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
void mpl_eh_timeout_handler(struct scsi_cmnd *);
#endif
static int host_cnt = 0;

#define DEFAULT_DEVICE_HASH_SIZE  1024
#define DEFAULT_PATH_HASH_SIZE  1024
#define DEFAULT_STUB_HASH_SIZE  1024

#define NUM_ENTRIES(x) (sizeof(x) / sizeof(x[0]))

extern struct mpl_plugin plugin_data_start;
extern struct mpl_plugin plugin_data_end;
extern int mpl_setup_multipulse(void);

LIST_HEAD(mpl_targets);
LIST_HEAD(mpl_hosts);
static LIST_HEAD(reaper_queue);

static LIST_HEAD(deferral_queue);
static spinlock_t deferral_queue_lock = SPIN_LOCK_UNLOCKED;
static int  mpl_thread_quit;
static DECLARE_MUTEX_LOCKED(mpl_thread_sleep);
static DECLARE_MUTEX_LOCKED(mpl_thread_gone);

static LIST_HEAD(eh_deferral_queue);
static spinlock_t eh_deferral_queue_lock = SPIN_LOCK_UNLOCKED;
static int  mpl_eh_thread_quit;
static DECLARE_MUTEX_LOCKED(mpl_eh_thread_sleep);
static DECLARE_MUTEX_LOCKED(mpl_eh_thread_gone);

kmem_cache_t *io_stub_cache;
struct mpl_device **device_hash;
int  device_hash_size = DEFAULT_DEVICE_HASH_SIZE;
spinlock_t device_hash_table_lock = SPIN_LOCK_UNLOCKED;
struct mpl_path **path_hash;
int  path_hash_size = DEFAULT_PATH_HASH_SIZE;
spinlock_t path_hash_table_lock = SPIN_LOCK_UNLOCKED;
struct io_stub **stub_hash;
int  stub_hash_size = DEFAULT_STUB_HASH_SIZE;
spinlock_t stub_hash_table_lock = SPIN_LOCK_UNLOCKED;

spinlock_t mpl_module_spin_lock = SPIN_LOCK_UNLOCKED;
unsigned long mpl_module_flag = 0;

/* alternate paths are standby paths */
int  mpl_auto_configure = MPL_PC_TYPE_FAILOVER;

int  mpl_auto_controller = 1;  /* Enabled by default */
int  mpl_auto_lun_balance = 1; /* Enabled by default */

int (* mpl_base_queuecommand)(struct scsi_cmnd *, void (*done)(struct scsi_cmnd *));
int (* mpl_base_release)(struct Scsi_Host *);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
Scsi_Host_Template *mpl_template;
int (* mpl_base_detect)(Scsi_Host_Template *);
/* In the 2.6 driver, lun queue depth is controled entirely by scsi mid-layer */
void  (* mpl_base_select_queue_depths)(struct Scsi_Host *, struct scsi_device *);
#else
struct scsi_host_template *mpl_template;
int (* mpl_base_detect)(struct scsi_host_template *);
int   (* mpl_base_slave_alloc)(struct scsi_device *sdev);
void  (* mpl_base_slave_destroy)(struct scsi_device *sdev);
#endif

int (* mpl_base_abort_handler)(struct scsi_cmnd *);
int (* mpl_base_device_reset_handler)(struct scsi_cmnd *);
int (* mpl_base_bus_reset_handler)(struct scsi_cmnd *);
#define MPL_ABORT_EXACT             0x0
#define MPL_ABORT_DEVICE            0x1
#define MPL_ABORT_HOST              0x2
#define MPL_ABORT_TYPE_MASK         0xf
#define MPL_ABORT_IGNORE_ABORTED    0x10

void mpl_done(struct scsi_cmnd *);
int mpl_abort_handler(struct scsi_cmnd *);
int mpl_do_abort(struct scsi_cmnd *);
int mpl_do_tmo(struct io_stub *);
int mpl_do_dev_reset(struct io_stub *);
int mpl_log_verbose = 0;
int mpl_enable_tgt_qdepth = 1;
volatile int driver_unloading = 0;
atomic_t int_cmd_cnt = ATOMIC_INIT(0);
int mpl_hbeat_interval = MPL_DFT_HBEAT_INTERVAL;
int mpl_hbeat_fail_interval = MPL_DFT_HBEATFAIL_INTERVAL;
int mpl_max_hbeat_retry = MPL_DFT_HBEAT_RETRY;
int mpl_nopath_tmo = 0;
int mpl_enable_nopath = 1;
int mpl_pfq_use_lun_notsup = 1;

uint32_t mpl_max_can_queue = 0xffffff;

static int initial_scsi_scan_done = 0;
static spinlock_t pend_queue_lock = SPIN_LOCK_UNLOCKED;
static struct mpl_device *mpl_match_device_by_serial(struct mpl_device *,
						 unsigned char *, int);
static void mpl_hbeat_timeout(unsigned long);
static void mpl_nopath_timeout(unsigned long);
static void mpl_hb_done(struct scsi_cmnd *cmd);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
MODULE_PARM(mpl_log_verbose, "i");
MODULE_PARM(mpl_hbeat_interval, "i");
MODULE_PARM(mpl_hbeat_fail_interval, "i");
MODULE_PARM(mpl_auto_configure, "i");
MODULE_PARM(mpl_auto_controller, "i");
MODULE_PARM(mpl_auto_lun_balance, "i");
MODULE_PARM(mpl_max_hbeat_retry, "i");
MODULE_PARM(mpl_enable_tgt_qdepth, "i");
MODULE_PARM(mpl_nopath_tmo, "i");
MODULE_PARM(mpl_enable_nopath, "i");
MODULE_PARM(mpl_pfq_use_lun_notsup, "i");
#else
int mpl_handle_qfull = 0;
module_param(mpl_log_verbose, int, 0);
module_param(mpl_hbeat_interval, int, 0);
module_param(mpl_hbeat_fail_interval, int, 0);
module_param(mpl_auto_configure, int, 0);
module_param(mpl_auto_controller, int, 0);
module_param(mpl_auto_lun_balance, int, 0);
module_param(mpl_max_hbeat_retry, int, 0);
module_param(mpl_enable_tgt_qdepth, int, 0);
module_param(mpl_handle_qfull, int, 0);
module_param(mpl_nopath_tmo, int, 0);
module_param(mpl_enable_nopath, int, 0);
module_param(mpl_pfq_use_lun_notsup, int, 0);
#endif
MODULE_PARM_DESC(mpl_log_verbose, "Mpl log verbose setting (default = 0)");
MODULE_PARM_DESC(mpl_hbeat_interval, "Seconds between Inquiry heartbeats on a device (default = 30)");
MODULE_PARM_DESC(mpl_hbeat_fail_interval, "Seconds between Inquiry heartbeats on a failed path (default = 1)");
MODULE_PARM_DESC(mpl_auto_configure, "Automatically configured discovered paths into MultiPulse devices");
MODULE_PARM_DESC(mpl_auto_controller, "Make active controller or preferred paths primary for autoconfiguration");
MODULE_PARM_DESC(mpl_auto_lun_balance, "Balance primary paths across a target for autoconfiguration");
MODULE_PARM_DESC(mpl_max_hbeat_retry, "Max retries for a failed hbeat cmd");
MODULE_PARM_DESC(mpl_enable_tgt_qdepth, "Enable target queue depth per plugin");
MODULE_PARM_DESC(mpl_nopath_tmo, "Override plugin setting for time to wait when no paths are available");
MODULE_PARM_DESC(mpl_enable_nopath, "Enable setting for time to wait when no paths are available");
MODULE_PARM_DESC(mpl_pfq_use_lun_notsup, "Use LUN not supported PFQ when hiding paths from SCSI Mid-layer");
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
MODULE_PARM_DESC(mpl_handle_qfull, "Tell SCSI Layer to reduce qdepth on QFULL");
#endif

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
static int
mpl_slave_alloc(struct scsi_device *scsi_devs)
{
	int ret;
	struct mpl_host   *host;
	struct mpl_path   *path;

	mpl_module_lock();
	host  = mpl_find_host_by_scsi_host(scsi_devs->host);
	if (!host) {
		mpl_module_unlock();
		return -ENXIO;
	}
	path = mpl_find_or_create_path(host,
					scsi_devs->id, 
					scsi_devs->lun);
	mpl_module_unlock();

	ret = mpl_base_slave_alloc(scsi_devs);
	return ret;
}

static void
mpl_slave_destroy(struct scsi_device *sdev)
{
	mpl_base_slave_destroy(sdev);
	return;
}

#endif

/* The following 2 functions support a global lock on the mpl module */
void
mpl_module_lock(void)
{
	unsigned long flag;

	flag = 0;
	spin_lock_irqsave(&mpl_module_spin_lock, flag);
	mpl_module_flag = flag;
	return;
}

void
mpl_module_unlock(void)
{

	unsigned long flag;

	flag = mpl_module_flag;
	spin_unlock_irqrestore(&mpl_module_spin_lock, flag);
	return;
}

/* The following functions support a generic module logging mechanism */
char *
mpl_print_host(struct mpl_host   *host)
{
	static char hostbuf[32];

	snprintf(hostbuf, sizeof(hostbuf), "mplHost:%d sHost:%d",
		 host->host_no,
		 host->scsi_host->host_no);
	return(hostbuf);
}

char *
mpl_print_device(struct mpl_device *device)
{
	static char devbuf[32];

	snprintf(devbuf, sizeof(devbuf), "Dev:%d:%d:%d",
		 device->id.host_no,
		 device->id.tgt_no, device->id.lun_no);
	return(devbuf);
}

char *
mpl_print_path(struct mpl_path   *path)
{
	static char pathbuf[32];

	snprintf(pathbuf, sizeof(pathbuf), "Path:%d:%d:%d",
		 path->host->host_no,
		 path->target, path->lun);
	return(pathbuf);
}

char *
mpl_print_wwpn(unsigned char *wwpn)
{
	static char wwpnbuf[32];

	snprintf(wwpnbuf, sizeof(wwpnbuf), "%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x",
		 wwpn[0], wwpn[1], wwpn[2], wwpn[3], 
		 wwpn[4], wwpn[5], wwpn[6], wwpn[7]); 
	return(wwpnbuf);
}

/* The mpl module uses 2 different kernel threads for deferred processing
 * of of events, mpl_thread and mpl_eh_thread. 
 * The following routines support waking up mpl_thread and adding/removing
 * events to mpl_thread's deferral_queue.
 */
static inline void
mpl_wake_thread(void)
{
	up(&mpl_thread_sleep);
}

void
mpl_add_deferral(struct io_stub *stub, enum deferral_reason reason)
{
	unsigned long flags;
	struct mpl_device *device = stub->path->device;

	if ((reason == MPL_INTERNAL_CMD) || (reason == MPL_HEARTBEAT_CMD)) {
		mpl_dec_can_queue(stub);
	}
	stub->reason = reason;
	spin_lock_irqsave(&deferral_queue_lock, flags);
	atomic_inc(&device->deferral_count);
	list_add_tail(&stub->queue, &deferral_queue);
	spin_unlock_irqrestore(&deferral_queue_lock, flags);
				/* Waking the thread up for blocked paths
				 * causes enough scheduler churn to hang the
				 * system. */
	if (!stub->path->host->blocked && !stub->path->lld_blocked)
		mpl_wake_thread();
}

/* dequeue first deferral io_stub, return NULL if list empty */
static inline struct io_stub *
mpl_get_first_deferral(void)
{
	struct list_head *first;
	struct io_stub *stub = NULL;
	unsigned long flags;

	spin_lock_irqsave(&deferral_queue_lock, flags);
	first = list_first(&deferral_queue);
	if (first) {
		stub = list_entry(first, struct io_stub, queue);
		list_del(&stub->queue);
	}
	spin_unlock_irqrestore(&deferral_queue_lock, flags);

	/* The device deferral_count NEEDS to be decremented outside
	 * this call if successful.
	 */
	return stub;
}

static struct io_stub *
mpl_find_stub_remove_deferral(struct scsi_cmnd *cmd)
{
	struct io_stub *stub;
	struct io_stub *rv = NULL;
	unsigned long flags;

	spin_lock_irqsave(&deferral_queue_lock, flags);
	list_for_each_entry(stub, &deferral_queue, queue) {
		if (stub->cmd == cmd) {
			rv = stub;
			list_del(&stub->queue);
			mpl_inc_can_queue(stub);
			break;
		}
	}
	spin_unlock_irqrestore(&deferral_queue_lock, flags);
	return rv;
}

/* The mpl module uses 2 different kernel threads for deferred processing
 * of of events, mpl_thread and mpl_eh_thread. 
 * The following routines support waking up mpl_eh_thread and adding/removing
 * events to mpl_eh_thread's eh_deferral_queue.
 */
static inline void
mpl_wake_eh_thread(void)
{
	up(&mpl_eh_thread_sleep);
}

void
mpl_add_eh_deferral(struct io_stub *stub, enum deferral_reason reason)
{
	unsigned long flags;
	struct mpl_device *device = stub->path->device;

	stub->reason = reason;
	spin_lock_irqsave(&eh_deferral_queue_lock, flags);
	atomic_inc(&device->eh_deferral_count);
	list_add_tail(&stub->queue, &eh_deferral_queue);
	spin_unlock_irqrestore(&eh_deferral_queue_lock, flags);
	mpl_wake_eh_thread();
}

/* dequeue first eh_deferral io_stub, return NULL if list empty */
static inline struct io_stub *
mpl_get_first_eh_deferral(void)
{
	struct list_head *first;
	struct io_stub *stub = NULL;
	unsigned long flags;

	spin_lock_irqsave(&eh_deferral_queue_lock, flags);
	first = list_first(&eh_deferral_queue);
	if (first) {
		stub = list_entry(first, struct io_stub, queue);
		list_del(&stub->queue);
	}
	spin_unlock_irqrestore(&eh_deferral_queue_lock, flags);

	/* The device eh_deferral_count NEEDS to be decremented outside
	 * this call if successful.
	 */
	return stub;
}

static struct io_stub *
mpl_find_stub_remove_eh_deferral(struct scsi_cmnd *cmd)
{
	struct io_stub *stub;
	struct io_stub *rv = NULL;
	unsigned long flags;

	spin_lock_irqsave(&eh_deferral_queue_lock, flags);
	list_for_each_entry(stub, &eh_deferral_queue, queue) {
		if (stub->cmd == cmd) {
			rv = stub;
			list_del(&stub->queue);
			break;
		}
	}
	spin_unlock_irqrestore(&eh_deferral_queue_lock, flags);
	return rv;
}

/* dequeue first pending io_stub, return NULL if list empty */
struct io_stub *
mpl_get_first_pend_io(struct mpl_device *device)
{
	struct list_head *first;
	struct io_stub *stub = NULL;
	unsigned long flags;

	spin_lock_irqsave(&pend_queue_lock, flags);
	first = list_first(&device->pending_ios);
	if (first) {
		stub = list_entry(first, struct io_stub, queue);
		list_del(&stub->queue);
		atomic_dec(&device->pending_count);
	}
	spin_unlock_irqrestore(&pend_queue_lock, flags);
	return stub;
}


struct mpl_host *
mpl_find_host(int n)
{
	struct mpl_host *host;

	list_for_each_entry(host, &mpl_hosts, host_list) {
		if (host->host_no == n) {
			return host;
		}
	}
	return NULL;
}

struct mpl_target *
mpl_find_target(struct mpl_path *path)
{
      struct mpl_target *tgtp;

      list_for_each_entry(tgtp, &mpl_targets, target_list) {
              if ((tgtp->host == path->host) &&
                  (tgtp->id == path->target)) {
                      return tgtp;
              }
      }
      return NULL;
}

struct mpl_host *
mpl_find_or_create_host(int n)
{
	struct mpl_host *host;

	host = mpl_find_host(n);
	if (host == NULL) {
		host = kmalloc(sizeof *host, GFP_ATOMIC);
		if (!host) {
			mpl_print(0, MPL_ERR "Cannot allocate host mem\n");
			return NULL;
		}
		memset(host, 0, sizeof(*host));
		host->host_no = n;
		INIT_LIST_HEAD(&host->devices);

		list_add_tail(&host->host_list, &mpl_hosts);
	}
	return host;
}

struct mpl_target *
mpl_find_or_create_target(struct mpl_host *host, struct mpl_path *path)
{
	struct mpl_target *tgtp;
	struct mpl_path *pathp;

	tgtp = mpl_find_target(path);
	if (tgtp)
		goto out;

	tgtp = kmalloc(sizeof(*tgtp), GFP_ATOMIC);
	if (!tgtp)
		goto out;
	memset(tgtp, 0, sizeof(*tgtp));
	INIT_LIST_HEAD(&tgtp->paths);

	tgtp->host = path->host;
	tgtp->id = path->target;

	list_add_tail(&tgtp->target_list, &mpl_targets);
out:
	/* Check to see if target already points to this path */
	list_for_each_entry(pathp, &tgtp->paths, tgt_path_list) {
		if (pathp == path) {
			return tgtp;
		}
	}
	list_add_tail(&path->tgt_path_list, &tgtp->paths);
	atomic_inc(&tgtp->path_ref_count);
	if (path->is_default) {
		atomic_inc(&tgtp->prim_path_ref_count);
	}
	path->tgt = tgtp;
	return tgtp;
}

struct mpl_host *
mpl_find_host_by_scsi_host(struct Scsi_Host *scsi_host)
{
	struct mpl_host *host, *ret_host;

	ret_host = NULL;
	list_for_each_entry(host, &mpl_hosts, host_list) {
		if (host->scsi_host == scsi_host) {
			ret_host = host;
			break;
		}
	}
	return ret_host;
}

static inline int
mpl_mangle_host_pointer(struct mpl_host *host)
{
	return (unsigned long) host >> 8;
}

static inline int
mpl_device_hash_fn(struct mpl_host *host, int target, int lun)
{
	int h;
				/* For lack of imagination... */
	h = mpl_mangle_host_pointer(host) + target * 128 + lun;
	return h & (device_hash_size - 1);

}

static inline int
mpl_path_hash_fn(struct mpl_host *host, int target, int lun)
{
	int h;
				/* For lack of imagination... */
	h = mpl_mangle_host_pointer(host) + target * 128 + lun;
	return h & (path_hash_size - 1);

}

struct mpl_device *
mpl_find_device(struct mpl_host *host, int target, int lun)
{
	struct mpl_device *device;
	int hash = mpl_device_hash_fn(host, target, lun);
	unsigned long flags;

	spin_lock_irqsave(&device_hash_table_lock, flags);
	for (device = device_hash[hash]; device; device = device->next) {
		/* Look for a device that matches the specified presentation id */
		if (device->id.host_no == host->host_no &&
		    device->id.tgt_no  == target  &&
		    device->id.lun_no  == lun)
		{
			spin_unlock_irqrestore(&device_hash_table_lock, flags);
			return device;
		}
	}
	spin_unlock_irqrestore(&device_hash_table_lock, flags);
	return NULL;
}

struct mpl_path *
mpl_find_path(struct mpl_host *host, int target, int lun)
{
	struct mpl_path *path;
	int hash = mpl_path_hash_fn(host, target, lun);
	unsigned long flags;

	spin_lock_irqsave(&path_hash_table_lock, flags);
	for (path = path_hash[hash]; path; path = path->next) {
		if (path->host == host &&
		    path->target  == target  &&
		    path->lun     == lun)
		{
			spin_unlock_irqrestore(&path_hash_table_lock, flags);
			return path;
		}
	}
	spin_unlock_irqrestore(&path_hash_table_lock, flags);
	return NULL;
}

void
mpl_set_dev_ctrl_header(struct mpl_device *device, int flag)
{
	if (flag & MPL_PC_CFG_AUTO_FAILBACK)
		device->auto_failback = 1;
	flag = flag & 0xf;
	switch (flag) {
	case MPL_PC_TYPE_FAILOVER:      /* alternate paths are standby paths */
	case MPL_PC_TYPE_ROUND_ROBIN:   /* alt paths balanced by round robin */
	case MPL_PC_TYPE_DYNAMIC_LUNQ:  /* Dynamic balancing algorithm used */
	case MPL_PC_TYPE_DYNAMIC_TGTQ:  /* Dynamic balancing algorithm used */
	case MPL_PC_TYPE_DYNAMIC_HBAQ:  /* Dynamic balancing algorithm used */
	case MPL_PC_TYPE_DYNAMIC_TIME:  /* Dynamic balancing algorithm used */
		device->path_cntl_type = flag;
		break;
	default:
		device->path_cntl_type = MPL_PC_TYPE_FAILOVER; /* default */
	}
	return;
}

static void
mpl_set_device_nopath(struct mpl_device *device)
{
	init_timer(&device->nopath);
	device->nopath.function = mpl_nopath_timeout;
	device->nopath.expires = 0;
	device->nopath.data = (unsigned long) device;
}

static void
mpl_set_path_hbeat(struct mpl_path *path)
{
	init_timer(&path->hbeat);
	path->hbeat.function = mpl_hbeat_timeout;
	path->hbeat.expires = 0;
	path->hbeat.data = (unsigned long) path;
}

static void
mpl_set_path_cmd_scheduler(struct mpl_path *path)
{
	init_timer(&path->sched_cmd);
	path->sched_cmd.function = mpl_path_schedule_cmd;
	path->sched_cmd.expires = 0;
	path->sched_cmd.data = (unsigned long) path;
}

struct mpl_device *
mpl_find_or_create_device(struct mpl_host *host, int channel,
			  int target, int lun)
{
	struct mpl_device *device;
	unsigned long flags;
	int device_hash_index;

	device = mpl_find_device(host, target, lun);
	if (device)
		goto out;

	/* Allocate a new device and the first
	 * path that gets us to it. */
	device = kmalloc(sizeof(*device), GFP_ATOMIC);
	if (!device)
		goto out;	
	memset(device, 0, sizeof(*device));

	device_hash_index = mpl_device_hash_fn(host, target, lun);
	
	device->host = host;
	device->failing_path = 0;
	device->dev_need_inqsn_data = 1;
	device->inquiry_heartbeat = 1;
	device->state = DEV_FAILED;

	/* By default, all devices / paths are "imported" for multipulse */
	device->configured = CONFIG_UNCONFIGURED;
	
	INIT_LIST_HEAD(&device->pending_ios);
	INIT_LIST_HEAD(&device->paths);
	mpl_set_device_nopath(device);
	if(mpl_nopath_tmo && mpl_enable_nopath) {
		device->nopath_tmo = mpl_nopath_tmo;
	}
	
	/* Take care of presentation id */
	device->id.host_no = host->host_no;
	device->id.tgt_no  = target;
	device->id.lun_no  = lun;
	
	list_add_tail(&device->devices, &device->host->devices);
	
	spin_lock_irqsave(&device_hash_table_lock, flags);
	device->next  = device_hash[device_hash_index];
	wmb();
	device_hash[device_hash_index] = device;
	spin_unlock_irqrestore(&device_hash_table_lock, flags);
	
out:
	return device;
}

void
mpl_inc_can_queue(struct io_stub *stub)
{
	struct mpl_path   *path = stub->path;
	struct mpl_device *device = path->device;
	struct Scsi_Host * sh;

	if (!(stub->flag & MPL_DEC_CAN_QUEUE))
		return;
	stub->flag &= ~MPL_DEC_CAN_QUEUE;

	sh = path->host->scsi_host;
	if (sh->can_queue < mpl_max_can_queue)
		sh->can_queue++;
	if (device->host != path->host) {
		sh = device->host->scsi_host;
		if (sh->can_queue < mpl_max_can_queue)
			sh->can_queue++;
	}
	return;
}

void
mpl_dec_can_queue(struct io_stub *stub)
{
	struct mpl_path   *path = stub->path;
	struct mpl_device *device = path->device;
	struct Scsi_Host * sh;

	if (stub->flag & MPL_DEC_CAN_QUEUE)
		return;
	stub->flag |= MPL_DEC_CAN_QUEUE;
	sh = path->host->scsi_host;
	if (sh->can_queue)
		sh->can_queue--;

	if (device->host != path->host) {
		sh = device->host->scsi_host;
		if (sh->can_queue)
			sh->can_queue--;
	}
	return;
}

/* Presentation path is always path 0.
 * This could translate to hbaN in the base driver.
 */
int
mpl_is_presentation_path(struct mpl_path *path)
{
	return !mpl_is_path_configured(path) ||
		path == list_entry(list_first(&path->device->paths),
				   struct mpl_path,
				   path_list);
}


/* Detects if there is an outstanding command that will change the
 * array controller priority / ownership of a LUN.
 */
int
mpl_is_out_lun_own_chg(struct mpl_device *device)
{
	struct mpl_path *path;

	list_for_each_entry(path, &device->paths, path_list) {
		if (path->flag & PATH_OUT_CHG_LUN_OWN_CMD) {
			return 1;
		}
	}
	return 0;
}


/* By default, primary path is the same as presentation path,
 * until changed by a library call.  Primary path is identified
 * by the path->is_default setting.
 */
struct mpl_path *
mpl_get_primary_path(struct mpl_device *device)
{
	struct mpl_path *path;

	list_for_each_entry(path, &device->paths, path_list) {
		if (mpl_is_path_primary(path)) {
			return path;
		}
	}
	return 0;
}

int
mpl_set_device_auto_failback(struct mpl_device *device, int boolean)
{
	device->auto_failback = boolean;
	return 0;
}

int
mpl_set_device_optimized(struct mpl_device *device, int boolean)
{
	device->optimized_only = boolean;
	return 0;
}

static void
mpl_revoke_other_primaries(struct mpl_path *path)
{
	struct mpl_device *device = path->device;
	struct mpl_path *p;


	list_for_each_entry(p, &device->paths, path_list) {
		if (p != path) {
			if(p->is_default)
				atomic_dec(&p->tgt->prim_path_ref_count);
			p->is_default = 0;
			if(path->state == PATH_ONLINE)
				path->state = PATH_STANDBY;
		}
	}
}

int
mpl_set_path_primary(struct mpl_path *path, int boolean)
{
	struct mpl_path *next_path;
	int new_primary = !path->is_default;
	int status = 0;

	mpl_print(MPL_IOCTL, MPL_INFO "Set primary path for %s %s: "
		  "current = %d: %d %d\n",
		  mpl_print_device(path->device),
		  mpl_print_path(path),
		  mpl_is_path_primary(path),
		  boolean,
		  new_primary);

	if (!boolean) {		/* Let someone grab primary, but don't
				 * arbitrarily try to give up primary.
				 */
		return 0;
	}

	/* Cannot make inactive controller primary */
	if (new_primary && !path->active_controller) {
		status = -EBUSY;
		new_primary = 0;
	}

	/* For ALUA, cannot make non-optimized path primary */
	if (new_primary && path->device->alua_support &&
	   (path->sas != MPL_ALUA_ACTIVE_OPT)) {
		/* Add code here to send STPG */
	}

	if (new_primary) {
		path->is_default = 1;
		atomic_inc(&path->tgt->prim_path_ref_count);
		mpl_print(MPL_IOCTL, MPL_INFO "New primary path for %s %s\n",
			mpl_print_device(path->device), mpl_print_path(path));
		mpl_revoke_other_primaries(path);
		mpl_path_state_machine(path, PATH_UP);
		list_for_each_entry (next_path, &path->device->paths, path_list) {
			if (path != next_path)
				mpl_path_state_machine(next_path, PATH_UP);
		}
	}
	return status;
}

struct mpl_path *
mpl_get_next_path(struct mpl_path *path)
{
	struct mpl_device *device = path->device;
	struct mpl_path *next_path;
	struct list_head *l;

	l = path->path_list.next;
	if (l == &device->paths) {
		l = l->next;
	}

	next_path = list_entry(l, struct mpl_path, path_list);
	return next_path;
}

struct mpl_path *
mpl_get_next_avail_path(struct mpl_path *path, struct mpl_path *first_path)
{
	struct mpl_device *device = path->device;
	struct mpl_path *next_path;
	struct list_head *l;

	do {
		l = path->path_list.next;
		if (l == &device->paths) {
			l = l->next;
		}

		next_path = list_entry(l, struct mpl_path, path_list);
		if (next_path->active_controller &&
		   (next_path->state == PATH_ONLINE ||
		    next_path->state == PATH_STANDBY)) {
			/* Skip path if the optimized bit for the path is set,
			 * and this path is not optimized,
			 * and there are optimized paths available.
			 */
			if(mpl_is_optimized(device) &&
			   next_path->sas != MPL_ALUA_ACTIVE_OPT &&
			   atomic_read(&device->opt_path_cnt)) {
				path = next_path;
				continue;
			}
			return next_path == first_path ? NULL : next_path;
		}
		path = next_path;
	} while (path != first_path);
	return 0;
}

struct mpl_path *
mpl_find_or_create_path(struct mpl_host *host, int target, int lun)
{
	struct mpl_device *device;
	struct mpl_path   *path;

	/* If the path already exists, return it */
	path = mpl_find_path(host, target, lun);
	if (path) {
		return path;
	}

	/* Otherwise create a new device with this path */
	device = mpl_find_or_create_device(host, 0, target, lun);
	if (!device) {
		mpl_print(0, MPL_ERR "find/create device for %s failed\n",
		mpl_print_host(host));
		return 0;
	}

	path = mpl_add_path(device, host, 0, target, lun);
	if (!path) {
		mpl_print(0, MPL_ERR "find/create path for %s failed\n",
		       mpl_print_device(device));
		return 0;
	}
	device->last_path = path;
	return path;
}

/* NOTE:
 *    adding a new path to an existing mpl_device.
 *
 *    device->host == device->1st_path->host
 *    param "host" for this call applies for this new path ONLY
 *    Therefore, device->host and this "host" should NOT be the same.
 */
struct mpl_path *
mpl_add_path(struct mpl_device *device, struct mpl_host *host,
	int channel, int target, int lun)
{
	struct mpl_path   *path;
	unsigned long flags;
	int path_hash_index = mpl_path_hash_fn(host, target, lun);

	path = kmalloc(sizeof(*path), GFP_ATOMIC);
	if (path == NULL) {
		mpl_print(0, MPL_ERR "Could not alloc path\n");
		return 0;
	}

	memset(path, 0, sizeof(*path));
	init_timer(&path->sdev_block);
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
	path->sdev = kmalloc(sizeof(struct scsi_device), GFP_ATOMIC);
	if (path->sdev) {
		memset(path->sdev, 0, sizeof(struct scsi_device));
		path->sdev->host =  host->scsi_host;
		path->sdev->id = target;
		path->sdev->lun = lun;
		path->sdev->channel = channel;
		path->sdev->queue_depth = mpl_get_lun_q_depth(host);
		path->private_sdev = 1;
		mpl_module_unlock();
		if (mpl_base_slave_alloc(path->sdev)) {
			path->private_sdev = 0;
			kfree(path->sdev);
			path->sdev = NULL;
		}
		mpl_module_lock();
	}
#endif
	path->device  = device;
	path->channel = channel;
	path->target  = target;
	path->lun     = lun;
	path->host    = host;
	path->active_controller = 1;
	path->sas = MPL_ALUA_ACTIVE_OPT;
	atomic_inc(&device->opt_path_cnt);

	list_add_tail(&path->path_list, &device->paths);

	spin_lock_irqsave(&path_hash_table_lock, flags);
	path->next = path_hash[path_hash_index];
	wmb();
	path_hash[path_hash_index] = path;
	spin_unlock_irqrestore(&path_hash_table_lock, flags);

	mpl_set_path_hbeat(path);
	mpl_set_path_cmd_scheduler(path);
	path->state = PATH_FAILED;
	path->hbstate = HB_STOPPED;
	path->configured = CONFIG_UNCONFIGURED;

	mpl_find_or_create_target(host, path);
	return path;
}

void
mpl_free_path(struct mpl_path *path)
{
	unsigned long flags;
	int  hash;

	mpl_path_state_machine(path, PATH_DISABLE);
	mpl_set_path_configured(path, CONFIG_UNCONFIGURED);
	hash = mpl_path_hash_fn(path->host, path->target, path->lun);
	if (path->tgt_port_info) {
		mpl_put_scsi_buf(path->tgt_port_info);
		path->tgt_port_info = 0;
	}

	list_del(&path->path_list);

	list_del(&path->tgt_path_list);
	atomic_dec(&path->tgt->path_ref_count);
	if (path->is_default) {
		atomic_dec(&path->tgt->prim_path_ref_count);
	}
	if(!atomic_read(&path->tgt->path_ref_count)) {
		list_del(&path->tgt->target_list);
		kfree(path->tgt);
	}

	spin_lock_irqsave(&path_hash_table_lock, flags);
	if (path_hash[hash] == path) {
		path_hash[hash] = path->next;
	} else {
		struct mpl_path *prev = path_hash[hash];

		while (prev->next && prev->next != path)
			prev = prev->next;
		if (!prev->next) {
			mpl_print(0, MPL_ERR "Cannot find path %p "
				"(%s %s) in path hash.\n",
				path, mpl_print_device(path->device),
				mpl_print_path(path));
			
		} else {
			prev->next = path->next;
		}
	}
	if (del_timer(&path->sdev_block)) {
		mpl_print(0, MPL_ERR "mpl_free_path sdev_block timer active for path %p <%d/%d/%d> \n",
			  path,
			  (path->host == NULL) ? -1 : path->host->host_no,
			  path->target, path->lun);
	}
	spin_unlock_irqrestore(&path_hash_table_lock, flags);
	if (path->private_sdev) {
		path->private_sdev = 0;
		kfree(path->sdev);
		path->sdev = NULL;
	}
	kfree(path);
}

static int
mpl_scan_for_device_hash(struct mpl_device *device)
{
	struct mpl_device *d;
	unsigned long flags;
	int  hash;
				/* If we've lost all paths, need to
				 * scan the hash table to find the
				 * device entry. */
	spin_lock_irqsave(&device_hash_table_lock, flags);
	for (hash = 0; hash < device_hash_size; ++hash) {
		for (d = device_hash[hash]; d; d = d->next) {
			if (d == device) {
				goto out;
			}
		}
	}
	hash = -1;
out:
	spin_unlock_irqrestore(&device_hash_table_lock, flags);
	return hash;
}

static int
mpl_device_is_active(struct mpl_device *device)
{
	struct mpl_path *path;

	if (atomic_read(&device->ref_count)) {
		return 1;
	}
	list_for_each_entry(path, &device->paths, path_list) {
		if (atomic_read(&path->io_count) ||
		    atomic_read(&path->internal_count)) {
			return 1;
		}
	}
	return 0;
}

static void
mpl_remove_device_from_device_hash(struct mpl_device *device)
{
	struct mpl_path *path = mpl_first_path(device);
	unsigned long flags;
	int hash;

	hash = path ? mpl_device_hash_fn(device->host, path->target, path->lun)
		: mpl_scan_for_device_hash(device);
	if (hash < 0) {
		mpl_print(0, MPL_ERR "mpl: device %s not found in hash: %p\n",
			  mpl_print_device(device),
			  device);
		BUG();
	}
	spin_lock_irqsave(&device_hash_table_lock, flags);
	if (!device_hash[hash]) {
		mpl_print(0, MPL_ERR "No such device (%s) in device hash.\n",
			  mpl_print_device(path->device));
	} else if (device_hash[hash] == device) {
		device_hash[hash] = device->next;
	} else {
		struct mpl_device *prev;

		prev = device_hash[hash];
		while (prev->next != NULL && prev->next != device) {
			prev = prev->next;
		};
		if (prev->next == NULL) {
			mpl_print(0, MPL_ERR "Cannot find device %p "
				"(%s) in device hash.\n",
				device, mpl_print_device(path->device));
		} else {
			prev->next = device->next;
		}
	}
	spin_unlock_irqrestore(&device_hash_table_lock, flags);
}

int
mpl_free_device(struct mpl_device *device)
{
	struct list_head *l, *n;

	if (device == NULL) {
		mpl_print(0, MPL_ERR "No device to free\n");
		return 1;
	}

	if (mpl_device_is_active(device)) {
		return 1;
	}

	mpl_device_state_machine(device, NULL, DEV_DISABLE);
	/* free all paths */
	list_for_each_safe(l, n, &device->paths) {
		mpl_free_path(list_entry(l, struct mpl_path, path_list));
	}
	mpl_set_device_configured(device, CONFIG_UNCONFIGURED);
	mpl_module_unlock();
	del_timer_sync(&device->nopath);
	mpl_module_lock();

	list_del(&device->devices);
	kfree(device);
	return 0;
}

static int
mpl_stub_hash_fn(struct scsi_cmnd *cmd)
{
	return cmd->serial_number & (stub_hash_size - 1);
}

int mpl_find_stub_in_hash(struct io_stub *item)
{
	struct io_stub *stub, *prev;
	int hash = mpl_stub_hash_fn(item->cmd);
	unsigned long  flags;
	int rc = 0;

	spin_lock_irqsave(&stub_hash_table_lock, flags);
	stub = stub_hash[hash];
	if (stub && stub == item) {
		rc = 1;
		goto unlock_return;
	}
	if (!stub) {
		goto unlock_return;
	}
	prev = stub;
	for (stub = stub->next; stub; stub = stub->next) {
		if (stub == item) {
			rc = 1;
			goto unlock_return;
		}
		prev = stub;
	}

unlock_return:
	spin_unlock_irqrestore(&stub_hash_table_lock, flags);
	return rc;
}

void
mpl_add_stub_to_hash(struct io_stub *stub)
{
	int hash = mpl_stub_hash_fn(stub->cmd);
	unsigned long flags;
	static int dup;

	if (mpl_find_stub_in_hash(stub)) {
		if (dup % 32 == 0) {
			mpl_print(0, MPL_ERR "Duplicate stub add: %p from  %p, "
				  "%s %s, cmd = %02x %02x %02x %02x\n",
		       stub, (&stub)[-1],
		       mpl_print_device(stub->path->device),
		       mpl_print_path(stub->path),
		       stub->cmd->cmnd[0], stub->cmd->cmnd[1],
		       stub->cmd->cmnd[2], stub->cmd->cmnd[3]);
			WARN_ON(1);
		}
		++dup;
		return;
	}
	spin_lock_irqsave(&stub_hash_table_lock, flags);
	stub->next = stub_hash[hash];
	wmb();
	stub_hash[hash] = stub;
	spin_unlock_irqrestore(&stub_hash_table_lock, flags);
}

struct io_stub *
mpl_locate_stub_hash(struct scsi_cmnd *cmd)
{
	struct io_stub *stub, *prev;
	int hash = mpl_stub_hash_fn(cmd);
	unsigned long  flags;

	spin_lock_irqsave(&stub_hash_table_lock, flags);
	stub = stub_hash[hash];
	if (!stub) {
		goto unlock_return;
	}
	
	if (stub->cmd == cmd) {
		goto unlock_return;
	}
	prev = stub;
	for (stub = stub->next; stub; stub = stub->next) {
		if (stub->cmd == cmd) {
			goto unlock_return;
		}
		prev = stub;
	}

unlock_return:
	spin_unlock_irqrestore(&stub_hash_table_lock, flags);
	return stub;
}

struct io_stub *
mpl_find_stub_hash(struct scsi_cmnd *cmd)
{
	struct io_stub *stub;

	stub = mpl_locate_stub_hash(cmd);
	if (!stub) {
		mpl_print(0, MPL_ERR "No stub found: %p from  %p, %d/%d/%d "
		  "cmd = %02x %02x %02x %02x\n",
		  cmd, (&cmd)[-1],
		  mpl_host_no_from_cmd(cmd),
		  mpl_target_no_from_cmd(cmd),
		  mpl_lun_no_from_cmd(cmd),
		  cmd->cmnd[0], cmd->cmnd[1], cmd->cmnd[2], cmd->cmnd[3]);
	}
	return stub;
}

static struct io_stub *
mpl_find_stub_remove_hash_quiet(struct scsi_cmnd *cmd)
{
	struct io_stub *stub, *prev;
	int hash = mpl_stub_hash_fn(cmd);
	unsigned long  flags;

	spin_lock_irqsave(&stub_hash_table_lock, flags);
	stub = stub_hash[hash];
	if (!stub) {
		goto unlock_return;
	}
	
	if (stub->cmd == cmd) {
		stub_hash[hash] = stub->next;
		goto unlock_return;
	}
	prev = stub;
	for (stub = stub->next; stub; stub = stub->next) {
		if (stub->cmd == cmd) {
			prev->next = stub->next;
			goto unlock_return;
		}
		prev = stub;
	}

unlock_return:
	spin_unlock_irqrestore(&stub_hash_table_lock, flags);
	return stub;
}

struct io_stub *
mpl_find_stub_remove_quiet(struct scsi_cmnd *cmd)
{
	struct io_stub *stub;

	stub = mpl_find_stub_remove_hash_quiet(cmd);
	if (!stub) {
		stub = mpl_find_stub_remove_deferral(cmd);
	}
	if (!stub) {
		stub = mpl_find_stub_remove_eh_deferral(cmd);
	}
	return stub;
}

struct io_stub *
mpl_find_stub_remove(struct scsi_cmnd *cmd)
{
	struct io_stub *stub;

	stub = mpl_find_stub_remove_quiet(cmd);
	if (!stub) {
		mpl_print(0, MPL_ERR "No stub to remove: %p from  %p, %d/%d/%d "
			  "cmd = %02x %02x %02x %02x\n",
			  cmd, (&cmd)[-1],
			  mpl_host_no_from_cmd(cmd),
			  mpl_target_no_from_cmd(cmd),
			  mpl_lun_no_from_cmd(cmd),
			  cmd->cmnd[0], cmd->cmnd[1], cmd->cmnd[2], cmd->cmnd[3]);
	}
	return stub;
}

struct io_stub *
mpl_find_stub_pend_io(struct scsi_cmnd *cmd)
{
	struct mpl_host *host;
	struct mpl_device *device;
	struct io_stub *stub;
	int  i;
	unsigned long flags;

	stub = NULL;
	/* First check pending_ios queue for ALL devices */
	for (i = 0; ; i++) {
		host = mpl_find_host(i);
		if (host == NULL) {
			break;
		}
		list_for_each_entry(device, &host->devices, devices) {
			spin_lock_irqsave(&pend_queue_lock, flags);
			list_for_each_entry(stub, &device->pending_ios, queue) {
				if (stub->cmd == cmd) {
					spin_unlock_irqrestore(&pend_queue_lock, flags);
					return stub;
				}
			}
			spin_unlock_irqrestore(&pend_queue_lock, flags);
		}
	}
	return NULL;
}

struct io_stub *
mpl_find_stub_remove_pend_io(struct scsi_cmnd *cmd, int flag)
{
	struct mpl_host *host;
	struct mpl_device *device;
	struct io_stub *stub;
	int  i;
	int  rm = 0;
	unsigned long flags;
	int ignore_aborted_cmd = flag & MPL_ABORT_IGNORE_ABORTED;

	flag &= MPL_ABORT_TYPE_MASK;
	/* First check pending_ios queue for ALL devices */
	for (i = 0; ; i++) {
		host = mpl_find_host(i);
		if (host == NULL) {
			break;
		}

		if (flag == MPL_ABORT_HOST &&
		    mpl_scsi_host_from_cmd(cmd) == host->scsi_host &&
		    cmd->device->host == host->scsi_host) {
			rm = 1; /* remove everything on this host */
		}

		list_for_each_entry(device, &host->devices, devices) {
			spin_lock_irqsave(&pend_queue_lock, flags);
			list_for_each_entry(stub, &device->pending_ios, queue) {
				if ((flag == MPL_ABORT_DEVICE) &&
		    		   (cmd->device == stub->cmd->device)) {
					rm = 1; /* remove everything on this device */
				}
				if ((flag == MPL_ABORT_EXACT) &&
		    			(cmd == stub->cmd)) {
					rm = 1; /* remove this specific IO */
				}
				if(rm) {
					list_del(&stub->queue);
					atomic_dec(&device->pending_count);
					spin_unlock_irqrestore(&pend_queue_lock, flags);
					return stub;
				}
			}
			spin_unlock_irqrestore(&pend_queue_lock, flags);
		}
	}

	/* Next lets check if its on the deferral list */
	spin_lock_irqsave(&deferral_queue_lock, flags);
	list_for_each_entry(stub, &deferral_queue, queue) {
		if (!stub->done) {
			continue;  /* skip internal commands */
		}
		/* skip already aborted commands */
		if ((ignore_aborted_cmd) && 
		    (stub->reason == MPL_ABORTED_CMD)) { 
			continue;
		}
		if ((flag == MPL_ABORT_DEVICE) &&
    		   (cmd->device == stub->cmd->device)) {
			rm = 1; /* remove everything on this device */
		}
		if ((flag == MPL_ABORT_EXACT) &&
    			(cmd == stub->cmd)) {
			rm = 1; /* remove this specific IO */
		}
		if(rm) {
			list_del(&stub->queue);
			atomic_dec(&stub->path->device->deferral_count);
			spin_unlock_irqrestore(&deferral_queue_lock, flags);
			return stub;
		}
	}
	spin_unlock_irqrestore(&deferral_queue_lock, flags);

	/* Next lets check if its on the eh_deferral list */
	spin_lock_irqsave(&eh_deferral_queue_lock, flags);
	list_for_each_entry(stub, &eh_deferral_queue, queue) {
		if (!stub->done) {
			continue;  /* skip internal commands */
		}
		/* This command needs abort to the base driver */
		if (stub->reason == MPL_DO_TMO) {
			continue;
		}
		if ((flag == MPL_ABORT_DEVICE) &&
    		   (cmd->device == stub->cmd->device)) {
			rm = 1; /* remove everything on this device */
		}
		if ((flag == MPL_ABORT_EXACT) &&
    			(cmd == stub->cmd)) {
			rm = 1; /* remove this specific IO */
		}
		if (rm) {
			list_del(&stub->queue);
			atomic_dec(&stub->path->device->eh_deferral_count);
			spin_unlock_irqrestore(&eh_deferral_queue_lock, flags);
			return stub;
		}
	}
	spin_unlock_irqrestore(&eh_deferral_queue_lock, flags);
	return 0;
}

static struct mpl_plugin *
mpl_find_plugin(struct scsi_cmnd *cmd, struct mpl_device *device)
{
	struct mpl_plugin *plugin;
	uint8_t *request = cmd->request_buffer;

	for (plugin = &plugin_data_start;
	     plugin < &plugin_data_end;
	     ++plugin) {
		if (plugin->probe && plugin->probe(cmd)) {
			device->scsi_type = request[0] & 0x1f;
			memcpy(device->vendor, request + 8, 8);
			memcpy(device->product, request + 16, 16);
			memcpy(device->rev, request + 32, 4);
			return plugin;
		}
	}
	return NULL;
}

void
mpl_cmd_restore(struct scsi_cmnd *cmd, struct io_stub *stub)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	cmd->host = stub->scsi_host;
	cmd->target = stub->target;
	cmd->lun = stub->lun;
#else
	cmd->device = stub->sdev;
#endif
}

void
mpl_free_internal_stub(struct io_stub *stub)
{
	if (stub->cmd->request_buffer) {
		mpl_put_scsi_buf(stub->cmd->request_buffer);
	}
	mpl_inc_can_queue(stub);
	mpl_put_scsi_cmd(stub->cmd);
	mpl_stub_free(stub);
	return;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
void
mpl_throttle_up_queue_full(struct io_stub *stub)
{
	struct scsi_device *sdev;
	struct mpl_device *device;
	unsigned short depth;
	int tag_type;
#ifndef MSG_ORDERED_TAG
#define MSG_ORDERED_TAG 0x22
#define MSG_SIMPLE_TAG  0x20
#endif	
	if ((!stub->sdev) || (!stub->path->device)){
		return;
	}
	device = stub->path->device;
	sdev = stub->sdev;

	depth = sdev->queue_depth;
	if((!sdev->last_queue_full_count) ||    // never changed yet ???
	   (!device->queue_max) ||		// not init yet ???
	   (device->queue_max == depth)){  	// up to max
		return;
	}

	if ((jiffies >> 4) > sdev->last_queue_full_time + 500){
		depth++;
		if(sdev->ordered_tags)
			tag_type = MSG_ORDERED_TAG;
		else
			tag_type = MSG_SIMPLE_TAG;

		scsi_adjust_queue_depth(sdev, tag_type, depth);
		sdev->last_queue_full_time = jiffies >> 4;
	}
}

void
mpl_throttle_down_queue_full(struct io_stub *stub)
{
	struct scsi_device *sdev;
	struct mpl_device *device;
	unsigned short current_depth, half_depth;

	if (!stub->sdev || !stub->path->device) {
		return;
	}
	device = stub->path->device;
	sdev = stub->sdev;
	if (!sdev->last_queue_full_count) {  /*  never changed ? */
		device->queue_max = sdev->queue_depth;
	}

	half_depth = device->queue_max >> 1;
	current_depth = sdev->queue_depth;
	if (current_depth == device->queue_max || current_depth > half_depth) {
		current_depth = half_depth;
		scsi_track_queue_full(sdev, current_depth);
	} else if (current_depth > 3){
		--current_depth;
		scsi_track_queue_full(sdev, current_depth);
	}
}
#endif

void
mpl_call_done_free_stub(struct io_stub *stub, int needs_lock)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	unsigned long flags;
#else
	struct mpl_path *presentation = mpl_first_path(stub->path->device);
#endif
	if (!mpl_is_presentation_path(stub->path)) {
		mpl_cmd_restore(stub->cmd, stub);
	}
	stub->cmd->host_scribble = NULL;
	stub->cmd->scsi_done = stub->done;
	mpl_module_unlock();
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	if (needs_lock) {
		spin_lock_irqsave(&io_request_lock, flags);
		stub->done(stub->cmd);
		spin_unlock_irqrestore(&io_request_lock, flags);
	}
	else {
		stub->done(stub->cmd);
	}
#else
	if (mpl_handle_qfull){
		if ((stub->cmd->result == SAM_STAT_TASK_SET_FULL) ||
		   (stub->cmd->result == SAM_STAT_BUSY)) {
			mpl_throttle_down_queue_full(stub);
		} else {
			mpl_throttle_up_queue_full(stub);
		}
	}

	if (presentation) {
		stub->sdev->queue_depth = presentation->sdev->queue_depth;
		stub->sdev->ordered_tags = presentation->sdev->ordered_tags;
		stub->sdev->simple_tags = presentation->sdev->simple_tags;
	}

	stub->cmd->device = stub->sdev;
	stub->cmd->eh_timeout.function = stub->timeout_handler;
	stub->done(stub->cmd);
#endif
	mpl_module_lock();
	mpl_stub_free(stub);
}

void
mpl_add_pend_io(struct io_stub *stub, struct mpl_device *device, int need_lock)
{
	unsigned long flags;

	/* Double check to ensure we are in the right state */
	if (((device->state == DEV_FAILED) && !timer_pending(&device->nopath)) ||
		(device->state == DEV_DISABLED)) {
		if (stub->done) {
			stub->cmd->result = ScsiResult(DID_NO_CONNECT, 0);
			mpl_call_done_free_stub(stub, need_lock);
		} else {
			mpl_free_internal_stub(stub);
		}
		return;
	}
	stub->cmd->host_scribble = (unsigned char *)stub;
	spin_lock_irqsave(&pend_queue_lock, flags);
	list_add_tail(&stub->queue, &device->pending_ios);
	atomic_inc(&device->pending_count);
	spin_unlock_irqrestore(&pend_queue_lock, flags);
	return;
}

static void
mpl_retry_pending_io(struct io_stub *stub)
{
	struct mpl_device *device = stub->path->device;
	struct mpl_path *path = device->last_path;

	memset(&stub->cmd->sense_buffer[0], 0, SCSI_SENSE_BUFFERSIZE);
	if (((device->state == DEV_FAILED) && !timer_pending(&device->nopath)) ||
		(device->state == DEV_DISABLED)) {
		if (stub->done) {
			stub->cmd->result = ScsiResult(DID_NO_CONNECT, 0);
			mpl_call_done_free_stub(stub, 1);
		} else {
			mpl_free_internal_stub(stub);
		}
	} else if (path->host->blocked || path->lld_blocked ||
		   timer_pending(&device->nopath) ||
		   mpl_queuecommand_alternate_path(device, path, stub)) {
		mpl_add_pend_io(stub, device, 1);
	}
	return;
}

static int is_inq83_valid(struct scsi_cmnd  *cmd)
{
	int response_length = cmd->request_bufflen - cmd->resid;
	uint8_t  *data = (uint8_t *)cmd->request_buffer;
	uint32_t size = data[7];

	return (cmd->result == 0 &&
		(data[0] >> 5) == 0 &&          /* peripheral qualifier */
		response_length > 8 &&          /* truncation */
		response_length >= (size + 8));
}

static int
mpl_compare_serial_number(struct mpl_device *device, uint8_t *serial, int size)
{
	return device->serial_size != size || memcmp(device->serial_number,
						     serial, size);
}

int
mpl_setup_inq_lun_skip(struct scsi_cmnd *cmd)
{
	unsigned char *request = cmd->request_buffer;
	uint8_t *sns = cmd->sense_buffer;

	if ((cmd->cmnd[0] == INQUIRY) && (cmd->cmnd[2] == 0)) {
		/* INQ Page 0 from midlayer scan */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#if defined(RHEL_FC)
		if (cmd->device->lun == 0) {
			*request++ = 0x3;   /* Processor type device */
			memset(sns, 0, SCSI_SENSE_BUFFERSIZE);
			cmd->result = ScsiResult(DID_OK, 0);
			return cmd->result;
		} else {
			memset(request, 0, cmd->request_bufflen); 
			*request++ = 0x60;  /* Lun not supported (pqb = 3) */
		}
#else /* not RHEL_FC -> is SLES_FC */
		memset(request, 0, cmd->request_bufflen);
		*request++ = 0x3f;   /* Lun supported but not available (pqb = 1) */
#endif
#else
		memset(request, 0, cmd->request_bufflen); 
		if (cmd->device->lun == 0)
			*request++ = 0x3;   /* Processor type device */
		else {
			if (mpl_pfq_use_lun_notsup)
				*request++ = 0x60; /* Lun not supported (pqb = 3) */
			else
				*request++ = 0x3f; /* Lun supported not available (pqb = 1) */
		}
#endif
		*request++ = 0x0;
		*request++ = 0x3;
		*request++ = 0x12;
		*request++ = 0x1f;  /* 2.6 midlayer expects atleast 36 byte 
				      inquiry response */
		*request++ = 0x0;
		*request++ = 0x0;
		*request++ = 0x2;
		memset(sns, 0, SCSI_SENSE_BUFFERSIZE);
		cmd->result = ScsiResult(DID_OK, 0);
	}
	else {
		cmd->result = ScsiResult(DID_NO_CONNECT, 0);
	}
	return cmd->result;
}

int 
mpl_no_alternate_path (struct mpl_device *device, struct mpl_path *path)
{
	struct mpl_path *p;
	
	list_for_each_entry(p, &device->paths, path_list) {
		if(p != path){
			return (0);
		}
	}
	return(1);
}

static int
mpl_auto_primary_select(struct mpl_device *device, struct mpl_path *path)
{
	int use_path = 0;
	int cnt, pcnt;
	struct mpl_path *prim_path;

	/*  Are we using first path assigned is primary path */
	if(!mpl_auto_controller && !mpl_auto_lun_balance)
		return 0;

	/* First make sure we meet auto_controller requirements */
	if(mpl_auto_controller) {
		if(device->active_passive) {
			/* active - passive uses first active controller */
			if (path->active_controller) {
				use_path = 1;
			}
		}
		else {
			/* active - active uses first optimized controller,
			 * if there are any.
			 */
			if ((path->sas == MPL_ALUA_ACTIVE_OPT) ||
			     !(atomic_read(&device->opt_path_cnt))) {
				use_path = 1;
			}
		}
		if(!use_path) {
			/* if this is the only path available, use it anyway */
			use_path = mpl_no_alternate_path(device, path);
			if(!use_path)
				return 0;
		}
	}

	/* Is primary path already assigned */
	if((prim_path = mpl_get_primary_path(device))) {
		/* If the primary path is not preferred, switch */
		if (device->alua_support &&
		    (path->sas == MPL_ALUA_ACTIVE_OPT) &&
		    (prim_path->sas != MPL_ALUA_ACTIVE_OPT)) {
			use_path = 1;
			mpl_revoke_other_primaries(path);
			goto primit;
		}
		if(mpl_auto_lun_balance) {
			pcnt = atomic_read(&prim_path->tgt->prim_path_ref_count);
			cnt = atomic_read(&path->tgt->prim_path_ref_count);

			/* Balance primary path across targets */
			if ((pcnt > 0) && ((pcnt - 1) >  cnt)) {
				use_path = 1;
				mpl_revoke_other_primaries(path);
				goto primit;
			}
		}
		return 0;
	}


primit:
	if (use_path) {
		path->is_default = 1;
		atomic_inc(&path->tgt->prim_path_ref_count);

		mpl_print(MPL_INIT, MPL_INFO " Set Primary %s for %s: tgt cnt %d\n",
			  mpl_print_device(path->device),
			  mpl_print_path(path),
			  atomic_read(&path->tgt->prim_path_ref_count));
	}
	return use_path;
}

static int
mpl_export_list(struct mpl_path *path)
{
	/* Check if device is on the export List
	 * return (1) true if it is.
	 */
	if (path->configured == CONFIG_UNAVAILABLE) {
		mpl_print(MPL_INIT, MPL_INFO " %s UNAVAILABLE %s \n",
			  mpl_print_path(path),
			  mpl_print_device(path->device));
		return 1;
	}
	return 0;
}

static int
mpl_hbeat_inquirysn(struct io_stub *stub)
{
	struct scsi_cmnd  *cmd = stub->cmd;
	struct mpl_path   *path = stub->path;
	struct mpl_device *device = path->device;
	uint8_t *request = cmd->request_buffer;
	int  rv = 0, serial_size = 0;
	uint8_t serialbuff[SCSI_INQ_BUFF_LEN];

	if (driver_unloading || path->state == PATH_DISABLED) {
		goto inqsn_exit;
	}

	if (cmd->result != 0) {
				/* cmd failed */
		mpl_print(MPL_RECOVERY, MPL_WARNING "InqSN for %s %s failed "
			  "but this should have been processed earlier"
			  " %x cnt %x: %02x/%02x/%02x\n", 
			  mpl_print_device(path->device),
			  mpl_print_path(path),
			  cmd->result, atomic_read(&int_cmd_cnt),
			  cmd->sense_buffer[2], cmd->sense_buffer[12],
			  cmd->sense_buffer[13]);
		rv = -1;
		goto inqsn_exit;
	}

	if (!is_inq83_valid(cmd)) {
				/* ignore inq 83 response that makes
				 * no sense to us. */
		mpl_print(MPL_HEARTBEAT, MPL_WARNING
			  "Invalid InqSN response for %s %s, ignoring.\n",
			  mpl_print_device(path->device),
			  mpl_print_path(path));
		goto inqsn_exit;
	}

	if (device->plugin &&
	    device->plugin->inq_validation) {
		if (device->plugin->inq_validation(path, request, serialbuff,
		    &serial_size)){
			mpl_print(MPL_HEARTBEAT, MPL_WARNING
			  "Invalid InqSN Data for %s %s, ignoring.\n",
			  mpl_print_device(path->device),
			  mpl_print_path(path));
			goto inqsn_exit;
		}
	}
	else {
		if(mpl_get_validation_id(request, serialbuff, &serial_size)){
			mpl_print(MPL_HEARTBEAT, MPL_WARNING
			  "Invalid InqSN Data for %s %s, ignoring.\n",
			  mpl_print_device(path->device),
			  mpl_print_path(path));
			goto inqsn_exit;
		} 
	} 

	if(!device->serial_size){
		struct mpl_device *d;
				/* If we already have this device
				 * defined, we need to redirect local
				 * pointers to the older device and
				 * prepare to drop the current
				 * device.  */
		if (mpl_export_list(path)) {
			mpl_set_path_configured(path, CONFIG_UNAVAILABLE);
		}
		else if (path->delayed_cmd && mpl_auto_configure &&
			 !device->scsi_type) {
			mpl_set_device_configured(device, CONFIG_CONFIGURED);
			mpl_set_path_configured(path,CONFIG_CONFIGURED);

			/* The first path we find will be the primary */
			if(!mpl_auto_controller && !mpl_auto_lun_balance) {
				path->is_default = 1;
				atomic_inc(&path->tgt->prim_path_ref_count);
			}
			mpl_set_dev_ctrl_header(device, mpl_auto_configure);

			/*  set auto failback by default for auto configured
			 * devices. 
			 */
			device->auto_failback = 1;
		}	

		/* Save necessary info from inq response */
		device->dev_need_inqsn_data = 0;
		device->serial_size = serial_size;
		memcpy(device->serial_number, serialbuff, serial_size);

		d = mpl_match_device_by_serial(device, serialbuff, serial_size);
 		mpl_print(MPL_RECOVERY, MPL_WARNING "InqSN data needed for "
			"%s %s (%d) type %d config %d/%d\n", 
			mpl_print_device(device),
			mpl_print_path(path),
			(d != device),
		    	device->scsi_type, device->configured, path->configured);

		if (path->delayed_cmd) {

			if (mpl_auto_configure && !device->scsi_type) {
				/* Change primary path based on mpl_auto_controller
				 * and mpl_auto_lun_balance.
				 */
				mpl_auto_primary_select(d, path);
			}

			/* This is only if mpl_match_device_by_serial moved the path
			 * to a different device, thus its not the presentation path.
			 * The first path to a device will always be the presentation path.
			 */
			if ((path->device != device) &&
			    (path->configured != CONFIG_UNAVAILABLE)) {
				mpl_setup_inq_lun_skip(path->delayed_cmd->cmd);
			}

			mpl_call_done_free_stub(path->delayed_cmd, 1);
			path->delayed_cmd = NULL;
		}

	} else if (mpl_compare_serial_number(device, serialbuff, serial_size) != 0) {
		mpl_print(MPL_RECOVERY, MPL_WARNING "InqSN data for %s %s: "
			  "did not match\n", 
			  mpl_print_device(device),
			  mpl_print_path(path));
		rv = -1;
	} else {
		mpl_print(MPL_RECOVERY, MPL_INFO "InqSN Heartbeat success "
			  "for %s %s\n",
			  mpl_print_device(device),
			  mpl_print_path(path));

				/* inq sn data matched, but if path is
				 * ONLINE or DISABLED, do nothing */
	}


inqsn_exit:

	if (path->delayed_cmd) {
		mpl_call_done_free_stub(path->delayed_cmd, 1);
		path->delayed_cmd = NULL;
	}
	return rv;
}

/* default to 30 second heart beat */
static void
mpl_hbeat_timeout(unsigned long data)
{
	struct mpl_path *path = (struct mpl_path *)data;

	mpl_module_lock();
	/* Don't send heart beat and don't restart timer if the path has been deconfigured.
	 * For initial Inquiry scan, let unconfigured paths thru.
	 */
	if (!mpl_is_path_configured(path) && !path->device->dev_need_inqsn_data) {
		path->hbstate = HB_STOPPED;
	}
	else {
		path->flag &= ~PATH_FAST_POLL;
		mpl_hbeat_state_machine(path, HB_TIMEOUT);
	}
	mpl_module_unlock();
	return;
}

/* default to 30 second heart beat */
static void
mpl_nopath_timeout(unsigned long data)
{
	struct mpl_device *device = (struct mpl_device *)data;

	mpl_module_lock();

	mpl_print(MPL_IOPATH, MPL_INFO "NoPath tmo for %s: state %d",
	  mpl_print_device(device), device->state);

	if (device->state == DEV_FAILED) {
		mpl_device_process_pendq(device);
	}
	mpl_module_unlock();
	return;
}

int
mpl_quiesce_device(struct mpl_device *device, int tmo)
{
	struct mpl_path *path;
	int  n, cnt;

	cnt = 0;
	do {
		n = 0;
		list_for_each_entry(path, &device->paths, path_list) {
			n += atomic_read(&path->io_count);
			n += atomic_read(&path->internal_count);
		}
		if (n) {
			mpl_module_unlock();
			set_current_state(TASK_UNINTERRUPTIBLE);
			schedule_timeout(2);
			mpl_module_lock();
			cnt += 2;
		}
		if (tmo && (cnt > tmo))
			break;
	} while (n);
	return n;
}

#define configure_transition(from, to) ((from << 3) | to)

enum mpl_config_state
mpl_set_path_configured(struct mpl_path *path,
			enum mpl_config_state configured)
{
	switch (configure_transition(path->configured, configured)) {
	case configure_transition(CONFIG_UNCONFIGURED, CONFIG_UNCONFIGURED):
	case configure_transition(CONFIG_CONFIGURED,   CONFIG_CONFIGURED):
	case configure_transition(CONFIG_UNAVAILABLE,  CONFIG_UNAVAILABLE):
				/* Do nothing. */
		break;

	case configure_transition(CONFIG_UNCONFIGURED, CONFIG_CONFIGURED):
	case configure_transition(CONFIG_UNAVAILABLE,  CONFIG_CONFIGURED):
	case configure_transition(CONFIG_UNCONFIGURED, CONFIG_UNAVAILABLE):
	case configure_transition(CONFIG_UNAVAILABLE,  CONFIG_UNCONFIGURED):
		path->static_config = 0;
		path->configured = configured;
		break;

	case configure_transition(CONFIG_CONFIGURED, CONFIG_UNAVAILABLE):
		return -EBUSY;

	case configure_transition(CONFIG_CONFIGURED, CONFIG_UNCONFIGURED):
		if (mpl_is_device_configured(path->device) &&
		    mpl_is_path_primary(path))
			return -EBUSY;
		mpl_path_state_machine(path, PATH_DISABLE);

		path->is_default = 0;
		path->static_config = 0;
		path->configured = configured;
		break;

	default:
		mpl_print(0, MPL_ERR "Invalid path configuration transition: "
			  "%#x to %#x, %s %s\n",
			  path->configured,
			  configured,
			  mpl_print_device(path->device),
			  mpl_print_path(path));
		break;
	}
	return 0;
}

void
dump_stub_hash(void)
{
	struct io_stub *stub;
	int  i;

	for (i = 0; i < stub_hash_size; ++i) {
		int j = 0;

		stub = stub_hash[i];
		while(stub) {
			if (++j > 16) {
				printk(KERN_ALERT "...\n");
				break;       
			}
			printk(KERN_ALERT "stub %p -> cmd %p\n",
			       stub, stub->cmd);
			stub = stub->next;
		}
	}
}

void
mpl_internal_done(struct scsi_cmnd *cmd)
{
	struct io_stub *stub;
	struct mpl_path *path;
	struct mpl_device *device;

        /* If timer fired, this command will be handled by timeout handler */
        if (!del_timer(&cmd->eh_timeout)) {
                return;
        }

	/* io_request_lock is held when called by lpfc driver */
	mpl_module_lock();

	stub = mpl_find_stub_remove(cmd);
	if (!stub) {
		dump_stub_hash();
		BUG();
	}

	path = stub->path;
	device = path->device;
	if ((cmd->cmnd[0] == REPORT_TGT_PORT_GROUP) ||
	    (cmd->cmnd[0] == SET_TGT_PORT_GROUP)) {
		device->tpg_cmd_inp = 0;
	}


	if (driver_unloading) {
		mpl_free_internal_stub(stub);
		atomic_dec(&path->internal_count);
		atomic_dec(&int_cmd_cnt);
		mpl_module_unlock();
		return;
	}

	mpl_add_deferral(stub, MPL_INT_CMD);
	mpl_module_unlock();
	return;
}

void
mpl_done(struct scsi_cmnd *cmd)
{
	struct io_stub *stub;
	struct mpl_path *path;
	struct mpl_device *device;
	uint32_t cnt;
	uint32_t outio = 1;
	unsigned long tm, avg;

	/* io_request_lock is held when called by lpfc driver */
	mpl_module_lock();

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
	/* If the timer is already fired, do nothing */
	if (!del_timer(&cmd->eh_timeout)) {
		mpl_module_unlock();
		return;
	} else {
		cmd->eh_timeout.expires = jiffies + cmd->timeout_per_command;
		add_timer(&cmd->eh_timeout);
	}
#endif

	stub = mpl_find_stub_remove(cmd);
	if (!stub) {
		dump_stub_hash();
		mpl_module_unlock();
		BUG();
	}

	if (stub->reason == MPL_DO_ABORT || stub->reason == MPL_DO_TMO) {
		mpl_add_stub_to_hash(stub);
		mpl_module_unlock();
		return;
	}

	cmd->host_scribble = (unsigned char *)stub;
	path = stub->path;
	device = path->device;

	atomic_dec(&path->io_count);
	atomic_dec(&path->tgt->io_count);
	atomic_dec(&device->io_count);
	atomic_dec(&path->host->io_count);
	if (!(atomic_read(&path->io_count))) {
		/* If there are no outstanding IOs on this path,
		 * zero the avg time.
		 */
		atomic_set(&path->avg_cnt, 0);
		path->avg_time = 0;
		outio = 0; /* outstanding IOs? */
	}

	/* 
	   2.6 kernel issue SCSI_CACHE_SYNC command while unloading LLD 
	   the mpulse driver should complete it correctly.
	*/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	if (driver_unloading) {
		mpl_free_internal_stub(stub);
		atomic_dec(&int_cmd_cnt);
		mpl_module_unlock();
		return;
	}
#endif
	if ((cmd->cmnd[0] == INQUIRY) && (cmd->cmnd[2] == 0) && !path->seen_ext_inq) {
		/* INQ Page 0 from midlayer scan */
		path->seen_ext_inq = 1;
		mpl_add_deferral(stub, MPL_INQ_CMD);
		mpl_module_unlock();
		return;
	} else if (cmd->result) {
		if (path->state == PATH_DISABLED) {
			mpl_add_deferral(stub, MPL_RETRY_CMD);
			mpl_module_unlock();
			return;
		}
		else if (mpl_is_path_configured(path) || (stub->cmd->cmnd[0] == TEST_UNIT_READY)) {
			/* All TURs should go thru plugin diagnostic check */
			mpl_add_deferral(stub, MPL_FAILED_CMD);
			mpl_module_unlock();
			return;
		}
	}

	/* Calculate average times for good IOs */
	if (outio) {
		cnt = atomic_read(&path->avg_cnt);
		tm = jiffies - stub->start_time;
		avg = ((path->avg_time * cnt) + tm) / (cnt + 1);
		atomic_inc(&path->avg_cnt);
		path->avg_time = avg;
	}

	mpl_call_done_free_stub(stub, 0);
	mpl_module_unlock();
	return;
}

struct mpl_path *
mpl_round_robin_path(struct mpl_device *device, struct scsi_cmnd *cmd)
{
	struct mpl_path   *path;
	struct mpl_path   *first_path;

	/* For now just implement round robin, 1 IO at a time */
	path = device->last_path;
	first_path = path;
	if((path = mpl_get_next_avail_path(path, first_path))) {
		device->last_path = path;
	}
	return device->last_path;
}

struct mpl_path *
mpl_load_balance_time(struct mpl_device *device, struct scsi_cmnd *cmd)
{
	struct mpl_path   *path;
	struct mpl_path   *first_path;
	struct mpl_path   *low_path;
	unsigned long      tm, path_tm;

	low_path = NULL;
	tm = 0xffffffff;

	/* Load balance based on the lowest path average IO time */

	/* The LUN paths are from a specific HBA Port WWPN to a specific
	 * Controller Port WWPN LUN lun_id.
	 */
	path = device->last_path;
	first_path = path;
	while((path = mpl_get_next_avail_path(path, first_path))) {
		path_tm =  path->avg_time;
		if (path_tm < tm) {
			low_path = path;
			tm = path_tm;
		}
	}
	if(low_path) {
		device->last_path = low_path;
	}
	return device->last_path;
}

struct mpl_path *
mpl_load_balance_lun(struct mpl_device *device, struct scsi_cmnd *cmd)
{
	struct mpl_path   *path;
	struct mpl_path   *first_path;
	struct mpl_path   *low_path;
	uint32_t           count, path_count;

	low_path = NULL;
	count = 0xffffffff;

	/* Load balance based on the LUN path outstanding IO */

	/* The LUN paths are from a specific HBA Port WWPN to a specific
	 * Controller Port WWPN LUN lun_id.
	 */
	path = device->last_path;
	first_path = path;
	while((path = mpl_get_next_avail_path(path, first_path))) {
		path_count =  atomic_read(&path->io_count);
		if (path_count < count) {
			low_path = path;
			count = path_count;
		}
	}
	if(low_path) {
		device->last_path = low_path;
	}
	return device->last_path;
}

struct mpl_path *
mpl_load_balance_target(struct mpl_device *device, struct scsi_cmnd *cmd)
{
	struct mpl_path   *path;
	struct mpl_path   *first_path;
	struct mpl_path   *low_path;
	uint32_t           count, target_count;

	low_path = NULL;
	count = 0xffffffff;

	/* Load balance based on the target path outstanding IO */

	/* The target paths are all paths from a specific HBA Port WWPN
	 * to a specific Controller Port WWPN.
	 */
	path = device->last_path;
	first_path = path;
	while((path = mpl_get_next_avail_path(path, first_path))) {
		target_count =  atomic_read(&path->tgt->io_count);
		if (target_count < count) {
			low_path = path;
			count = target_count;
		}
	}
	if(low_path) {
		device->last_path = low_path;
	}
	return device->last_path;
}

struct mpl_path *
mpl_load_balance_hba(struct mpl_device *device, struct scsi_cmnd *cmd)
{
	struct mpl_path   *path;
	struct mpl_path   *first_path;
	struct mpl_path   *low_path;
	uint32_t           count, hba_count;

	low_path = NULL;
	count = 0xffffffff;

	/* Load balance based on the hba path outstanding IO */

	/* The hba paths are all paths from a specific HBA Port WWPN.
	 */
	path = device->last_path;
	first_path = path;
	while((path = mpl_get_next_avail_path(path, first_path))) {
		hba_count =  atomic_read(&path->host->io_count);
		if (hba_count < count) {
			low_path = path;
			count = hba_count;
		}
	}
	if(low_path) {
		device->last_path = low_path;
	}
	return device->last_path;
}

int
mpl_get_device_outio(struct mpl_device *device)
{
	struct mpl_path   *path;
	struct mpl_path   *first_path;
	uint32_t           count;

	count = 0;

	/* The hba paths are all paths from a specific HBA Port WWPN.
	 */
	path = device->last_path;
	first_path = path;
	do {
		count +=  atomic_read(&path->tgt->io_count);
		path = mpl_get_next_path(path);
	} while (path != first_path);

	return count;
}

/* any non zero return code indicates error.  exact value not percolated,
 * but we differentiate anyway */
int
mpl_queuecommand(struct scsi_cmnd *cmd, void (*done)(struct scsi_cmnd *))
{
	struct mpl_host   *host;
	struct mpl_device *device;
	struct mpl_path   *path;
	struct io_stub    *stub;
	struct mpl_path *xmit_path;
	int  rv = DID_NO_CONNECT;

	mpl_module_lock();
	host  = mpl_find_host_by_scsi_host(mpl_scsi_host_from_cmd(cmd));
	if (!host) {
		cmd->result = ScsiResult(rv, 0);
		goto out_error;
	}

	path = mpl_find_or_create_path(host, mpl_target_no_from_cmd(cmd),
				       mpl_lun_no_from_cmd(cmd));
	if (!path) {
		mpl_print(0, MPL_ERR "queuecommand failed to find path %d/%d/%d\n",
			  mpl_host_no_from_cmd(cmd),
			  mpl_target_no_from_cmd(cmd),
			  mpl_lun_no_from_cmd(cmd));
		cmd->result = ScsiResult(rv, 0);
		goto out_error;
	}

	device = path->device;
	if (device->plugin &&
	    device->plugin->pre_queue_filter &&
	    device->plugin->pre_queue_filter(path, cmd, done)) {
		mpl_module_unlock();
		return 0;
	}

	if ((!mpl_is_presentation_path(path)) &&
	    (cmd->cmnd[0] != REPORT_LUNS)) { /* Just let Report LUN go thru */

		mpl_print(MPL_IOPATH, MPL_WARNING 
			"queuecommand rejected non presentation I/O %d/%d/%d\n",
			  mpl_host_no_from_cmd(cmd),
			  mpl_target_no_from_cmd(cmd),
			  mpl_lun_no_from_cmd(cmd));
		if (cmd->cmnd[0] == INQUIRY) {
			/* We don't want to return DID_NO_CONNECT because we want
			 * the SCSI Mid-layer to keep scanning this device for other LUNs.
			 */
			mpl_setup_inq_lun_skip(cmd);

		}
		else {
			cmd->result = ScsiResult(rv, 0);
		}
		goto out_error;
	}

#if 0
	if (mpl_is_presentation_path(path) && path->private_sdev) {
		path->private_sdev = 0;
		kfree(path->sdev);
		path->sdev = cmd->device;
	}
#endif

	if ( mpl_enable_tgt_qdepth) {
		if (device->tgt_queue_depth && mpl_is_path_configured(path) &&
		    (mpl_get_device_outio(device) >= device->tgt_queue_depth)) {
			mpl_module_unlock();
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
			return SCSI_MLQUEUE_DEVICE_BUSY;
#else
			cmd->result = ScsiResult(DID_BUS_BUSY, 0);
			cmd->retries = 0; /* Always retry this case */
			goto out_error;
#endif
		}
	}

	stub = mpl_stub_alloc(cmd, path, done);
	if (!stub) {
		cmd->result = ScsiResult(DID_BUS_BUSY, 0);
		goto out_error;
	}

	stub->scsi_host = mpl_scsi_host_from_cmd(cmd);
	stub->target = mpl_target_no_from_cmd(cmd);
	stub->lun = mpl_lun_no_from_cmd(cmd);
	if (mpl_is_path_configured(path))
		xmit_path = device->last_path;
	else
		xmit_path = path;

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
        stub->timeout_handler = cmd->eh_timeout.function;
	cmd->eh_timeout.function = (void (*)(unsigned long)) mpl_eh_timeout_handler;
#endif
	atomic_inc(&device->host->tot_count);
	atomic_inc(&device->tot_count);
	atomic_inc(&path->tgt->tot_count);
	atomic_inc(&path->tot_count);

	if (path->seen_ext_inq) {

		if (device->state == DEV_FAILING ||
			(xmit_path->host->blocked || xmit_path->lld_blocked)) {
			mpl_add_pend_io(stub, device, 0);
			mpl_module_unlock();
			return 0;
		}

		if (mpl_is_path_configured(path) &&
		    ((device->state == DEV_FAILED) ||
		     (device->state == DEV_DISABLED))) {

			if ((device->state == DEV_FAILED) &&
				timer_pending(&device->nopath)) {
				mpl_add_pend_io(stub, device, 0);
				mpl_module_unlock();
				return 0;
			}
			mpl_stub_free(stub);
			cmd->result = ScsiResult(rv, 0);
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
			cmd->eh_timeout.function = stub->timeout_handler;
#endif			
			goto out_error;
		}

		switch(device->path_cntl_type) {
		case 0: /* default */
		case MPL_PC_TYPE_FAILOVER:
			break;
		case MPL_PC_TYPE_ROUND_ROBIN:
			xmit_path = mpl_round_robin_path(device, cmd);
			break;
		case MPL_PC_TYPE_DYNAMIC_LUNQ:
			xmit_path = mpl_load_balance_lun(device, cmd);
			break;
		case MPL_PC_TYPE_DYNAMIC_TGTQ:
			xmit_path = mpl_load_balance_target(device, cmd);
			break;
		case MPL_PC_TYPE_DYNAMIC_HBAQ:
			xmit_path = mpl_load_balance_hba(device, cmd);
			break;
		case MPL_PC_TYPE_DYNAMIC_TIME:
			xmit_path = mpl_load_balance_time(device, cmd);
			break;
		default:
			mpl_stub_free(stub);
			cmd->result = ScsiResult(rv, 0);
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
			cmd->eh_timeout.function = stub->timeout_handler;
#endif			
			goto out_error;
		}
		if (!mpl_is_presentation_path(xmit_path)) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
			cmd->host = xmit_path->host->scsi_host;
			cmd->target = xmit_path->target;
			cmd->lun = xmit_path->lun;
#endif
		}

	}

	/* updating device's io_count since now we know we own it */
	atomic_inc(&xmit_path->host->io_count);
	atomic_inc(&xmit_path->device->io_count);
	atomic_inc(&xmit_path->tgt->io_count);
	atomic_inc(&xmit_path->io_count);

	rv = mpl_do_base_queuecommand(device, xmit_path, stub);

                                /* If base driver cannot process this cmd,
				 * then perform cleanup for this cmd before
				 * returning.  */
	if (rv) {
				/* If base driver cannot process this cmd,
				 * then perform cleanup for this cmd before
				 * returning.  */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
		stub->cmd->device = stub->sdev;
		cmd->eh_timeout.function = stub->timeout_handler;
#endif
		atomic_dec(&xmit_path->host->io_count);
		atomic_dec(&xmit_path->device->io_count);
		atomic_dec(&xmit_path->tgt->io_count);
		atomic_dec(&xmit_path->io_count);
		mpl_stub_free(stub);
	}

	mpl_module_unlock();
	return rv;

out_error:
	mpl_module_unlock();
	done(cmd);
	return 0;
}

static inline int
mpl_logical_unit_not_supported(unsigned char *sense)
{
	return (sense[2]  == 0x05 &&		/* sense key */
		sense[12] == 0x25 &&		/* asc */
		sense[13] == 0x00);		/* asq */
}

static inline int
mpl_illegal_request(unsigned char *sense)
{
	return (sense[2] == 0x05);
}

static void
mpl_no_plugin(struct scsi_cmnd *cmd, struct mpl_device *device,
	      struct mpl_path *path)
{				/* simply used for diagnostic output. */
	uint8_t vendor[8 + 1];
	uint8_t product[16 + 1];
	uint8_t rev[4 + 1];

	trim_to(vendor, cmd->request_buffer + 8, sizeof(vendor));
	trim_to(product, cmd->request_buffer + 16, sizeof(product));
	trim_to(rev, cmd->request_buffer + 32, sizeof(rev));
	mpl_print(0, MPL_ERR "Cannot find plugin for %s %s, %s, %s, %s\n",
		  mpl_print_device(device),
		  mpl_print_path(path),
		  vendor, product, rev);
}

static void
mpl_evaluate_config_path_control(struct mpl_path *path)
{
	struct mpl_device *device;
	int device_flags;
	int path_flags;
	int path_cntl;


	if (!path->tgt || !path->host) {
		mpl_print(0, MPL_ERR "Missing host or target: %p, %p\n",
			  path->tgt, path->host);
		return;
	}
	
	if (mpl_get_device_flags_by_path(&device_flags, path->host->wwpn,
					 path->tgt->wwpn, path->lun))
		return;
	if (mpl_get_path_flags(&path_flags, path->host->wwpn,
			       path->tgt->wwpn, path->lun))
		return;

	device = path->device;
	if (path_flags & MPL_CFG_PATH_EXPORT) {
		mpl_set_path_configured(path,CONFIG_UNAVAILABLE);
		/* We don't care about the device for an unavailable path */
		return;
	} else if (mpl_auto_configure) {
		return;
	}

	mpl_set_path_configured(path,CONFIG_CONFIGURED);

	path_cntl = device_flags & 0x0f;
	switch (path_cntl) {
	case MPL_PC_TYPE_ROUND_ROBIN:
		path->traffic_ratio = path_flags;
				/* Intentional fall through */
	case MPL_PC_TYPE_DYNAMIC_LUNQ:
	case MPL_PC_TYPE_DYNAMIC_TGTQ:
	case MPL_PC_TYPE_DYNAMIC_HBAQ:
	case MPL_PC_TYPE_DYNAMIC_TIME:
		device->path_cntl_type = path_cntl;
		break;
	case MPL_PC_TYPE_FAILOVER:
		device->path_cntl_type = path_cntl;
		device->auto_failback = (device_flags & 
					 MPL_PC_CFG_AUTO_FAILBACK) != 0;
		if (path_flags & 1)
			path->is_default = 1;
		break;
	default:
		mpl_print(0, MPL_ERR "Bad path control type (%#x) for %s\n",
			  path_cntl, mpl_print_device(device));
		break;
	}
	mpl_set_device_static_config(device);
	mpl_set_path_static_config(path);
}

static void
mpl_process_inquiry(struct io_stub *stub)
{
	struct mpl_path   *path = stub->path;
	struct mpl_path   *nextpath;
	struct mpl_device *device = path->device;
	struct scsi_cmnd *cmd = stub->cmd;
	uint8_t *request = cmd->request_buffer;
	uint8_t *sns = cmd->sense_buffer;
	int  failure;

	/* Path this cmd just finished on might not be the original
	 * path.
	 */

	if (mpl_find_stub_in_hash(stub)) {
		mpl_print(0, MPL_ERR "Inquiry stub still in hash: %p, %s %s\n",
		       stub, mpl_print_device(path->device),
			mpl_print_path(path));
	}

	mpl_print(MPL_RECOVERY, MPL_INFO "Inquiry Page 0 "
			  "done %s %s - result %x, sns: "
			  "%x/%x/%x\n",
			  mpl_print_device(device),
			  mpl_print_path(path),
			  cmd->result, sns[2],
			  sns[12], sns[13]);

	failure = (cmd->result || (request[0] == 0x7f) ||
		   mpl_illegal_request(sns)  ||
		   ((request[0] & 0xE0) == 0x20));
	if (failure) {

		if (!mpl_is_device_configured(device) ||
		    (mpl_auto_configure && !initial_scsi_scan_done)) {
			mpl_remove_device_from_device_hash(device);
			list_del(&device->devices);
			list_add_tail(&device->devices, &reaper_queue);
		}
	} else {
		device->scsi_type = request[0] & 0x1f;

		/* TPGS support */
		path->tpgs = (request[5] & MPL_ALUA_TPGS_MASK);

		if (path->tpgs) {
			mpl_print(MPL_ALUA, MPL_INFO "Inquiry P0 ALUA support x%x: %s %s\n",
				path->tpgs,
		  		mpl_print_device(device),
		  		mpl_print_path(path));

			/* path->tpgs will be used later in the HSM to decide whether or not
			 * send RTPG before InquirySN.
			 */
		}

		if (device->plugin == NULL) {
			device->plugin = mpl_find_plugin(cmd, device);
			if (!device->plugin) {
				mpl_no_plugin(cmd, device, path);
				goto bypass_snooping;
			}
		}

		if (stub->done) {
			if (mpl_is_device_configured(device)) {
				list_for_each_entry(nextpath, &device->paths, path_list) {
					if(nextpath != path) {
						nextpath->flag |= PATH_NEEDS_INQUIRYP0;
						nextpath->retries = 1;
						mpl_restart_sched_cmd(nextpath, jiffies + 1);
					}
				}
			}
		}
		if(!path->got_wwpn_data && path->host->hba) {
			/* Get the WWPN for this path */
			mpl_get_tgt_wwinfo(path->host->hba, path->target, path->tgt->wwpn, path->tgt->wwnn);
			path->got_wwpn_data = 1;
			mpl_evaluate_config_path_control(path);
		}

		/* Check for tape device */
		if (device->scsi_type == 1) {
			/* Tape devices are just pass-thru */
			mpl_set_path_configured(path, CONFIG_UNAVAILABLE);
			goto bypass_snooping;
		}

		mpl_path_state_machine(path, PATH_UP);

		if (!failure && stub->done && !device->serial_size) {
			/* Make sure the HBeatSM has started */
			if (path->hbstate == HB_TUR) {

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
				/* 
				   During the initial lun scan, timeout value is
				   only 5 seconds. If we are delaying the completion
				   of this command increase the timeout.
				*/
				mod_timer(&stub->cmd->eh_timeout, jiffies + 30*HZ );
#endif
				path->delayed_cmd = stub;
				return;
			}
		}
	}
	
bypass_snooping:
	if (stub->done) {
		mpl_call_done_free_stub(stub, 1);
	}
}

/* Move paths from one device to another. 
 */

static void
mpl_adjust_ref_count(struct mpl_device *device, 
				 struct mpl_path *path)
{
	struct io_stub *stub, *prev;
	unsigned long flags;
	int i;

	spin_lock_irqsave(&stub_hash_table_lock, flags);
	for (i = 0; i < stub_hash_size; i++) {
		stub = stub_hash[i];
		while(stub) {
			prev = stub;
			stub = stub->next;
			if (prev && prev->path == path) {
				atomic_inc(&path->device->ref_count);
				atomic_dec(&device->ref_count);
			}
		}
	}
	spin_unlock_irqrestore(&stub_hash_table_lock, flags);
}

static void
mpl_redirect_pending_ios(struct mpl_device *dest, struct mpl_device *src)
{
	struct io_stub    *stub;

	while ((stub = mpl_get_first_pend_io(src))) {
		mpl_add_pend_io(stub, dest, 1);
	}
	if (dest->state != DEV_FAILING) {
		mpl_device_process_pendq(dest);
	}
}

static void
mpl_redirect_paths(struct mpl_device *dest, struct mpl_device *src)
{
	struct mpl_path *path;

	while (!list_empty(&src->paths)) {
		path = mpl_first_path(src);
		if (path->sas == MPL_ALUA_ACTIVE_OPT) {
			atomic_dec(&path->device->opt_path_cnt);
			atomic_inc(&dest->opt_path_cnt);
		}
		path->device = dest;
		if (mpl_auto_configure)
			path->is_default = 0;
		if(path->state == PATH_ONLINE)
			path->state = PATH_STANDBY;
		list_del(&path->path_list);
		list_add_tail(&path->path_list, &dest->paths);

		/* Fix up device reference counts related to
		 * any outstanding io_stubs
		 */
		mpl_adjust_ref_count(src, path);
	}
}

				/* Consolidate paths on two devices, return
				 * the survivor. */
static struct mpl_device *
mpl_consolidate_devices(struct mpl_device *device, struct mpl_device *alt)
{
				/* Calls routines that takes either the device
				 * hash lock or the stub hash lock.  So
				 * neither lock may be held on entry. */

				/* Theoretically only one device could be
				 * marked configured.  If it is the alternate
				 * device, switch roles so we end up with the
				 * device set up at initialization time. */
	if (!mpl_is_device_configured(device) &&
	    mpl_is_device_configured(alt)) {
		struct mpl_device *tmp;

		tmp = device;
		device = alt;
		alt = tmp;
	}
	mpl_redirect_paths(device, alt);
	mpl_redirect_pending_ios(device, alt);
	mpl_remove_device_from_device_hash(alt);
	list_del(&alt->devices);
	list_add_tail(&alt->devices, &reaper_queue);
	return device;
}

/* Search for existing device with newly probed serial number.  If
 * found, move current device path(s) to existing device.  If not
 * found, update this device serial number.  Return old device, if it
 * existed, current device if not. */

static struct mpl_device *
mpl_match_device_by_serial(struct mpl_device *device, uint8_t *serial, int size)
{
	struct mpl_host *host;
	struct mpl_device *d;
	int  i;

	if (!device->serial_size) {
		return device;
	}
	for (i = 0; ; i++) {
		host = mpl_find_host(i);
		if (host == NULL) {
			break;
		}
		list_for_each_entry(d, &host->devices, devices) {
			if (d != device &&
			    !mpl_compare_serial_number(d, serial, size)) {
				d = mpl_consolidate_devices(d, device);
				return d;
			}
		}
	}
	return device;
}

void
mpl_process_internal_cmd(struct io_stub *stub)
{
	struct scsi_cmnd *cmd = stub->cmd;
	struct mpl_path *path = stub->path;
	struct mpl_device *device = path->device;

	mpl_print(MPL_INTERNAL_CMD, MPL_INFO "Cmpl cmd %x %x for %s %s "
		  "result %x sk %x asc %x ascq %x\n",
		  cmd->cmnd[0], cmd->cmnd[2],
		  mpl_print_device(device), mpl_print_path(path),
		  cmd->result, cmd->sense_buffer[2],
		  cmd->sense_buffer[12], cmd->sense_buffer[13]); 

	if (mpl_is_power_on_reset(cmd->result, cmd->sense_buffer) ||
	    mpl_is_queue_full(cmd->result) ||
	    mpl_is_queue_busy(cmd->result)) {
		/* Just retry 29/00 and QFULL errors */
		atomic_dec(&int_cmd_cnt);
		memset(&cmd->sense_buffer[0], 0, SCSI_SENSE_BUFFERSIZE);
		path->retries++;
		mpl_set_cmd_timer(stub, MPL_INTERNAL_CMD_TMO);
		mpl_add_deferral(stub, MPL_INTERNAL_CMD);
		return;
	}

	if (mpl_is_inquiry_page_0(cmd)) {
		/* Inquiry P0 is not generated by a plugin, the mpl module
		 * generates this when processing a Inquiry P0 on a configured
		 * presentation path. Thus all types of devices should take
		 * the action coded in mpl_process_inquiry().
		 */
		mpl_process_inquiry(stub);
	} else if (device->plugin) {

		/* Always call plugin for plugin generated internal commands */
	 	device->plugin->diagnose_results(stub, PLUG_INTERNAL_CMD);
	}

	mpl_free_internal_stub(stub);
	atomic_dec(&int_cmd_cnt);
	return;
}


static inline int
mpl_is_retryable(struct scsi_cmnd  *cmd)
{
				/* This is a quick hack to fix CR 15564.  We
				 * need a more robust mechanism that allows
				 * the diagnose plugin to determine whether
				 * the I/O is retryable or not. */
	return (cmd->result == 2 && 
		(cmd->sense_buffer[2] & 0xf) == 0x02 &&
		cmd->sense_buffer[12] == 0x04);
}

static void
mpl_expedite_hbeat(struct mpl_path *path)
{
	struct mpl_path   *hpath;
	unsigned long flags;
	int i;

	/* If path is currently functional and is in process of going down */
	if ((path->state != PATH_ONLINE) &&
	    (path->state != PATH_STANDBY)) 
		return;

	/* Check all other paths on this target by expediting their hbeat */
	spin_lock_irqsave(&path_hash_table_lock, flags);
	for (i=0;i<path_hash_size;i++) {
redohash:
		for (hpath = path_hash[i]; hpath; hpath = hpath->next) {
			if (hpath->host == path->host &&
		    	    hpath->target == path->target)
			{
				if ((hpath->hbstate == HB_WAITING) &&
				   !(hpath->flag & PATH_FAST_POLL)) {
					hpath->flag |= PATH_FAST_POLL;
					spin_unlock_irqrestore(
						&path_hash_table_lock, flags);
					mpl_module_unlock();
					mod_timer(&hpath->hbeat, jiffies + HZ);
					mpl_module_lock();
					spin_lock_irqsave(
						&path_hash_table_lock, flags);
					goto redohash;
				}
			}
		}
	}
	spin_unlock_irqrestore(&path_hash_table_lock, flags);
	return;
}

void
mpl_process_hbeat_cmd(struct io_stub *stub)
{
	struct scsi_cmnd *cmd = stub->cmd;
	struct mpl_path *path = stub->path;
	struct mpl_device *device = path->device;
	enum mpl_hb_event event = HB_SUCCEED;
	int  diagnose_reject = 0;
	uint8_t *datap;

	if (cmd->result == ScsiResult(DID_NO_CONNECT, 0)) {
		mpl_expedite_hbeat(path);
		event = HB_FAIL;
	} else if (mpl_is_queue_full(cmd->result) ||
		   mpl_is_queue_busy(cmd->result)) {
		event = HB_BUSY;
	} else if (device->plugin) {
		diagnose_reject = device->plugin->diagnose_results(stub, PLUG_HBEAT_CMD);
		if (diagnose_reject)
			event = HB_FAIL;
		else if (mpl_is_inquiry_page_83(cmd) && mpl_hbeat_inquirysn(stub)) {
			event = HB_FAIL;
		}
	}
	if (event == HB_FAIL) {
		if (path->delayed_cmd) {
			mpl_call_done_free_stub(path->delayed_cmd, 1);
			path->delayed_cmd = NULL;
		}
	}

	if ((event == HB_SUCCEED) &&
		(path->hbstate == HB_RTPG)) {
		if (path->tgt_port_info) {
			mpl_put_scsi_buf(path->tgt_port_info);
			path->tgt_port_info = 0;
		}
		datap = mpl_get_scsi_buf(); 
		if (datap) {
			path->tgt_port_info = cmd->request_buffer;
			cmd->request_buffer = datap;
		}
	}

	mpl_free_internal_stub(stub);
	mpl_hbeat_state_machine(path, event);
	return;
}

static void
mpl_process_failed_cmd(struct io_stub *stub)
{
	struct scsi_cmnd  *cmd = stub->cmd;
	struct mpl_path   *path = stub->path;
	struct mpl_device *device = path->device;
	int retry_cmd = 0;
	int diag_res;

	if (device->state == DEV_FAILED) {
		if (timer_pending(&device->nopath)) {
			mpl_add_pend_io(stub, device, 1);
			return;
		}
		/* All paths to device are already down,
		 * just return the failed command to the SCSI layer.
		 */
		mpl_call_done_free_stub(stub, 1);
		return;
	}

	if (cmd->result == ScsiResult(DID_NO_CONNECT, 0)) {
		mpl_expedite_hbeat(path);
		goto fover;
	}
	/* All TURs should go thru plugin diagnostic check */
	if (device->plugin && 
	    (diag_res = device->plugin->diagnose_results(stub,PLUG_SCSI_CMD))) {
		if (diag_res == 2) { /* retry with no failover */
			retry_cmd = 1;
		} else {
fover:
			/* Attempt retry with failover */
			mpl_path_state_machine(path, PATH_DOWN);
			retry_cmd = 1;
		}
	}

	/* If the path I got this error on is FAILED/DISABLED, and there
	 * are other online / standy paths available, retry.
	 */
	if ((path->state == PATH_FAILED) ||
		 (path->state == PATH_DISABLED)) {
		if (device->state != DEV_FAILED)
			retry_cmd = 1;
	}

	if (retry_cmd) {

		mpl_print(MPL_IOPATH, MPL_INFO "RETRY cmd %x %x for %s %s "
		  "result %x sk %x asc %x ascq %x\n",
		  cmd->cmnd[0], cmd->cmnd[2],
		  mpl_print_device(device), mpl_print_path(path),
		  cmd->result, cmd->sense_buffer[2],
		  cmd->sense_buffer[12], cmd->sense_buffer[13]); 

		mpl_add_pend_io(stub, device, 1);
	} else {
		mpl_call_done_free_stub(stub, 1);
	}
	return;
}

/*
   This function is called from mpl_eh_thread context due to internal 
   command timeout and is called with module_lock held.
*/
int
mpl_do_int_tmo(struct io_stub *stub, void (*done)(struct scsi_cmnd *))
{
	struct mpl_host *host;
	struct mpl_path *path;
	struct scsi_cmnd *cmd = stub->cmd;
	int ret;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	host  = mpl_find_host_by_scsi_host(cmd->host);
#else
	host  = mpl_find_host_by_scsi_host(cmd->device->host);
#endif

	/* path this cmd just finished on. Might not be
	 * the original path.
	 */
	path = stub->path;

	WARN_ON(mpl_find_stub_remove_quiet(cmd));

	if (driver_unloading)
		goto out;

	mpl_module_unlock();
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	spin_lock_irq(&io_request_lock);
#else
	spin_lock_irq(cmd->device->host->host_lock);
#endif
	ret = mpl_base_abort_handler(cmd);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	spin_unlock_irq(&io_request_lock);
                /* Since 7.X driver doesn't wait for cmd cmpl in abort handler,
                 * delay a bit before returning back to SCSI layer.
                 */
	set_current_state(TASK_UNINTERRUPTIBLE);
	schedule_timeout(HZ);
#else
	spin_unlock_irq(cmd->device->host->host_lock);
#endif
	mpl_module_lock();

out:
	atomic_dec(&path->internal_count);
	mpl_inc_can_queue(stub);
	
	/* Call process cmpl with error */
	cmd->result = ScsiResult(DID_ERROR, 0);
	if(stub->reason == MPL_DO_HBEAT_TMO) {
		atomic_dec(&int_cmd_cnt);
		mpl_process_hbeat_cmd(stub);
	}
	else {
		mpl_process_internal_cmd(stub);
	}

	mpl_print(0, MPL_ERR "Internal cmd 0x%x tmo for %s %s reason %d pstate %d dstate %d\n",
		cmd->cmnd[0],
		mpl_print_device(path->device),
		mpl_print_path(path),
		stub->reason, path->state, path->device->state);

	return 0;
}


static void
mpl_process_aborted_cmd(struct io_stub *stub)
{
	mpl_call_done_free_stub(stub, 1);
	return;
}

int
mpl_set_path_online(struct mpl_path *path)
{
	mpl_path_state_machine(path, PATH_ENABLE);
	return 0;
}

int
mpl_set_path_offline(struct mpl_path *path)
{
	mpl_path_state_machine(path, PATH_DISABLE);
	return 0;
}

int
mpl_clear_path_statistics(struct mpl_path *path)
{
				/* TODO: UI needs some path statistics */
	return 0;
}

static void
mpl_reap_queue(void)
{
	struct mpl_device *device, *tmp;
	struct mpl_path   *path;
	struct io_stub    *stub;

	list_for_each_entry_safe(device, tmp, &reaper_queue, devices) {
				/* Kill any outstanding mid layer I/Os */
		while ((stub = mpl_get_first_pend_io(device))) {
			stub->cmd->result = ScsiResult(DID_NO_CONNECT, 0);

			WARN_ON(mpl_find_stub_remove(stub->cmd) != stub);
			mpl_call_done_free_stub(stub, 1);

		}

		/* Make sure HBeat is stopped */
		list_for_each_entry(path, &device->paths, path_list) {
			mpl_hbeat_state_machine(path, HB_STOP);
		}
				/* note: mpl_free_device does the list_del to
				 * remove this device from the reaper queue */
		mpl_free_device(device);
	}
}

static int
mpl_thread(void *arg)
{
	struct io_stub *stub;
	unsigned long dflags;
	siginitsetinv(&current->blocked, 0);
	lock_kernel();
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	daemonize();
#else
	daemonize("mpl_thread");
#endif
	snprintf(current->comm, sizeof(current->comm), "mpl_thread");
	unlock_kernel();
	mpl_module_lock();
	while (mpl_thread_quit == 0) {

		mpl_module_unlock();
		down_interruptible(&mpl_thread_sleep);
		mpl_module_lock();

		stub = mpl_get_first_deferral();
		while (stub) {
			struct mpl_path *path = stub->path;
			struct mpl_device *device = path->device;
			switch (stub->reason) {
			case MPL_INQ_CMD:
				mpl_process_inquiry(stub);
				break;
			case MPL_INT_CMD:
				mpl_process_internal_cmd(stub);
				atomic_dec(&path->internal_count);
				mpl_inc_can_queue(stub);
				break;
			case MPL_HB_CMD:
				mpl_process_hbeat_cmd(stub);
				break;
			case MPL_FAILED_CMD:
				mpl_process_failed_cmd(stub);
				break;
			case MPL_RETRY_CMD:
				mpl_retry_pending_io(stub);
				break;
			case MPL_HEARTBEAT_CMD:
				mpl_send_internal_completion(stub->path, stub,
							     mpl_hb_done);
				break;
			case MPL_INTERNAL_CMD:
				mpl_send_internal_completion(stub->path, stub,
						mpl_internal_done);
				break;
			case MPL_ABORTED_CMD:
				mpl_process_aborted_cmd(stub);
				break;
			default:
				mpl_print(0, MPL_ERR "mpl_thread: bad reason "
					  "code in io_stub: %p\n",
					  stub);
				mpl_stub_free(stub);
				break;
			}

			/* maintain deferral count until after the I/O
			 * has been completely processed
			 */
			spin_lock_irqsave(&deferral_queue_lock, dflags);
			atomic_dec(&device->deferral_count);
			spin_unlock_irqrestore(&deferral_queue_lock, dflags);

			stub = mpl_get_first_deferral();
		}
		mpl_reap_queue();

	}
	mpl_module_unlock();
	up(&mpl_thread_gone);
	return 0;
}

static int
mpl_eh_thread(void *arg)
{
	struct io_stub *stub;
	unsigned long dflags;

	siginitsetinv(&current->blocked, 0);
	lock_kernel();
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	daemonize();
#else
	daemonize("mpl_eh_thread");
#endif
	snprintf(current->comm, sizeof(current->comm), "mpl_eh_thread");
	unlock_kernel();
	mpl_module_lock();
	while (mpl_eh_thread_quit == 0) {

		mpl_module_unlock();
		down_interruptible(&mpl_eh_thread_sleep);
		mpl_module_lock();

		stub = mpl_get_first_eh_deferral();
		while (stub) {
			struct mpl_path *path = stub->path;
			struct mpl_device *device = path->device;

			switch (stub->reason) {
			case MPL_DO_TMO:
				mpl_do_tmo(stub);
				break;
			case MPL_DO_INT_TMO:
				mpl_do_int_tmo(stub, mpl_internal_done);
				break;
			case MPL_DO_HBEAT_TMO:
				mpl_do_int_tmo(stub, mpl_hb_done);
				break;
			case MPL_DO_ABORT:
				mpl_do_abort(stub->cmd);
				break;
			case MPL_DO_DEV_RESET:
				mpl_do_dev_reset(stub);
				mpl_stub_free(stub);
				break;
			default:
				mpl_print(0, MPL_ERR "mpl_eh_thread: bad reason "
					  "code in io_stub: %p\n",
					  stub);
				mpl_stub_free(stub);
				break;
			}

			/* maintain eh_deferral count until after the I/O
			 * has been completely processed
			 */
			spin_lock_irqsave(&eh_deferral_queue_lock, dflags);
			atomic_dec(&device->eh_deferral_count);
			spin_unlock_irqrestore(&eh_deferral_queue_lock, dflags);

			stub = mpl_get_first_eh_deferral();
		}
		mpl_reap_queue();

	}
	mpl_module_unlock();
	up(&mpl_eh_thread_gone);
	return 0;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
/* In the 2.6 driver, lun queue depth is controled entirely by scsi mid-layer */
void
mpl_select_queue_depths(struct Scsi_Host *host, struct scsi_device *scsi_devs)
{
	struct scsi_device *device;

	mpl_base_select_queue_depths(host, scsi_devs);

	/* After the base driver queue depth routine sets the scsi layer
	 * queue depth for each device, subtract 1 to allow room for
	 * an adminstrative mpl command to be sent (like inq hbeat)
	 */
	for (device = scsi_devs; device != 0; device = device->next) {
		if (device->host == host && device->queue_depth > 1) {
				device->queue_depth--;
		}
	}
}
#endif

int
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
mpl_detect(Scsi_Host_Template *tpnt)
#else
mpl_detect(struct scsi_host_template *tpnt)
#endif
{
	int ret;
	struct mpl_host *host;
	struct Scsi_Host *scsi_host;

	ret = mpl_base_detect(tpnt);

	mpl_module_lock();
	/* After detect returns, fix up select_queue_depths
	 * function in scsi host to call mpl_select_queue_depths
	 */
	list_for_each_entry(host, &mpl_hosts, host_list) {
		scsi_host = host->scsi_host;
		host->hba = (struct lpfc_hba *)scsi_host->hostdata[0];
		mpl_get_hba_wwpn(host->hba, &host->wwpn[0]);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
		/* In the 2.6 driver, lun queue depth is controlled entirely
		 * by scsi mid-layer */
		if(scsi_host) {
			if(scsi_host->select_queue_depths) {
				mpl_base_select_queue_depths =
					scsi_host->select_queue_depths;
				scsi_host->select_queue_depths =
					mpl_select_queue_depths;
			}
		}
#endif
	}
	mpl_module_unlock();
	return ret;
}

int
mpl_release(struct Scsi_Host *shost)
{
	struct mpl_host *host;
	struct mpl_device *device;
	struct mpl_path *path;
	int ret;

	mpl_module_lock();
	driver_unloading = 1;
	/* Lets cancel all our timers before we unload the base driver */
	list_for_each_entry(host, &mpl_hosts, host_list) {
		list_for_each_entry(device, &host->devices, devices) {
			mpl_device_state_machine(device, NULL, DEV_DISABLE);
			list_for_each_entry(path, &device->paths, path_list) {
				mpl_path_state_machine(path, PATH_DISABLE);
				mpl_set_path_configured(path, CONFIG_UNCONFIGURED);
			}
			mpl_set_device_configured(device, CONFIG_UNCONFIGURED);
			mpl_quiesce_device(device, 30);
		}
	}
	mpl_module_unlock();
	ret = mpl_base_release(shost);
	return ret;
}
/*
   This function is called from mpl_eh_thread context due to  a scsi_cmnd
   timeout and is called with module_lock held.
*/
int
mpl_do_tmo(struct io_stub *stub)
{
	int ret;
	struct mpl_path *path;

	mpl_module_unlock();
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	spin_lock_irq(&io_request_lock);
#else
	spin_lock_irq(stub->cmd->device->host->host_lock);
#endif
	ret = mpl_base_abort_handler(stub->cmd);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	spin_unlock_irq(&io_request_lock);
                /* Since 7.X driver doesn't wait for cmd cmpl in abort handler,
                 * delay a bit before returning back to SCSI layer.
                 */
	set_current_state(TASK_UNINTERRUPTIBLE);
	schedule_timeout(HZ);
#else
	spin_unlock_irq(stub->cmd->device->host->host_lock);
#endif
	mpl_module_lock();
	WARN_ON(mpl_find_stub_remove_quiet(stub->cmd));

	path = stub->path;
	atomic_dec(&path->io_count);
	atomic_dec(&path->tgt->io_count);
	atomic_dec(&path->device->io_count);
	atomic_dec(&path->host->io_count);

	stub->cmd->result = ScsiResult(DID_TIME_OUT, 0);
	stub->cmd->eh_timeout.data = (unsigned long)stub->cmd;
	stub->cmd->eh_timeout.expires = jiffies + stub->cmd->timeout_per_command;
	add_timer(&stub->cmd->eh_timeout);

	mpl_process_aborted_cmd(stub);
	return 0;
}

/* This routine will handle a internal module initiated abort.
   It is called from mpl_eh_thread context to abort a specific IO
   and is called with module_lock held.
*/
int
mpl_do_abort(struct scsi_cmnd *cmd)
{
	int ret;

	mpl_module_unlock();
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	spin_lock_irq(&io_request_lock);
#else
	spin_lock_irq(cmd->device->host->host_lock);
#endif

	ret = mpl_abort_handler(cmd);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	spin_unlock_irq(&io_request_lock);
#else
	spin_unlock_irq(cmd->device->host->host_lock);
#endif

	mpl_module_lock();
	return ret;
}

/* This routine will handle a internal module initiated device reset.
   It is called from mpl_eh_thread context to reset a specific physical device
   and is called with module_lock held.
*/
int
mpl_do_dev_reset(struct io_stub *stub)
{
	struct scsi_cmnd *cmd;
	struct mpl_path *path=NULL, *xmit_path;
	int ret = 0, i;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	struct Scsi_Host *tmp_host;
	unsigned int tmp_target;
	unsigned int tmp_lun;
#else
	struct scsi_device *tmp_sdev;
#endif

	if (stub) {
		cmd = stub->cmd;
		mpl_print(MPL_IOPATH, MPL_INFO "DEVICE RESET %p, %d/%d/%d stub %p\n",
			  cmd, mpl_host_no_from_cmd(cmd),
			  mpl_target_no_from_cmd(cmd),
			  mpl_lun_no_from_cmd(cmd), stub);
		path = stub->path;
		if (mpl_is_path_configured(path))
			xmit_path = path->device->last_path;
		else
			xmit_path = path;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
		tmp_host = cmd->host;
		tmp_target = cmd->target;
		tmp_lun = cmd->lun;
		cmd->host = xmit_path->host->scsi_host;
		cmd->target = xmit_path->target;
		cmd->lun = xmit_path->lun;
#else
		tmp_sdev = cmd->device;
		cmd->device = xmit_path->sdev;
#endif

		mpl_module_unlock();
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
		spin_lock_irq(&io_request_lock);
		ret = mpl_base_device_reset_handler(cmd);
		spin_unlock_irq(&io_request_lock);
#else
		spin_lock_irq(cmd->device->host->host_lock);
		ret = mpl_base_device_reset_handler(cmd);
		spin_unlock_irq(cmd->device->host->host_lock);
#endif
		mpl_module_lock();

		/* Restore the command fields */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
		cmd->host = tmp_host;
		cmd->target = tmp_target;
		cmd->lun = tmp_lun;
#else
		cmd->device = tmp_sdev;
#endif
	}

	/* Wait for ALL IOs on this device to complete */
	if (path) {
		i = 0;
		while(atomic_read(&path->device->io_count)) {
			mpl_module_unlock();
			set_current_state(TASK_UNINTERRUPTIBLE);
			schedule_timeout(HZ);
			mpl_module_lock();
			if (++i >= 30)
				break;
		}
	}
	return ret;
}

int
mpl_abort_handler(struct scsi_cmnd *cmd)
{
	struct io_stub *stub;
	struct mpl_path *path;
	int ret, i;

	/* lock is held when called by SCSI layer */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	spin_unlock_irq(&io_request_lock);
#else
	spin_unlock_irq(cmd->device->host->host_lock);
#endif

	mpl_module_lock();
	stub = mpl_find_stub_remove_pend_io(cmd, MPL_ABORT_EXACT);

	mpl_print(MPL_IOPATH, MPL_INFO "ABORT %p, %d/%d/%d stub %p\n",
		  cmd, mpl_host_no_from_cmd(cmd),
		  mpl_target_no_from_cmd(cmd),
		  mpl_lun_no_from_cmd(cmd), stub);
	if (stub) {
		cmd->result = ScsiResult(DID_ERROR, 0);
		mpl_add_deferral(stub, MPL_ABORTED_CMD);

		/* Wait for This IO to complete */
		i = 0;
		while (cmd->host_scribble) {
			mpl_module_unlock();
			set_current_state(TASK_UNINTERRUPTIBLE);
			schedule_timeout(HZ);
			mpl_module_lock();
			if (++i >= 30)
				break;
		}
		mpl_module_unlock();
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
		spin_lock_irq(&io_request_lock);
#else
		spin_lock_irq(cmd->device->host->host_lock);
#endif
		ret = SUCCESS;
	} else {
		/* Wait for the eh_deferral thread to complete all the
		 * timedout commands. We're holding the module_lock, which
		 * blocks irqs.  So eh_deferral_queue_lock needs to either
		 * leave irq settings alone or do lock_irqsave,
		 * unlock_irq_restore.  We opt for the former. */
		i = 0;
		spin_lock(&eh_deferral_queue_lock);
		while (!list_empty(&eh_deferral_queue)) {
			spin_unlock(&eh_deferral_queue_lock);
			mpl_module_unlock();
			set_current_state(TASK_UNINTERRUPTIBLE);
			schedule_timeout(HZ);
			mpl_module_lock();
			spin_lock(&eh_deferral_queue_lock);
			if (++i >= 30)
				break;
		}
		spin_unlock(&eh_deferral_queue_lock);

		/* Command is owned by lpfc driver. We need to issue abort to
		 *  the lpfc driver and then restore the original sdev
		 *  structure. mpl_done can be called while abort is
		 *  pending. However, mpl_done is not guaranteed. We need to
		 *  protect against this. */
		stub = mpl_find_stub_hash(cmd);
		if (!stub) {
			/* command got completed while abort pending. Do
			 * nothing.  return SUCCESS */
			mpl_module_unlock();
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
			spin_lock_irq(&io_request_lock);
#else
			spin_lock_irq(cmd->device->host->host_lock);
#endif
			return SUCCESS;
		}
		/* Going to abort the command mpl_done should not handle this
		 * command's completion */
		stub->reason = MPL_DO_ABORT;

		mpl_module_unlock();
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
		spin_lock_irq(&io_request_lock);
#else
		spin_lock_irq(cmd->device->host->host_lock);
#endif
		ret = mpl_base_abort_handler(cmd);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
		spin_unlock_irq(&io_request_lock);
		/* Since 7.X driver doesn't wait for cmd cmpl in abort handler,
		 * delay a bit before returning back to SCSI layer.
		 */
		set_current_state(TASK_UNINTERRUPTIBLE);
		schedule_timeout(HZ);
		spin_lock_irq(&io_request_lock);
#endif
#if  LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
		spin_unlock_irq(cmd->device->host->host_lock);
#endif
		mpl_module_lock();
		WARN_ON(mpl_find_stub_remove_quiet(cmd));
		path = stub->path;
		atomic_dec(&path->io_count);
		atomic_dec(&path->tgt->io_count);
		atomic_dec(&path->device->io_count);
		atomic_dec(&path->host->io_count);
		if (!mpl_is_presentation_path(stub->path)) {
			mpl_cmd_restore(stub->cmd, stub);
		}
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
		stub->cmd->device = stub->sdev;
		stub->cmd->eh_timeout.function =
			stub->timeout_handler;
#endif

		mpl_stub_free(stub);
		mpl_module_unlock();
#if  LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
		spin_lock_irq(cmd->device->host->host_lock);
#endif
	}
	return ret;
}

int
mpl_device_reset_handler(struct scsi_cmnd *cmd)
{
	struct io_stub *stub;
	int ret, i;
	struct mpl_path *path=NULL, *xmit_path;
	struct mpl_host *host=NULL;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	struct Scsi_Host *tmp_host;
	unsigned int tmp_target;
	unsigned int tmp_lun;
#else
	struct scsi_device *tmp_sdev;
#endif

	/* lock is held when called by SCSI layer */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	spin_unlock_irq(&io_request_lock);
#else
	spin_unlock_irq(cmd->device->host->host_lock);
#endif
	mpl_module_lock();

	host = mpl_find_host_by_scsi_host(cmd->device->host);

	if (!host) {
		mpl_module_unlock();
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
		spin_lock_irq(&io_request_lock);
#else
		spin_lock_irq(cmd->device->host->host_lock);
#endif
		return FAILED;
	}

	path = mpl_find_path(host, cmd->device->id, cmd->device->lun);
	if (!path) {
		mpl_module_unlock();
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
		spin_lock_irq(&io_request_lock);
#else
		spin_lock_irq(cmd->device->host->host_lock);
#endif
		return FAILED;
	}

	if (mpl_is_path_configured(path))
		xmit_path = path->device->last_path;
	else
		xmit_path = path;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	tmp_host = cmd->host;
	tmp_target = cmd->target;
	tmp_lun = cmd->lun;
	cmd->host = xmit_path->host->scsi_host;
	cmd->target = xmit_path->target;
	cmd->lun = xmit_path->lun;
#else
	tmp_sdev = cmd->device;
	cmd->device = xmit_path->sdev;
#endif

	stub = mpl_find_stub_remove_pend_io(cmd, 
					    MPL_ABORT_DEVICE | MPL_ABORT_IGNORE_ABORTED);
	mpl_print(MPL_IOPATH, MPL_INFO "DEVICE RESET %p, %d/%d/%d stub %p\n",
		  cmd, mpl_host_no_from_cmd(cmd),
		  mpl_target_no_from_cmd(cmd),
		  mpl_lun_no_from_cmd(cmd), stub);
	while (stub) {
		stub->cmd->result = ScsiResult(DID_ERROR, 0);
		mpl_add_deferral(stub, MPL_ABORTED_CMD);
		stub = mpl_find_stub_remove_pend_io(cmd, MPL_ABORT_DEVICE | MPL_ABORT_IGNORE_ABORTED);
	}
	mpl_module_unlock();
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	spin_lock_irq(&io_request_lock);
	ret = mpl_base_device_reset_handler(cmd);
	spin_unlock_irq(&io_request_lock);
#else
	spin_lock_irq(cmd->device->host->host_lock);
	ret = mpl_base_device_reset_handler(cmd);
	spin_unlock_irq(cmd->device->host->host_lock);
#endif
	mpl_module_lock();

	/* Wait for ALL IOs on this device to complete */
	i = 0;
	while(atomic_read(&xmit_path->device->io_count)) {
		mpl_module_unlock();
		set_current_state(TASK_UNINTERRUPTIBLE);
		schedule_timeout(HZ);
		mpl_module_lock();
		if (++i >= 30)
			break;
	}

	/* Wait for the deferral thread to complete all the aborted commands.
	 * We're holding the module_lock, which blocks irqs.  So
	 * deferral_queue_lock needs to either leave irq settings alone or do
	 * lock_irqsave, unlock_irq_restore.  We opt for the former. */
	i = 0;
	spin_lock(&deferral_queue_lock);
	while (!list_empty(&deferral_queue)) {
		spin_unlock(&deferral_queue_lock);
		mpl_module_unlock();
		set_current_state(TASK_UNINTERRUPTIBLE);
		schedule_timeout(HZ);
		mpl_module_lock();
		spin_lock(&deferral_queue_lock);
		if (++i >= 30)
			break;
	}
	spin_unlock(&deferral_queue_lock);

	/* Wait for the eh_deferral thread to complete all the timed out
	 * commands.  We're holding the module_lock, which blocks irqs.  So
	 * eh_deferral_queue_lock needs to either leave irq settings alone or
	 * do lock_irqsave, unlock_irq_restore.  We opt for the former.*/
	i = 0;
	spin_lock(&eh_deferral_queue_lock);
	while (!list_empty(&eh_deferral_queue)) {
		spin_unlock(&eh_deferral_queue_lock);
		mpl_module_unlock();
		set_current_state(TASK_UNINTERRUPTIBLE);
		schedule_timeout(HZ);
		mpl_module_lock();
		spin_lock(&eh_deferral_queue_lock);
		if (++i >= 30)
			break;
	}
	spin_unlock(&eh_deferral_queue_lock);

	/* Restore the command fields */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	cmd->host = tmp_host;
	cmd->target = tmp_target;
	cmd->lun = tmp_lun;
#else
	cmd->device = tmp_sdev;
#endif

	mpl_module_unlock();
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	spin_lock_irq(&io_request_lock);
#else
	spin_lock_irq(cmd->device->host->host_lock);
#endif
	return ret;
}

int
mpl_bus_reset_handler(struct scsi_cmnd *cmd)
{
	struct io_stub *stub;
	int ret, i;
	struct mpl_path *path=NULL, *xmit_path;
	struct mpl_host *host=NULL;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	struct Scsi_Host *tmp_host;
	unsigned int tmp_target;
	unsigned int tmp_lun;
#else
	struct scsi_device *tmp_sdev;
#endif
	/* lock is held when called by SCSI layer */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	spin_unlock_irq(&io_request_lock);
#else
	spin_unlock_irq(cmd->device->host->host_lock);
#endif
	mpl_module_lock();

	host  = mpl_find_host_by_scsi_host(cmd->device->host);
	if (!host) {
		mpl_module_unlock();
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
		spin_lock_irq(&io_request_lock);
#else
		spin_lock_irq(cmd->device->host->host_lock);
#endif
		return FAILED;
	}

	path = mpl_find_path(host, cmd->device->id, cmd->device->lun);
	if (!path) {
		mpl_module_unlock();
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
		spin_lock_irq(&io_request_lock);
#else
		spin_lock_irq(cmd->device->host->host_lock);
#endif
		return FAILED;
	}

	if (mpl_is_path_configured(path))
		xmit_path = path->device->last_path;
	else
		xmit_path = path;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	tmp_host = cmd->host;
	tmp_target = cmd->target;
	tmp_lun = cmd->lun;
	cmd->host = xmit_path->host->scsi_host;
	cmd->target = xmit_path->target;
	cmd->lun = xmit_path->lun;
#else
	tmp_sdev = cmd->device;
	cmd->device = xmit_path->sdev;
#endif

	stub = mpl_find_stub_remove_pend_io(cmd,
					    MPL_ABORT_HOST | MPL_ABORT_IGNORE_ABORTED);

	mpl_print(MPL_IOPATH, MPL_INFO "BUS RESET %p, %d/%d/%d stub %p\n",
		  cmd, mpl_host_no_from_cmd(cmd),
		  mpl_target_no_from_cmd(cmd),
		  mpl_lun_no_from_cmd(cmd), stub);
	while (stub) {
		stub->cmd->result = ScsiResult(DID_ERROR, 0);
		mpl_add_deferral(stub, MPL_ABORTED_CMD);
		stub = mpl_find_stub_remove_pend_io(cmd, 
						    MPL_ABORT_HOST | MPL_ABORT_IGNORE_ABORTED);
	}
	mpl_module_unlock();
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	spin_lock_irq(&io_request_lock);
	ret = mpl_base_bus_reset_handler(cmd);
	spin_unlock_irq(&io_request_lock);
#else
	spin_lock_irq(cmd->device->host->host_lock);
	ret = mpl_base_bus_reset_handler(cmd);
	spin_unlock_irq(cmd->device->host->host_lock);
#endif
	mpl_module_lock();

	/* Wait for ALL IOs on this host to complete */
	i = 0;
	while (atomic_read(&xmit_path->host->io_count)) {
		mpl_module_unlock();
		set_current_state(TASK_UNINTERRUPTIBLE);
		schedule_timeout(HZ);
		mpl_module_lock();
		if (++i >= 30)
			break;
	}

 	/* Wait for the deferral thread to complete all the aborted commands.
	 * We're holding the module_lock, which blocks irqs.  So
	 * deferral_queue_lock needs to either leave irq settings alone or do
	 * lock_irqsave, unlock_irq_restore.  We opt for the former. */
	i = 0;
	spin_lock(&deferral_queue_lock);
	while (!list_empty(&deferral_queue)) {
		spin_unlock(&deferral_queue_lock);
		mpl_module_unlock();
		set_current_state(TASK_UNINTERRUPTIBLE);
		schedule_timeout(HZ);
		mpl_module_lock();
		spin_lock(&deferral_queue_lock);
		if (++i >= 30)
			break;
	}
	spin_unlock(&deferral_queue_lock);

 	/* Wait for the eh_deferral thread to complete all the timedout
	 * commands We're holding the module_lock, which blocks irqs.  So
	 * eh_deferral_queue_lock needs to either leave irq settings alone or
	 * do lock_irqsave, unlock_irq_restore.  We opt for the former. */
	i = 0;
	spin_lock(&eh_deferral_queue_lock);
	while (!list_empty(&eh_deferral_queue)) {
		spin_unlock(&eh_deferral_queue_lock);
		mpl_module_unlock();
		set_current_state(TASK_UNINTERRUPTIBLE);
		schedule_timeout(HZ);
		mpl_module_lock();
		spin_lock(&eh_deferral_queue_lock);
		if (++i >= 30)
			break;
	}
	spin_unlock(&eh_deferral_queue_lock);

	/* Restore the command fields */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	cmd->host = tmp_host;
	cmd->target = tmp_target;
	cmd->lun = tmp_lun;
#else
	cmd->device = tmp_sdev;
#endif

	mpl_module_unlock();
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	spin_lock_irq(&io_request_lock);
#else
	spin_lock_irq(cmd->device->host->host_lock);
#endif
	return ret;
}

/* 
 *  This function is called when all the low level hosts are detached from
 *  the multipulse driver. This function cleans up all the hosts, devices,
 *  and paths when a low level driver is detached.
 */
static void
mpl_destroy_all_hosts(void) 
{
	unsigned long flags;
	struct list_head *l, *n;
	struct mpl_host *host;
	struct mpl_target *tgtp;
	struct mpl_device *device, *tmp;
	struct io_stub *stub, *prev;
	int  i;

	while (atomic_read(&int_cmd_cnt)) {
		/* give it a chance */
		set_current_state(TASK_UNINTERRUPTIBLE);
		schedule_timeout(2);
	}
	mpl_module_lock();
	if (stub_hash) {
		spin_lock_irqsave(&stub_hash_table_lock, flags);
		for (i = 0; i < stub_hash_size; ++i) {
			stub = stub_hash[i];
			if (!stub) {
				continue;
			}
			while (stub) {
				prev = stub;
				stub = stub->next;
				mpl_print(0, MPL_ERR "mpl_destroy_all_hosts found active "
					  "stub while unloading lpfc driver stub %p",
					  prev);
				/* free resources associated with prev */
				if (!prev->done || prev->done == mpl_hb_done ||
				    prev->done == mpl_internal_done) {
					/* Internal command */
					mpl_module_unlock();
					del_timer_sync(&prev->cmd->eh_timeout);
					mpl_module_lock();
					
					if (prev->cmd->request_buffer) {
						mpl_put_scsi_buf(prev->cmd->request_buffer);
					}
					mpl_put_scsi_cmd(prev->cmd);
					atomic_dec(&int_cmd_cnt);
				}
				stub_hash[i] = stub;
				mpl_stub_free(prev);
			}
			stub_hash[i] = 0;
		}
		spin_unlock_irqrestore(&stub_hash_table_lock, flags);
	}
	
	if (device_hash) {
		spin_lock_irqsave(&device_hash_table_lock, flags);
		for (i = 0; i < device_hash_size; ++i) {
			struct mpl_path *path;
			struct mpl_device *next_dev;
			
			device = device_hash[i];
			spin_unlock_irqrestore(&device_hash_table_lock, flags);

			while (device) {
				mpl_device_state_machine(device, NULL, DEV_DISABLE);
				list_for_each_safe(l, n, &device->paths) {
					path = list_entry(l, struct mpl_path,
							  path_list);
					/* We don't really need to delink the
					 * path since we're going to free the
					 * list header when we
					 * kfree(device) 
					 */
					mpl_free_path(path);
				}
				next_dev = device->next;
				mpl_set_device_configured(device, CONFIG_UNCONFIGURED);
				mpl_module_unlock();
				del_timer_sync(&device->nopath);
				mpl_module_lock();
				list_del(&device->devices);
				kfree(device);
				device = next_dev;
			}
			spin_lock_irqsave(&device_hash_table_lock, flags);
			device_hash[i] = 0;
		}
		spin_unlock_irqrestore(&device_hash_table_lock, flags);
	}

	/* Double check to ensure reaper queue is empty */
	list_for_each_entry_safe(device, tmp, &reaper_queue, devices) {
		mpl_device_state_machine(device, NULL, DEV_DISABLE);
		list_for_each_safe(l, n, &device->paths) {
			struct mpl_path *path;
			path = list_entry(l, struct mpl_path, path_list);
			mpl_free_path(path);
		}
		mpl_set_device_configured(device, CONFIG_UNCONFIGURED);
		mpl_module_unlock();
		del_timer_sync(&device->nopath);
		mpl_module_lock();
		list_del(&device->devices);
		kfree(device);
	}

	list_for_each_safe(l, n, &mpl_targets) {
		tgtp = list_entry(l, struct mpl_target, target_list);
		list_del(&tgtp->target_list);
		kfree(tgtp);
	}

	list_for_each_safe(l, n, &mpl_hosts) {
		host = list_entry(l, struct mpl_host, host_list);
		list_del(&host->host_list);
		kfree(host);
	}

	INIT_LIST_HEAD(&deferral_queue);
	INIT_LIST_HEAD(&eh_deferral_queue);

	mpl_module_unlock();
}

static void 
mpl_timeout_blocked_host(unsigned long data)
{
	struct mpl_host *host = (struct mpl_host *)data;
	struct mpl_path *path;
	uint32_t hash;
	unsigned long flags;

	mpl_print(MPL_RECOVERY, MPL_INFO "mpl_timeout_blocked_host %d\n",
		  host->host_no);

	if (mpl_host_is_offline(host->hba)) {
		mpl_module_lock();
		host->blocked = 0;
		spin_lock_irqsave(&path_hash_table_lock, flags);
		for(hash=0; hash<path_hash_size; hash++) {
			for (path = path_hash[hash]; path; path = path->next) {
				if (path->host == host) {
					spin_unlock_irqrestore(&path_hash_table_lock, flags);
					mpl_path_state_machine(path, PATH_DOWN);
					spin_lock_irqsave(&path_hash_table_lock, flags);
				}
			}
		}
		spin_unlock_irqrestore(&path_hash_table_lock, flags);
		mpl_module_unlock();
		scsi_unblock_requests(host->scsi_host);
		mpl_wake_thread();
		return;
	}
	mod_timer(&host->host_block, jiffies + HZ);
	return;
}

/* Ensure can_queue for all mpl hosts are the same, ie the
 * minimum Scsi_Host value for can_queue.
 */
void
mpl_setup_can_queue(struct mpl_host *newhost)
{
	struct Scsi_Host * sh;
	struct mpl_host *host;

	sh = newhost->scsi_host;
	newhost->can_queue = sh->can_queue;
        sh->can_queue -= 10; /* Allow some space for mpl HeartBeat commands */

	if (!sh || (sh->can_queue >= mpl_max_can_queue) || !sh->can_queue)
		return;
	mpl_max_can_queue = sh->can_queue;

	/* A side effect of this is, if one HBA has a smaller can_queue, ALL
	 * HBAs will get that can_queue.
	 */
	list_for_each_entry(host, &mpl_hosts, host_list) {
		host->scsi_host->can_queue = sh->can_queue;
	}
	return;
}

void
scsi_block_requests_wrapped(struct Scsi_Host * sh)
{
	struct mpl_host *host;
	struct timer_list *timer;

	mpl_module_lock();
	host  = mpl_find_host_by_scsi_host(sh);
	mpl_module_unlock();
	if (!host) {
		return;
	}
	if (!host->blocked) {
		timer = &host->host_block;
		init_timer(timer);
		timer->data = (unsigned long)host;
		timer->expires = jiffies + HZ;
		timer->function = mpl_timeout_blocked_host;
		add_timer(timer);
	}
	scsi_block_requests(sh);
	host->blocked = 1;
	return;
}

void
scsi_unblock_requests_wrapped(struct Scsi_Host * sh)
{
	struct mpl_host *host;

	mpl_module_lock();
	host  = mpl_find_host_by_scsi_host(sh);
	mpl_module_unlock();
	if (!host) {
		return;
	}
	if (host->blocked) {
		del_timer_sync(&host->host_block);
	}
	host->blocked = 0;
	scsi_unblock_requests(sh);
	mpl_wake_thread();
	return;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
struct Scsi_Host *
scsi_register_wrapped(Scsi_Host_Template *tpnt, int j)
{
	struct mpl_host *host;

	mpl_module_lock();
	host = mpl_find_or_create_host(host_cnt);
	mpl_module_unlock();
	if (!host) {
		return NULL;
	}
	host->scsi_host = scsi_register(tpnt, j);
	mpl_setup_can_queue(host);
	host_cnt++;
	return host->scsi_host;
}

void
scsi_unregister_wrapped(struct Scsi_Host * sh)
{
	
	scsi_unregister(sh);
	host_cnt--;
	/* 
	   If all the hosts are unregistered the lpfc driver is unloaded.
	   Delete all the hosts/paths/devices and reset mpl_base_detect
	*/
	if (host_cnt == 0) {
		mpl_base_detect = NULL;
		mpl_destroy_all_hosts();
	}
	return;
}


int
scsi_register_module_wrapped(int module_type, Scsi_Host_Template *tpnt)
{
	int ret;

	if (mpl_base_detect == NULL) {
		driver_unloading = 0;
		mpl_template = tpnt;
		mpl_base_detect = tpnt->detect;
		tpnt->detect = mpl_detect;
		if(tpnt->cmd_per_lun > 1)
			tpnt->cmd_per_lun--;
		mpl_base_release = tpnt->release;
		tpnt->release = mpl_release;
		mpl_base_queuecommand = tpnt->queuecommand;
		tpnt->queuecommand = mpl_queuecommand;
		mpl_base_abort_handler = tpnt->eh_abort_handler;
		tpnt->eh_abort_handler = mpl_abort_handler;
		mpl_base_device_reset_handler = tpnt->eh_device_reset_handler;
		tpnt->eh_device_reset_handler = mpl_device_reset_handler;
		mpl_base_bus_reset_handler = tpnt->eh_bus_reset_handler;
		tpnt->eh_bus_reset_handler = mpl_bus_reset_handler;
	}

	ret = scsi_register_module(module_type, tpnt);
	initial_scsi_scan_done = 1;
	return ret;
}


int
scsi_unregister_module_wrapped(int module_type, Scsi_Host_Template *tpnt)
{
	int ret;

	ret = scsi_unregister_module(module_type, tpnt);
	return ret;
}

#else  /* 2.6 KERNEL */

struct Scsi_Host *
scsi_host_alloc_wrapped(struct scsi_host_template *tpnt, int sz)
{
	struct mpl_host *host;

	if (mpl_base_queuecommand == NULL) {
		driver_unloading = 0;
		mpl_template = tpnt;
/* NOTE: detect is different for 2.6, look into pci_register_driver stub */
		mpl_base_detect = tpnt->detect;
		tpnt->detect = mpl_detect;
		mpl_base_release = tpnt->release;
		tpnt->release = mpl_release;
		mpl_base_queuecommand = tpnt->queuecommand;
		tpnt->queuecommand = mpl_queuecommand;
		mpl_base_abort_handler = tpnt->eh_abort_handler;
		tpnt->eh_abort_handler = mpl_abort_handler;
		mpl_base_device_reset_handler = tpnt->eh_device_reset_handler;
		tpnt->eh_device_reset_handler = mpl_device_reset_handler;
		mpl_base_bus_reset_handler = tpnt->eh_bus_reset_handler;
		tpnt->eh_bus_reset_handler = mpl_bus_reset_handler;
		mpl_base_slave_alloc = tpnt->slave_alloc;
		tpnt->slave_alloc = mpl_slave_alloc;
		mpl_base_slave_destroy = tpnt->slave_destroy;
		tpnt->slave_destroy = mpl_slave_destroy;
		tpnt->eh_strategy_handler = mpl_eh_strategy_handler;
	}

	mpl_module_lock();
	host = mpl_find_or_create_host(host_cnt);
	mpl_module_unlock();
	if (!host) {
		return NULL;
	}
	host->scsi_host = scsi_host_alloc(tpnt, sz);
	host_cnt++;
	return host->scsi_host;
}

int 
scsi_add_host_wrapped(struct Scsi_Host *shost, 
		      struct device *dev)
{
	struct mpl_host *host;

	mpl_module_lock();
	host = mpl_find_host_by_scsi_host(shost);
	if (!host) {
		mpl_print(0, MPL_ERR "scsi_add_host_wrapped cannot find host for" 
			  "shost %p host_no %d \n",
			  shost, shost->host_no);
	}

	if (host) {
		host->hba = (struct lpfc_hba *)shost->hostdata[0];
		mpl_get_hba_wwpn(host->hba, &host->wwpn[0]);
	        mpl_setup_can_queue(host);
	}
	mpl_module_unlock();
	return scsi_add_host(shost, dev);

}
void
scsi_remove_host_wrapped(struct Scsi_Host *sh) {
	struct mpl_host *host;
	unsigned long flags;
	struct Scsi_Host *shost;
	struct scsi_device *sdev, *tmp;
	struct mpl_path *prsn_path;
	/* 
	   The lpfc driver is going away. Remove all the scsi_devices exported 
	   by mpulse driver now. Removing scsi_devices after some of hosts
	   are dettached can cause  SCSI_CACHE_SYNC command to fail if
	   presentation path is not available.
	*/	
	mpl_module_lock();
	if (!driver_unloading) {
		list_for_each_entry(host, &mpl_hosts, host_list) {
			shost = host->scsi_host;
			mpl_module_unlock();
			spin_lock_irqsave(shost->host_lock, flags);
			list_for_each_entry_safe(sdev, tmp, &shost->__devices, siblings) {
				spin_unlock_irqrestore(shost->host_lock, flags);
				if (sdev->sdev_state == SDEV_BLOCK) {
					mpl_module_lock();
					prsn_path = mpl_find_path(host, sdev->id,
								  sdev->lun);
					mpl_module_unlock();
					if (prsn_path)
						del_timer_sync(&prsn_path->sdev_block);
					scsi_internal_device_unblock(sdev);
				}
				scsi_remove_device(sdev);
				spin_lock_irqsave(shost->host_lock, flags);
			}
			spin_unlock_irqrestore(shost->host_lock, flags);
			mpl_module_lock();
		}
	}

	driver_unloading = 1;
	mpl_module_unlock();

	scsi_remove_host(sh);
}

void
scsi_host_put_wrapped(struct Scsi_Host *sh)
{
	struct mpl_host *host;
	struct mpl_device *device;
	struct mpl_path *path;

	mpl_module_lock();

	/* Lets cancel all our timers before freeing scsi_host */
	list_for_each_entry(host, &mpl_hosts, host_list) {
		if (host->scsi_host != sh)
			continue;
		list_for_each_entry(device, &host->devices, devices) {
			mpl_device_state_machine(device, NULL, DEV_DISABLE);
			list_for_each_entry(path, &device->paths, path_list) {
				mpl_path_state_machine(path, PATH_DISABLE);
				mpl_set_path_configured(path, CONFIG_UNCONFIGURED);
			}
			mpl_set_device_configured(device, CONFIG_UNCONFIGURED);
			mpl_quiesce_device(device, 30);
		}
	}

	mpl_module_unlock();	
	scsi_host_put(sh);
	mpl_module_lock();
	host_cnt--;
	/* 
	   If all the hosts are unregistered the lpfc driver is unloaded.
	   Delete all the hosts/paths/devices and reset mpl_base_queuecommand
	*/
	if (host_cnt == 0) {
		mpl_module_unlock();
		mpl_base_queuecommand = NULL;
		mpl_destroy_all_hosts();
	} else {
		mpl_module_unlock();
	}
	return;
}

static int
mpl_get_path_by_sdev(struct mpl_path **pathp, struct scsi_device *sdev)
{
	struct mpl_host *host;
	int configured = 0;

	mpl_module_lock();
	host  = mpl_find_host_by_scsi_host(sdev->host);
	if (host) {
		*pathp = mpl_find_or_create_path(host, sdev->id, sdev->lun);
		configured = *pathp && mpl_is_path_configured(*pathp);
				/* The path sdev may be locally manufactured,
				 * potentially before a valid sdev is
				 * available.  So make sure we've got a valid
				 * request_queue associated with the path's
				 * sdev before upcalls to the mid-layer. */
		(*pathp)->sdev->request_queue = sdev->request_queue;
	}
	mpl_module_unlock();
	return configured;
}

static void
mpl_propagate_queue_depth(struct mpl_device *device)
{
	struct mpl_path *presentation;
	struct mpl_path *path = NULL;

	mpl_module_lock();
	presentation = mpl_first_path(device);
	list_for_each_entry(path, &device->paths, path_list) {
		path->sdev->queue_depth = presentation->sdev->queue_depth;
		path->sdev->ordered_tags = presentation->sdev->ordered_tags;
		path->sdev->simple_tags = presentation->sdev->simple_tags;
	}
	mpl_module_unlock();
}


int
scsi_track_queue_full_wrapped(struct scsi_device *sdev, int cnt)
{
	struct mpl_path *path = NULL;
	int configured = 0;
	int ret = cnt;

	configured = mpl_get_path_by_sdev(&path, sdev);
	if  (!configured) {
		ret = scsi_track_queue_full(sdev, cnt);
	} else {
		struct mpl_path *presentation = mpl_first_path(path->device);

		if (presentation) {
			ret = scsi_track_queue_full(presentation->sdev, cnt);
			mpl_propagate_queue_depth(path->device);
		}
	}
	return ret;
}

void
scsi_adjust_queue_depth_wrapped(struct scsi_device *sdev, int tag, int cnt)
{
	struct mpl_path *path = NULL;
	int configured = 0;

	configured = mpl_get_path_by_sdev(&path, sdev);
	if  (!configured) {
		scsi_adjust_queue_depth(sdev, tag, cnt);
	} else {
		struct mpl_path *presentation = mpl_first_path(path->device);

		scsi_adjust_queue_depth(sdev, tag, cnt);
		if (presentation) {
			scsi_adjust_queue_depth(presentation->sdev, tag, cnt);
			mpl_propagate_queue_depth(path->device);
		}
	}
}

void 
mpl_eh_timeout_handler(struct scsi_cmnd *scmd)
{
	struct io_stub *stub;	
	struct mpl_host *host = NULL;
	struct mpl_path *path = NULL;

	mpl_module_lock();

	host  = mpl_find_host_by_scsi_host(scmd->device->host);

	if (host)
		path = mpl_find_path(host, 
				     scmd->device->id, 
				     scmd->device->lun);

	/* 
	 * If commands timeout allowed number of times under heavy I/O we need
	 * to retry this command.  We let INQUIRIES slide in order to avoid
	 * hanging during mid-layer discovery.
	*/
	if (scmd->cmnd[0] != INQUIRY && scmd->retries + 1 >= scmd->allowed) {
		scmd->retries = 0;
	}

	scmd->eh_timeout.function = (void (*)(unsigned long)) mpl_eh_timeout_handler;
	stub = mpl_find_stub_remove_pend_io(scmd, MPL_ABORT_EXACT);
	if (stub) {
		mpl_module_unlock();
		scmd->result = ScsiResult(DID_TIME_OUT, 0);
		scmd->eh_timeout.data = (unsigned long)scmd;
		scmd->eh_timeout.expires = jiffies + scmd->timeout_per_command;
		scmd->eh_timeout.function = stub->timeout_handler;
		add_timer(&scmd->eh_timeout);
		mpl_module_lock();
		mpl_process_aborted_cmd(stub);
		mpl_module_unlock();
		return;
	}
	stub = mpl_find_stub_remove_quiet(scmd);
	if (stub) {
		/* The command is pending. We need to abort it.*/
		scmd->eh_timeout.function = stub->timeout_handler;
		mpl_add_eh_deferral(stub, MPL_DO_TMO);
		mpl_module_unlock();
		return;
	} else if (path->delayed_cmd && path->delayed_cmd->cmd == scmd) {
		scmd->eh_timeout.data = (unsigned long)scmd;
		scmd->eh_timeout.expires = jiffies + scmd->timeout_per_command;
		add_timer(&scmd->eh_timeout);
		mpl_module_unlock();
		return;
	} else {

		/* Looks like mpl_done already completed this command.
		   We might not hit this condition. */

		mpl_print(0, MPL_ERR
			"IO timeout: IO not owned 0x%lx <%d/%d/%d>  done %p\n",
			scmd->serial_number, scmd->device->host->host_no,
			scmd->device->id, scmd->device->lun, scmd->scsi_done);

		mpl_module_unlock();
		return;
	}

}
/*
  This is a very simple error recovery. Abort all the
  commands errored and retry.
*/
int
mpl_eh_strategy_handler(struct Scsi_Host *shost)
{
	unsigned long flags;
	LIST_HEAD(eh_work_q);
	LIST_HEAD(eh_done_q);
	struct list_head *lh, *lh_sf;
	int ret;
	struct scsi_cmnd *scmd;
	spin_lock_irqsave(shost->host_lock, flags);
	list_splice_init(&shost->eh_cmd_q, &eh_work_q);
	spin_unlock_irqrestore(shost->host_lock, flags);

	/* Abort all the commands in the eh_work_q */
	list_for_each_safe(lh, lh_sf, &eh_work_q) {
		scmd = list_entry(lh, struct scsi_cmnd, eh_entry);

		if (scmd->serial_number == 0) {
			scmd->device->host->host_failed--;
			scmd->state = SCSI_STATE_BHQUEUE;
			scmd->eh_eflags = 0;
			continue;
		}

		spin_lock_irqsave(scmd->device->host->host_lock, flags);
		ret = mpl_abort_handler(scmd);
		spin_unlock_irqrestore(scmd->device->host->host_lock, flags);

		if (ret != SUCCESS) {
			mpl_print(0, MPL_ERR "mpl_abort_handler failed seial_number 0x%lx <%d/%d/%d> \n",
				  scmd->serial_number, scmd->device->host->host_no,
				  scmd->device->id, scmd->device->lun);
		}

		scmd->device->host->host_failed--;
		scmd->state = SCSI_STATE_BHQUEUE;
		scmd->eh_eflags = 0;
	}
	
	list_for_each_safe(lh, lh_sf, &eh_work_q) {
		scmd = list_entry(lh, struct scsi_cmnd, eh_entry);
		list_del_init(lh);
		if (++scmd->retries < scmd->allowed) {
			scsi_delete_timer(scmd);
			scmd->state = SCSI_STATE_MLQUEUE;
			spin_lock_irqsave(scmd->device->host->host_lock, flags);
			scmd->device->host->host_busy--;
			spin_unlock_irqrestore(scmd->device->host->host_lock, flags);
			
			spin_lock_irqsave(&scmd->device->sdev_lock, flags);
			scmd->device->device_busy--;
			spin_unlock_irqrestore(&scmd->device->sdev_lock, flags);

			blk_insert_request(scmd->device->request_queue, 
					   scmd->request, 1, scmd, 1);
		} else {
			if (!scmd->result)
				scmd->result |= (DRIVER_TIMEOUT << 24);
			scsi_finish_command(scmd);
		}
	}
	return 0;
}

static void 
mpl_timeout_blocked_sdev(unsigned long data)
{
	struct scsi_device *sdev = (struct scsi_device *)data;
	mpl_print(0, MPL_ERR "mpl_timeout_blocked_sdev : blocked target timed out <%d/%d/%d> \n",
		  sdev->host->host_no, sdev->id, sdev->lun);
	scsi_internal_device_unblock(sdev);
}

static int 
mpl_path_block(struct mpl_path *path, int tmo)
{
	struct mpl_path    *prsn_path;
	struct scsi_device *prsn_sdev=NULL, *tmp_sdev;
	struct timer_list *timer;
	enum mpl_device_state prev_state;

	mpl_module_lock();
	prev_state = path->device->state;
	path->lld_blocked = 1;
	mpl_expedite_hbeat(path); 
	mpl_module_unlock();

	/* Find the presentation path of this device */
	prsn_path = list_entry(list_first(&path->device->paths),
			       struct mpl_path,
			       path_list);

	/* Find the SCSI device for the presentation path */
	shost_for_each_device(tmp_sdev, prsn_path->host->scsi_host) {
		if ((tmp_sdev->id == prsn_path->target) &&
		    (tmp_sdev->lun == prsn_path->lun)) {
			prsn_sdev = tmp_sdev;
			scsi_device_put(prsn_sdev);
			break;
		}
	}

	if (!prsn_sdev) {
		mpl_print(0, MPL_ERR "mpl_device_block cannot find scsi device for path <%d/%d/%d>\n",
			  prsn_path->host->scsi_host->host_no, prsn_path->target, prsn_path->lun);
		return 0;
	}
	/* 
	   If multipulse driver is able to find another path do not
	   block. Else block the queue.
	*/
	if ((path->device->state == DEV_FAILED) &&
	    (prev_state != DEV_FAILED) &&
	    (prsn_sdev != NULL) &&
	    (prsn_sdev->sdev_state != SDEV_BLOCK)) {
		/* The lpfc driver blocks this target for the timeout period */
		scsi_internal_device_block(prsn_sdev);
		timer = &prsn_path->sdev_block;
		init_timer(timer);
		timer->data = (unsigned long)prsn_sdev;
		timer->expires = jiffies + tmo * HZ;
		timer->function = mpl_timeout_blocked_sdev;
		add_timer(timer);
	}
	return 0;
}

static int
mpl_path_unblock(struct mpl_path *path)
{
	mpl_module_lock();
	path->lld_blocked = 0;

	mpl_expedite_hbeat(path);
	mpl_module_unlock();
	return 0;
}

void
mpl_scsi_device_unblock(struct mpl_device *device)
{
	struct mpl_path    *prsn_path;
	struct scsi_device *prsn_sdev=NULL, *tmp_sdev;

	mpl_module_unlock();
	/* Find the presentation path of this device */
	prsn_path = list_entry(list_first(&device->paths),
			       struct mpl_path,
			       path_list);

	/* Find the SCSI device for the presentation path */
	shost_for_each_device(tmp_sdev, prsn_path->host->scsi_host) {
		if ((tmp_sdev->id == prsn_path->target) &&
		    (tmp_sdev->lun == prsn_path->lun)) {
			prsn_sdev = tmp_sdev;
			scsi_device_put(prsn_sdev);
			break;
		}
	}

	if (!prsn_sdev) {
		mpl_print(0, MPL_ERR "mpl_device_unblock cannot find scsi device for path <%d/%d/%d>\n",
			  prsn_path->host->scsi_host->host_no, prsn_path->target, prsn_path->lun);
		mpl_module_lock();
		return ;
	}

	if (prsn_sdev->sdev_state == SDEV_BLOCK) {
		del_timer_sync(&prsn_path->sdev_block);
		scsi_internal_device_unblock(prsn_sdev);
	}
	mpl_module_lock();
	return ;
}

#if defined(RHEL_FC)
static int 
mpl_device_block(struct device *dev, void *data)
{
	struct scsi_device *sdev = to_scsi_device(dev);
	struct scsi_target *st  = (struct scsi_target *) data;
	int timeout = fc_starget_dev_loss_tmo(st);
	struct mpl_path *path;
	struct mpl_host *host;
	uint32_t hash;
	unsigned long flags;

	mpl_module_lock();
	host  = mpl_find_host_by_scsi_host(sdev->host);
	if (!host) {
		mpl_print(0, MPL_ERR "mpl_device_block cannot find host <%d/%d/%d> \n",
			  sdev->host->host_no, sdev->id, sdev->lun);
		mpl_module_unlock();
		return 1;
	}

	spin_lock_irqsave(&path_hash_table_lock, flags);
	for(hash=0; hash<path_hash_size; hash++) {
		for (path = path_hash[hash]; path; path = path->next) {
			if (path->host == host &&
			    path->target  == sdev->id)
			{
				spin_unlock_irqrestore(&path_hash_table_lock, 
							flags);
				mpl_module_unlock();
				mpl_path_block(path, timeout);
				mpl_module_lock();
				spin_lock_irqsave(&path_hash_table_lock, 
						   flags);
			}
		}
	}
	spin_unlock_irqrestore(&path_hash_table_lock, flags);
	mpl_module_unlock();

	return 1;
}

void
fc_target_block_wrapped(struct scsi_target *st)
{
	device_for_each_child(&st->dev, st, mpl_device_block);
	return;
}
#else /* not RHEL_FC -> is SLES_FC */
void
fc_target_block_wrapped(struct Scsi_Host *sh,  uint16_t id,
			struct timer_list *tmr, uint32_t tmo)
{
	struct mpl_host *host;
	struct mpl_path *path;
	uint32_t hash;
	unsigned long flags;

	mpl_module_lock();
	host  = mpl_find_host_by_scsi_host(sh);
	if (!host) {
		mpl_print(0, MPL_ERR "fc_target_block_wrapped cannot find host <%d/%d> \n",
			  sh->host_no, id);
		mpl_module_unlock();
		return ;
	}
	spin_lock_irqsave(&path_hash_table_lock, flags);
	for(hash=0; hash<path_hash_size; hash++) {
		for (path = path_hash[hash]; path; path = path->next) {
			if (path->host == host &&
			    path->target  == id)
			{
				spin_unlock_irqrestore(&path_hash_table_lock, 
							flags);
				mpl_module_unlock();
				mpl_path_block(path, tmo);
				mpl_module_lock();
				spin_lock_irqsave(&path_hash_table_lock, 
						   flags);
			}
		}
	}
	spin_unlock_irqrestore(&path_hash_table_lock, flags);
	mpl_module_unlock();

	return;
}

#endif

#if defined(RHEL_FC)

static int
mpl_device_unblock(struct device *dev, void *data)
{
	struct scsi_device *sdev = to_scsi_device(dev);
	struct mpl_path *path;
	struct mpl_host *host;
	uint32_t hash;
	unsigned long flags;

	mpl_module_lock();
	host  = mpl_find_host_by_scsi_host(sdev->host);

	if (!host) {
		mpl_print(0, MPL_ERR "mpl_device_unblock cannot find host <%d/%d/%d> \n",
			sdev->host->host_no, sdev->id, sdev->lun);
		mpl_module_unlock();
		return 1;
	}

	spin_lock_irqsave(&path_hash_table_lock, flags);
	for(hash=0; hash<path_hash_size; hash++) {
		for (path = path_hash[hash]; path; path = path->next) {
			if (path->host == host &&
			    path->target  == sdev->id)
			{
				spin_unlock_irqrestore(&path_hash_table_lock, 
							flags);
				mpl_module_unlock();
				mpl_path_unblock(path);
				mpl_module_lock();
				spin_lock_irqsave(&path_hash_table_lock, 
						   flags);
			}
		}
	}
	spin_unlock_irqrestore(&path_hash_table_lock, flags);
	mpl_module_unlock();
        return 1;
}

void
fc_target_unblock_wrapped(struct scsi_target *st)
{
	mpl_module_lock();
	if (!driver_unloading) {
		mpl_module_unlock();
		device_for_each_child(&st->dev, st, mpl_device_unblock);
		mpl_wake_thread();
	} else
		mpl_module_unlock();
	return;
}
#else /* not RHEL_FC -> is SLES_FC */

void
fc_target_unblock_wrapped(struct Scsi_Host *sh, uint16_t id,
	struct timer_list *tmr)
{
	struct mpl_host *host;
	struct mpl_path *path;
	uint32_t hash;
	unsigned long flags;

	mpl_module_lock();
	host  = mpl_find_host_by_scsi_host(sh);

	if (!host) {
		mpl_print(0, MPL_ERR "fc_target_unblock_wrapped cannot find host <%d/%d> \n",
			sh->host_no, id);
		mpl_module_unlock();
		return ;
	}

	spin_lock_irqsave(&path_hash_table_lock, flags);
	for(hash=0; hash<path_hash_size; hash++) {
		for (path = path_hash[hash]; path; path = path->next) {
			if (path->host == host &&
			    path->target  == id)
			{
				spin_unlock_irqrestore(&path_hash_table_lock, 
							flags);
				mpl_module_unlock();
				mpl_path_unblock(path);
				mpl_module_lock();
				spin_lock_irqsave(&path_hash_table_lock, 
						   flags);
			}
		}
	}
	spin_unlock_irqrestore(&path_hash_table_lock, flags);
	mpl_module_unlock();

	return;
}

#endif

int
pci_register_driver_wrapped(struct pci_driver *dt)
{
	int ret;

	ret = pci_register_driver(dt);
/* Need to double check to see if scsi_scan for all HBA detected is done */
	initial_scsi_scan_done = 1;
	return ret;
}

void
pci_unregister_driver_wrapped(struct pci_driver *dt)
{
	pci_unregister_driver(dt);
	return;
}

#endif

static int __init
mpl_init(void)
{
	int  rv = -ENOMEM;

	printk(KERN_ERR "Emulex MultiPulse module version %s\n",
	       MPL_MODULE_VERSION);
	printk(KERN_ERR MPL_COPYRIGHT "\n");
	mpl_base_queuecommand = NULL;
	mpl_base_release = NULL;
	mpl_base_detect = NULL;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	mpl_base_select_queue_depths = NULL;
	/* In the 2.6 driver, lun queue depth is controlled entirely
	 * by scsi mid-layer */
#endif

	/* Initialize spinlocks */
	spin_lock_init(&mpl_module_spin_lock);
	spin_lock_init(&deferral_queue_lock);
	spin_lock_init(&eh_deferral_queue_lock);
	spin_lock_init(&device_hash_table_lock);
	spin_lock_init(&path_hash_table_lock);
	spin_lock_init(&stub_hash_table_lock);
	spin_lock_init(&pend_queue_lock);

	/* Verify configurable hbeat parameters */
	if ((mpl_hbeat_interval == 0) ||
	    (mpl_hbeat_interval > MPL_MAX_HBEAT_INTERVAL)) {
		mpl_print(0, MPL_ERR "mpl_hbeat_interval (%d)"
		   "is out of range [1-%d], resetting to default (%d)\n", 
		   mpl_hbeat_interval,
		   MPL_MAX_HBEAT_INTERVAL,  MPL_DFT_HBEAT_INTERVAL);
		mpl_hbeat_interval = MPL_DFT_HBEAT_INTERVAL;
	}
	if ((mpl_hbeat_fail_interval == 0) ||
	    (mpl_hbeat_fail_interval > MPL_MAX_HBEATFAIL_INTERVAL)) {
		mpl_print(0, MPL_ERR "mpl_hbeat_fail_interval (%d)"
		   "is out of range [1-%d], resetting to default (%d)\n", 
		   mpl_hbeat_fail_interval,
		   MPL_MAX_HBEATFAIL_INTERVAL,  MPL_DFT_HBEATFAIL_INTERVAL);
		mpl_hbeat_fail_interval = MPL_DFT_HBEATFAIL_INTERVAL;
	}
	if ((mpl_max_hbeat_retry == 0) ||
	    (mpl_max_hbeat_retry > MPL_MAX_HBEAT_RETRY)) {
		mpl_print(0, MPL_ERR "mpl_max_hbeat_retry (%d)"
		   "is out of range [1-%d], resetting to default (%d)\n", 
		   mpl_max_hbeat_retry,
		   MPL_MAX_HBEAT_RETRY,  MPL_DFT_HBEAT_RETRY);
		mpl_max_hbeat_retry = MPL_DFT_HBEAT_RETRY;
	}
	if(mpl_hbeat_fail_interval > mpl_hbeat_interval) {
		mpl_print(0, MPL_ERR "mpl_hbeat_fail_interval (%d)"
		   "is > mpl_hbeat_interval (%d), resetting to default (%d)\n",
		    mpl_hbeat_fail_interval, mpl_hbeat_interval,
		    MPL_DFT_HBEATFAIL_INTERVAL);
		mpl_hbeat_fail_interval = MPL_DFT_HBEATFAIL_INTERVAL;
	}

	io_stub_cache = kmem_cache_create("mpl_stub_io",
					  sizeof(struct io_stub), 0,
					  (SLAB_HWCACHE_ALIGN
#ifdef CONFIG_DEBUG_SLAB
					   | SLAB_POISON | SLAB_RED_ZONE
#endif
						  ),
					  NULL, NULL);
	if (!io_stub_cache)
		goto out;
				/* truncate hash_sizes to power of 2 */
	while (device_hash_size & (device_hash_size - 1)) {
		device_hash_size = device_hash_size & (device_hash_size - 1);
	}
	while (path_hash_size & (path_hash_size - 1)) {
		path_hash_size = path_hash_size & (path_hash_size - 1);
	}
	while (stub_hash_size & (stub_hash_size - 1)) {
		stub_hash_size = stub_hash_size & (stub_hash_size - 1);
	}
	device_hash = kmalloc(device_hash_size * sizeof(*device_hash),
			      GFP_KERNEL);
	if (!device_hash)
		goto out_free_cache;
	memset(device_hash, 0, device_hash_size * sizeof(*device_hash));
	path_hash = kmalloc(path_hash_size * sizeof(*path_hash),
			      GFP_KERNEL);
	if (!path_hash)
		goto out_free_device_hash;
	memset(path_hash, 0, path_hash_size * sizeof(*path_hash));
	stub_hash = kmalloc(stub_hash_size * sizeof(*stub_hash), GFP_KERNEL);
	if (!stub_hash)
		goto out_free_path_hash;
	memset(stub_hash, 0, stub_hash_size * sizeof(*stub_hash));
	rv = mpl_create_proc();
	if (rv)
		goto out_free_stub_hash;

	rv = mpl_create_mem_pool();
	if (rv)
		goto out_free_stub_hash;

	mpl_setup_multipulse();
	kernel_thread(mpl_thread, NULL, 0);
	kernel_thread(mpl_eh_thread, NULL, 0);

	goto out;

out_free_stub_hash:
	kfree(stub_hash);
	stub_hash = NULL;
out_free_path_hash:
	kfree(path_hash);
	path_hash = NULL;
out_free_device_hash:
	kfree(device_hash);
	device_hash = NULL;
out_free_cache:
	kmem_cache_destroy(io_stub_cache);
	io_stub_cache = NULL;
out:
	return rv;
}

static void __exit
mpl_exit(void)
{
	struct io_stub *stub;

	mpl_thread_quit = 1;
	mpl_wake_thread();
	down(&mpl_thread_gone);

	mpl_eh_thread_quit = 1;
	mpl_wake_eh_thread();
	down(&mpl_eh_thread_gone);

        mpl_destroy_all_hosts();

	mpl_module_lock();

	mpl_free_initial_data();
	if (device_hash) {
                kfree(device_hash);
                kfree(path_hash);
		device_hash = NULL;
	}

	mpl_remove_proc(); 

	/* Cleanup any unprocessed mpl_thread entries */
	while((stub = mpl_get_first_deferral())) {
	        mpl_inc_can_queue(stub);
		mpl_stub_free(stub);
	}
	while((stub = mpl_get_first_eh_deferral())) {
		mpl_stub_free(stub);
	}
	mpl_module_unlock();

	mpl_destroy_mem_pool();
	if (io_stub_cache) {
		kmem_cache_destroy(io_stub_cache);
		io_stub_cache = NULL;
	}
	if (stub_hash) {
		kfree(stub_hash);
		stub_hash = NULL;
	}
	return;
}

int debug_cnt = 0;

static void
mpl_hb_done(struct scsi_cmnd *cmd)
{
	struct io_stub *stub;
	struct mpl_path *path;
	struct mpl_device *device;

	/* If timer fired, this command will be handled by timeout handler */
        if (!del_timer(&cmd->eh_timeout)) {
		return;
	}

	mpl_module_lock();

	stub = mpl_find_stub_remove(cmd);
	if (!stub) {
		mpl_print(0, MPL_ERR "mpl: no heartbeat stub %d/%d/%d\n",
			  mpl_host_no_from_cmd(cmd),
			  mpl_target_no_from_cmd(cmd),
			  mpl_lun_no_from_cmd(cmd));
		mpl_module_unlock();
		return;
	}

	path = stub->path;
	device = path->device;
	mpl_print(MPL_INTERNAL_CMD, MPL_INFO "Cmpl cmd %x %x for %s %s "
		  "result %x sk %x asc %x ascq %x\n",
		  cmd->cmnd[0], cmd->cmnd[2],
		  mpl_print_device(device), mpl_print_path(path),
		  cmd->result, cmd->sense_buffer[2],
		  cmd->sense_buffer[12], cmd->sense_buffer[13]); 

	atomic_dec(&path->internal_count);
	mpl_inc_can_queue(stub);

	if (mpl_is_power_on_reset(cmd->result, cmd->sense_buffer) ||
	   (path->tpgs && mpl_is_sas_changed(cmd->result, cmd->sense_buffer)) ||
	   ((mpl_is_lu_not_ready(cmd->result, cmd->sense_buffer)) &&
		(path->hbstate == HB_TUR))) {
		/* Set the time out handler */
		if ( stub->cmd->cmnd[0] == INQUIRY ) {
			mpl_set_cmd_timer(stub, MPL_INQUIRY_CMD_TMO);
		} else if ( stub->cmd->cmnd[0] == TEST_UNIT_READY ) {
			mpl_set_cmd_timer(stub, MPL_TUR_CMD_TMO);
		} else {
			mpl_set_cmd_timer(stub, MPL_INTERNAL_CMD_TMO);
		}
		/* Just retry 29/00, QFULL and 04/01 errors on TUR */
		atomic_dec(&int_cmd_cnt);
		memset(&cmd->sense_buffer[0], 0, SCSI_SENSE_BUFFERSIZE);
		path->retries++;
		mpl_add_deferral(stub, MPL_HEARTBEAT_CMD);
		mpl_module_unlock();
		return;
	}

	if (cmd->cmnd[0] == REPORT_TGT_PORT_GROUP)  {
		device->tpg_cmd_inp = 0;
	}
	atomic_dec(&int_cmd_cnt);
	mpl_add_deferral(stub, MPL_HB_CMD);
	mpl_module_unlock();
	return;
}

#define hb_transition(state, event) ((state << 3) | event)

void
mpl_hbeat_state_machine(struct mpl_path *path, enum mpl_hb_event event)
{
	static char *states[] = {
		"HB_STOPPED",
		"HB_TUR",
		"HB_RTPG",
		"HB_INQSN",
		"HB_WAITING",
	};
	static char *events[] = {
		"HB_START",
		"HB_STOP",
		"HB_SUCCEED",
		"HB_FAIL",
		"HB_TIMEOUT",
		"HB_BUSY",
	};
	struct io_stub *stub;
	enum mpl_hb_state orig_hbstate = path->hbstate;
	int  expires = 0;

	if ((unsigned) path->hbstate > NUM_ENTRIES(states)) {
		mpl_print(0, MPL_ERR "Invalid path heartbeat state: %d\n",
			  path->hbstate);
		return;
	}
	if ((unsigned) event > NUM_ENTRIES(events)) {
		mpl_print(0, MPL_ERR "Invalid heartbeat event: %d\n", event);
		return;
	}

	mpl_print(MPL_HBSM, MPL_INFO "HbeatSM IN state %s, event %s "
		  "for %s %s blk %d (%p)\n",
		  states[path->hbstate], events[event],
		  mpl_print_device(path->device),
		  mpl_print_path(path),
		  (path->host->blocked || path->lld_blocked),
		  (&path)[-1]);

	switch (hb_transition(path->hbstate, event)) {
	/* HB_START event */
	case hb_transition(HB_TUR, HB_START):
	case hb_transition(HB_RTPG, HB_START):
	case hb_transition(HB_INQSN, HB_START):
		mpl_print(MPL_HEARTBEAT,
			  MPL_INFO "Heartbeat already started: %s, %s, "
			  "state %s, event %s from (%p)\n",
			  mpl_print_device(path->device),
			  mpl_print_path(path),
			  states[path->hbstate],
			  events[event],
			  (&path)[-1]);
		break;

	case hb_transition(HB_WAITING, HB_START):
		mpl_module_unlock();
		del_timer_sync(&path->hbeat);
		mpl_module_lock();
		/* Drop thru to initiate TUR */

	case hb_transition(HB_STOPPED, HB_START):
		/* Drop thru to initiate TUR */

	/* HB_TIMEOUT event */
	case hb_transition(HB_WAITING, HB_TIMEOUT):
		if (path->host->blocked || path->lld_blocked) {
			path->hbstate = HB_WAITING;
			expires = mpl_hbeat_fail_interval;
			break;
		}
		if (!path->host->scsi_host ||
		    !path->host->scsi_host->hostdata[0]) {
			path->hbstate = HB_WAITING;
			expires = mpl_hbeat_interval;
			break;
		}
		stub = mpl_setup_tur(path);
		if (!stub) {
			path->hbstate = HB_WAITING;
			expires = mpl_hbeat_interval;
		} else {
			path->hbstate = HB_TUR;
				/* Don't issue new commands in timer
				 * context.
				 */
			path->retries = 0;
			mpl_add_deferral(stub, MPL_HEARTBEAT_CMD);
		}
		break;

	case hb_transition(HB_STOPPED, HB_TIMEOUT):
	case hb_transition(HB_TUR, HB_TIMEOUT):
	case hb_transition(HB_RTPG, HB_TIMEOUT):
	case hb_transition(HB_INQSN, HB_TIMEOUT):
		break;

	case hb_transition(HB_TUR, HB_BUSY):
	case hb_transition(HB_RTPG, HB_BUSY):
	case hb_transition(HB_INQSN, HB_BUSY):
		path->hbstate = HB_WAITING;
		expires = mpl_hbeat_fail_interval;
		break;

	/* HB_SUCCEED event */
	case hb_transition(HB_TUR, HB_SUCCEED):
		if (path->device->alua_support && path->tpgs && 
			!path->device->tpg_cmd_inp &&
			!mpl_is_out_lun_own_chg(path->device)) {
			stub = mpl_setup_report_tgt_port(path);
			path->device->tpg_cmd_inp = 1;
			path->hbstate = HB_RTPG;
		}
		else {
			stub = mpl_setup_inquiry(path, (unsigned char)0x83);
			path->hbstate = HB_INQSN;
		}
		if (!stub) {
			path->hbstate = HB_WAITING;
			expires = mpl_hbeat_interval;
		} else {
			path->retries = 0;
			mpl_add_deferral(stub, MPL_HEARTBEAT_CMD);
		}
		break;

	case hb_transition(HB_RTPG, HB_SUCCEED):
		stub = mpl_setup_inquiry(path, (unsigned char)0x83);
		if (!stub) {
			path->hbstate = HB_WAITING;
			expires = mpl_hbeat_interval;
		} else {
			path->hbstate = HB_INQSN;
			path->retries = 0;
			mpl_add_deferral(stub, MPL_HEARTBEAT_CMD);
		}
		break;
		
	case hb_transition(HB_INQSN, HB_SUCCEED):
		path->failures = 0;
		if (mpl_is_path_configured(path)) {
			if ((!path->device->plugin) ||
				path->device->plugin->path_up(path)) {

				/* path_up plugin will return TRUE if path
				 * is really up. If extra steps need to be
				 * taken, PSM for PATH_UP will be called by
				 * plugin subsystem for device.
				 */
				mpl_path_state_machine(path, PATH_UP);

				/* If we were DEV_FAILED, PATH_UP will
				 * only bring us to DEV_FAILING.
				 */
				if (path->device->state == DEV_FAILING) {
					expires = mpl_hbeat_fail_interval;
					path->flag |= PATH_FAST_POLL;
				}
			}
			if (!expires)
				expires = mpl_hbeat_interval;
			path->hbstate = HB_WAITING;
		}
		else {
			path->hbstate = HB_STOPPED;
			mpl_path_state_machine(path, PATH_DOWN);
		}
		break;

	case hb_transition(HB_STOPPED, HB_SUCCEED):
	case hb_transition(HB_WAITING, HB_SUCCEED):
		break;

	/* HB_FAIL event */
	case hb_transition(HB_TUR, HB_FAIL):
	case hb_transition(HB_RTPG, HB_FAIL):
	case hb_transition(HB_INQSN, HB_FAIL):
		if (mpl_is_path_configured(path)) {
			path->hbstate = HB_WAITING;
			expires = mpl_hbeat_fail_interval;
			path->flag |= PATH_FAST_POLL;
			mpl_path_state_machine(path, PATH_DOWN);
		}
		else {
			/* TUR, RTPG or INQ while unconfigured only happens
			 * during initial Inquiry for a new path. In this case
			 * keep sending till we are successful or the internal
			 * retry exceeded.
			 */
			if (path->failures > mpl_max_hbeat_retry) {
				path->hbstate = HB_STOPPED;
				mpl_path_state_machine(path, PATH_DOWN);
				break;
			}
			path->hbstate = HB_WAITING;
			expires = mpl_hbeat_fail_interval;
			path->flag |= PATH_FAST_POLL;
			path->failures++;
		}
		break;

	case hb_transition(HB_WAITING, HB_FAIL):
		if (mpl_is_path_configured(path)) {
			path->hbstate = HB_WAITING;
			expires = mpl_hbeat_fail_interval;
			path->flag |= PATH_FAST_POLL;
			mpl_path_state_machine(path, PATH_DOWN);
		}
		else {
			path->hbstate = HB_STOPPED;
			mpl_path_state_machine(path, PATH_DOWN);
		}
		break;

	case hb_transition(HB_STOPPED, HB_FAIL):
		break;

	/* HB_STOP event */
	case hb_transition(HB_WAITING, HB_STOP):
	case hb_transition(HB_TUR, HB_STOP):
	case hb_transition(HB_INQSN, HB_STOP):
				/* We're called from path state machine to
				 * stop the heartbeat.  So, in the future, be
				 * carfeful of adding calls back into the path
				 * state machine here. */
		path->hbstate = HB_STOPPED;
		mpl_module_unlock();
		del_timer_sync(&path->hbeat);
		mpl_module_lock();

		mpl_path_state_machine(path, PATH_DOWN);
		break;

	case hb_transition(HB_STOPPED, HB_STOP):
		break;

	default:
		mpl_print(0, MPL_ERR "Odd heartbeat transition: %s, %s, "
			  "state %s, event %s from (%p)\n",
			  mpl_print_device(path->device),
			  mpl_print_path(path),
			  states[path->hbstate],
			  events[event],
			  (&path)[-1]);
	}
	if (path->hbstate == HB_WAITING) {
		mpl_module_unlock();
		mod_timer(&path->hbeat, jiffies + expires * HZ);
		mpl_module_lock();
	} else if (timer_pending(&path->hbeat)) {
				/* Don't del_timer_sync.  That can deadlocks
				 * us if this I/O completion interrupted
				 * __run_timers(). */
		mpl_module_unlock();
		del_timer(&path->hbeat);
		mpl_module_lock();
		mpl_print(MPL_TMO, MPL_INFO "State %s w/ active tmr on %s %s.  "
			  "Deleted..\n",
			  states[path->hbstate],
			  mpl_print_device(path->device),
			  mpl_print_path(path));
	}

	mpl_print(MPL_HBSM, MPL_INFO "HbeatSM state %s, event %s "
		  "result %s for %s %s (%p)\n",
		  states[orig_hbstate], events[event], states[path->hbstate],
		  mpl_print_device(path->device),
		  mpl_print_path(path),
		  (&path)[-1]);
	return;
}

#define path_transition(state, event) ((state << 3) | event)

void
mpl_path_state_machine(struct mpl_path *path, enum mpl_path_event event)
{
	static char *states[] = {
		"PATH_ONLINE",
		"PATH_FAILED",
		"PATH_STANDBY",
		"PATH_DISABLED",
	};
	static char *events[] = {
		"PATH_DOWN",
		"PATH_UP",
		"PATH_DISABLE",
		"PATH_ENABLE",
		"PATH_FAILBACK",
		"PATH_FAILOVER",
	};
	enum mpl_path_state orig_state = path->state;
	struct mpl_path *newpath;
	int log = 0;

	if ((unsigned) path->state > NUM_ENTRIES(states)) {
		mpl_print(0, MPL_ERR "Invalid path state: %d\n", path->state);
		return;
	}
	if ((unsigned) event > NUM_ENTRIES(events)) {
		mpl_print(0, MPL_ERR "Invalid path event: %d\n", event);
		return;
	}

	/* Filter out no action transitions */
	if ((event != PATH_UP) ||
		((path->state != PATH_ONLINE) && (path->state != PATH_STANDBY))) {
		log = 1;
		mpl_print(MPL_PSM, MPL_INFO "PSM IN %s %s state %s event %s (%p)\n",
			  mpl_print_device(path->device),
			  mpl_print_path(path),
			  states[orig_state],
			  events[event],
			  (&path)[-1]);
	}

	switch (path_transition(path->state, event)) {

	/* PATH_DOWN event */
	case path_transition(PATH_ONLINE, PATH_DOWN):
	case path_transition(PATH_STANDBY, PATH_DOWN):
				/* Must set state before plugin is called */
		path->state = PATH_FAILED;
		mpl_device_state_machine(path->device, path, DEV_PATH_DOWN);
		if (path->device->plugin) {
			path->device->plugin->path_down(path);
		}
		break;

	case path_transition(PATH_DISABLED, PATH_DOWN):
	case path_transition(PATH_FAILED, PATH_DOWN):
		path->flag &= ~PATH_JUST_ENABLED;
		break;


	/* PATH_DISABLE event */
	case path_transition(PATH_ONLINE, PATH_DISABLE):
	case path_transition(PATH_STANDBY, PATH_DISABLE):
		path->state = PATH_DISABLED;
		mpl_hbeat_state_machine(path, HB_STOP);
		mpl_device_state_machine(path->device, path, DEV_PATH_DOWN);

		mpl_module_unlock();
		del_timer_sync(&path->sched_cmd);
		mpl_module_lock();
		path->flag &= ~PATH_NEEDS_ALL_CMDS;

		path->flag = 0;
		if (path->device->plugin) {
			path->device->plugin->path_down(path);
		}
		break;

	case path_transition(PATH_FAILED, PATH_DISABLE):
		path->state = PATH_DISABLED;
		path->flag &= ~PATH_JUST_ENABLED;
		mpl_hbeat_state_machine(path, HB_STOP);

		mpl_module_unlock();
		del_timer_sync(&path->sched_cmd);
		mpl_module_lock();
		path->flag &= ~PATH_NEEDS_ALL_CMDS;
		break;

	case path_transition(PATH_DISABLED, PATH_DISABLE):
		break;

	/* PATH_UP event */
	case path_transition(PATH_FAILED, PATH_UP):
		mpl_device_state_machine(path->device, path, DEV_PATH_UP);
		path->state = (path == path->device->last_path
			       ? PATH_ONLINE
			       : PATH_STANDBY);

		/* If we just enabled the primary path, try to fail it back */
		if (mpl_is_path_primary(path) &&
			(path->flag & PATH_JUST_ENABLED))
			goto pfback;

		mpl_hbeat_state_machine(path, HB_START);
		break;

	case path_transition(PATH_ONLINE, PATH_UP):
	case path_transition(PATH_STANDBY, PATH_UP):
				/* Notify device we're still up. */
		mpl_device_state_machine(path->device, path, DEV_PATH_UP);
		path->state = (path == path->device->last_path
			       ? PATH_ONLINE
			       : PATH_STANDBY);
		break;

	case path_transition(PATH_DISABLED, PATH_UP):
		break;

	/* PATH_ENABLE event */
	case path_transition(PATH_DISABLED, PATH_ENABLE):
		path->state = PATH_FAILED;
		mpl_hbeat_state_machine(path, HB_START);
		path->flag |= PATH_JUST_ENABLED;
		break;

	case path_transition(PATH_FAILED, PATH_ENABLE):
	case path_transition(PATH_ONLINE, PATH_ENABLE):
		break;

	case path_transition(PATH_STANDBY, PATH_ENABLE):
		if (!mpl_is_path_primary(path)) {
			break;
		}
pfback:
		if(path->device->last_path != path)
			path->device->failing_path = path->device->last_path;
		path->device->last_path = path;
		/* Drop thru to FAILBACK */

	/* PATH_FAILBACK event */
	case path_transition(PATH_FAILED, PATH_FAILBACK):
	case path_transition(PATH_STANDBY, PATH_FAILBACK):
		/* Only one path can be online */
		list_for_each_entry(newpath, &path->device->paths, path_list) {
			if ((newpath != path) && (newpath->state == PATH_ONLINE)) {
				newpath->state = PATH_STANDBY;
			}
		}
		path->flag |= PATH_FAIL_BACK_EVT;
		path->state = PATH_STANDBY;
		if ((!path->device->plugin) ||
			path->device->plugin->path_up(path)) {

			/* path_up plugin will return TRUE if path
			 * is really up. If extra steps need to be
			 * taken, PSM for PATH_UP will be called by
			 * plugin subsystem for device.
			 */
			mpl_path_state_machine(path, PATH_UP);
		}
		break;

	case path_transition(PATH_ONLINE, PATH_FAILBACK):
	case path_transition(PATH_DISABLED, PATH_FAILBACK):
		break;

	/* PATH_FAILOVER event */
	case path_transition(PATH_STANDBY, PATH_FAILOVER):

		path->flag |= PATH_FAIL_OVER_EVT;
		path->state = PATH_ONLINE;
		/* Only one path can be online */
		list_for_each_entry(newpath, &path->device->paths, path_list) {
			if ((newpath != path) && (newpath->state == PATH_ONLINE)) {
				newpath->state = PATH_STANDBY;
			}
		}
		/* Restart Hbeat to force TUR */
		mpl_hbeat_state_machine(path, HB_START);
		break;

	case path_transition(PATH_FAILED, PATH_FAILOVER):
	case path_transition(PATH_ONLINE, PATH_FAILOVER):
	case path_transition(PATH_DISABLED, PATH_FAILOVER):
		break;
		
	default:
		mpl_print(0, MPL_ERR "Odd path state transition: %s, %s, "
			  "state %s, event %s from (%p)\n",
			  mpl_print_device(path->device),
			  mpl_print_path(path),
			  states[path->state],
			  events[event],
			  (&path)[-1]);
	}

	if (log) {
		mpl_print(MPL_PSM, MPL_INFO "PSM %s %s state %s event %s result %s (%p)\n",
			  mpl_print_device(path->device),
			  mpl_print_path(path),
			  states[orig_state],
			  events[event],
			  states[path->state],
			  (&path)[-1]);
	}
	return;
}

static int
mpl_select_next_path(struct mpl_device *device, struct mpl_path *first_path)
{
	struct mpl_path *path = first_path;
	int  found = 0;

				/* Find path on active controller if
				 * possible, preferring default path. */
	if (path->device->alua_support) {
		do {
			path = mpl_get_next_path(path);

			if (mpl_is_path_primary(path) &&
			   (path->tpgs && (path->sas == MPL_ALUA_ACTIVE_OPT)) &&
			   (path->state == PATH_ONLINE ||
			    path->state == PATH_STANDBY)) {
				found = 1;
				device->last_path = path;
				goto done;
			}
		} while (path != first_path);
		do {
			path = mpl_get_next_path(path);

			if ((path->tpgs && (path->sas == MPL_ALUA_ACTIVE_OPT)) &&
			    (path->state == PATH_ONLINE ||
			     path->state == PATH_STANDBY)) {
				found = 1;
				device->last_path = path;
				goto done;
			}
		} while (path != first_path);
	}
	else {
		do {
			path = mpl_get_next_path(path);

			if (mpl_is_path_primary(path) &&
			    path->active_controller &&
			    (path->state == PATH_ONLINE ||
			     path->state == PATH_STANDBY)) {
				found = 1;
				device->last_path = path;
				goto done;
			}
		} while (path != first_path);
		do {
			path = mpl_get_next_path(path);

			if (path->active_controller &&
			    (path->state == PATH_ONLINE ||
			     path->state == PATH_STANDBY)) {
				found = 1;
				device->last_path = path;
				goto done;
			}
		} while (path != first_path);
	}
	do {
		path = mpl_get_next_path(path);
		if (path->state == PATH_ONLINE ||
		    path->state == PATH_STANDBY) {
			found = 1;
				/* send start/stop on MSA */
			device->last_path = path;
			goto done;
		}
	} while (path != first_path);

done:
	return found;
}

#define device_action(state, event) ((state << 3) | event)

void
mpl_device_state_machine(struct mpl_device *device,
			 struct mpl_path *path,
			 enum mpl_device_event event)
{
	static char *states[] = {
		"DEV_ONLINE",
		"DEV_FAILING",
		"DEV_FAILED",
		"DEV_DISABLED",
	};
	static char *events[] = {
		"DEV_PATH_UP",
		"DEV_PATH_DOWN",
		"DEV_DISABLE",
	};
	struct mpl_path *last_path = device->last_path;
	enum mpl_device_state orig_state = device->state;
	int log = 0;

	if ((unsigned) device->state > NUM_ENTRIES(states)) {
		mpl_print(0, MPL_ERR "Invalid device state: %d\n",
			  device->state);
		return;
	}
	if ((unsigned) event > NUM_ENTRIES(events)) {
		mpl_print(0, MPL_ERR "Invalid device event: %d\n", event);
		return;
	}

	/* Filter out no action transitions */
	if ((event != DEV_PATH_UP) || (device->state != DEV_ONLINE) ||
		((path->state != PATH_ONLINE) && (path->state != PATH_STANDBY))) {
		log = 1;

		mpl_print(MPL_DSM, MPL_INFO "DSM IN %s state %s event %s (%p)\n",
			  mpl_print_device(device),
			  states[orig_state],
			  events[event],
			  (&device)[-1]);
	}

	switch (device_action(device->state, event)) {

	/* DEV_PATH_UP event */
	case device_action(DEV_ONLINE, DEV_PATH_UP):
		if (path == last_path) {
			mpl_device_process_pendq(device);
			break;
		}
		if (mpl_is_auto_failback(device) &&
		     mpl_is_path_primary(path)) {
			if(last_path != path)
				device->failing_path = last_path;
			device->last_path = path;
			/* If we changed to a new IO path */
			if (last_path != path) {
				mpl_path_state_machine(path, PATH_FAILBACK);
			}
		}
		break;

	case device_action(DEV_FAILED, DEV_PATH_UP):
		if(last_path != path)
			device->failing_path = last_path;
		device->last_path = path;
		device->state = DEV_FAILING;
		break;

	case device_action(DEV_FAILING, DEV_PATH_UP):
		if (path == last_path) {
			device->state = DEV_ONLINE;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
			mpl_scsi_device_unblock(device);
#endif
			if (timer_pending(&device->nopath)) {
				mpl_module_unlock();
				del_timer_sync(&device->nopath);
				mpl_module_lock();
			}
			mpl_device_process_pendq(device);
		}
		break;

	case device_action(DEV_DISABLED, DEV_PATH_UP):
		break;

	/* DEV_PATH_DOWN event */
	case device_action(DEV_ONLINE, DEV_PATH_DOWN):
		if (path != last_path) {
			break;
		}
		if (!mpl_select_next_path(device, path)) {
			device->state = DEV_FAILED;
			if (device->nopath_tmo) {
				mpl_module_unlock();
				mod_timer(&device->nopath,
					jiffies + device->nopath_tmo * HZ);
				mpl_module_lock();
			}
			else {
				mpl_device_process_pendq(device);
			}
		} else {
			device->failing_path = last_path;
			device->state = DEV_FAILING;
			/* If we changed to a new IO path */
			if (last_path != device->last_path) {
				mpl_path_state_machine(device->last_path, PATH_FAILOVER);
			}
		}

		break;

	case device_action(DEV_FAILING, DEV_PATH_DOWN):
		if (path != last_path) {
			break;
		}
		if (!mpl_select_next_path(device, path)) {
			device->state = DEV_FAILED;
			if (device->nopath_tmo) {
				mpl_module_unlock();
				mod_timer(&device->nopath,
					jiffies + device->nopath_tmo * HZ);
				mpl_module_lock();
			}
			else {
				mpl_device_process_pendq(device);
			}
		} else {
			device->failing_path = last_path;
			/* If we changed to a new IO path */
			if (last_path != device->last_path) {
				mpl_path_state_machine(device->last_path, PATH_FAILOVER);
			}
		}
		break;

	case device_action(DEV_DISABLED, DEV_PATH_DOWN):
	case device_action(DEV_FAILED, DEV_PATH_DOWN):
		break;

	/* DEV_DISABLE event */
	case device_action(DEV_ONLINE, DEV_DISABLE):
	case device_action(DEV_FAILING, DEV_DISABLE):
	case device_action(DEV_FAILED, DEV_DISABLE):
		device->state = DEV_DISABLED;
		mpl_device_process_pendq(device);
		break;

	case device_action(DEV_DISABLED, DEV_DISABLE):
		break;

	default:
		mpl_print(0, MPL_ERR "Odd device state/event: %s, "
			  "state %s, event %s from (%p)\n",
			  mpl_print_device(device),
			  states[device->state],
			  events[event],
			  (&path)[-1]);
	}
	
	if (log) {
		mpl_print(MPL_DSM, MPL_INFO "DSM %s state %s event %s "
			  "result %s newIO(%d)\n",
			  mpl_print_device(device),
			  states[orig_state],
			  events[event],
			  states[device->state],
			  (last_path != device->last_path));
	}
	return;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
EXPORT_SYMBOL(scsi_register_wrapped);
EXPORT_SYMBOL(scsi_unregister_wrapped);
EXPORT_SYMBOL(scsi_register_module_wrapped);
EXPORT_SYMBOL(scsi_unregister_module_wrapped);
#else
EXPORT_SYMBOL(scsi_host_alloc_wrapped);
EXPORT_SYMBOL(scsi_add_host_wrapped);
EXPORT_SYMBOL(scsi_host_put_wrapped);
EXPORT_SYMBOL(scsi_remove_host_wrapped);
EXPORT_SYMBOL(scsi_track_queue_full_wrapped);
EXPORT_SYMBOL(scsi_adjust_queue_depth_wrapped);
EXPORT_SYMBOL(pci_register_driver_wrapped);
EXPORT_SYMBOL(pci_unregister_driver_wrapped);
EXPORT_SYMBOL(fc_target_block_wrapped);
EXPORT_SYMBOL(fc_target_unblock_wrapped);
#endif
EXPORT_SYMBOL(scsi_block_requests_wrapped);
EXPORT_SYMBOL(scsi_unblock_requests_wrapped);

module_init(mpl_init);
module_exit(mpl_exit);
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Emulex Corporation - tech.support@emulex.com");
MODULE_DESCRIPTION("Emulex MultiPulse for Linux");
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
MODULE_VERSION(MPL_MODULE_VERSION);
#endif



/*
 * Local variables:
 *  compile-command: "gcc -D__KERNEL__ -I/usr/src/linux-`uname -r`/include -Wall -Wstrict-prototypes -Wno-trigraphs -O2 -fno-strict-aliasing -fno-common -Wno-unused -pipe -mpreferred-stack-boundary=2 -march=i686  -I. -DVARYIO=21 -DUSE_HIGHMEM_IO=2 -DUSE_SCHEDULER -I/lib/modules/`uname -r`/build/drivers/scsi -DMODULE -DMODVERSIONS -include /usr/src/linux-`uname -r`/include/linux/modversions.h -c lpfc_shim.c"
 * End:
 */
